﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using BenLincoln.DebugSymbols.SymDump;

namespace BenLincoln.DebugSymbols
{
    public class ObjectFileToSourceFileMap
    {
        public string ObjectFile;
        public Dictionary<string, int> SourceFilesWithCounts;

        public ObjectFileToSourceFileMap()
        {
            ObjectFile = "";
            SourceFilesWithCounts = new Dictionary<string, int>();
        }
    }

    public class CodeElementFileMapper
    {
        public Dictionary<string, ObjectFileToSourceFileMap> ObjectFileToSourceFileMapEntries;
        // TKTK: also build a map of memory offsets based on functions and SLD information, and use that as a secondary map 
        // for anything that's not categorized by the object file method
        // maybe use it as the primary method, since it should be more accurate WRT conflicts

        public CodeElementFileMapper()
        {
            ObjectFileToSourceFileMapEntries = new Dictionary<string, ObjectFileToSourceFileMap>();
        }

        public void GetDataFromCElement(BenLincoln.DebugSymbols.SymDump.CElement e)
        {
            if ((e.source_file != null) && (e.source_file != ""))
            {
                for (int i = 0; i < e.object_files.Count; i++)
                {
                    string objectFile = e.object_files[i];
                    ObjectFileToSourceFileMap oMap = new ObjectFileToSourceFileMap();
                    if (ObjectFileToSourceFileMapEntries.ContainsKey(objectFile))
                    {
                        oMap = ObjectFileToSourceFileMapEntries[objectFile];
                    }
                    else
                    {
                        oMap.ObjectFile = objectFile;
                    }
                    if (oMap.SourceFilesWithCounts.ContainsKey(e.source_file))
                    {
                        oMap.SourceFilesWithCounts[e.source_file]++;
                    }
                    else
                    {
                        oMap.SourceFilesWithCounts.Add(e.source_file, 1);
                    }
                    if (ObjectFileToSourceFileMapEntries.ContainsKey(objectFile))
                    {
                        ObjectFileToSourceFileMapEntries[objectFile] = oMap;
                    }
                    else
                    {
                        ObjectFileToSourceFileMapEntries.Add(objectFile, oMap);
                    }
                }
            }
        }

        public void GetAllDataFromSymbolCollection(BenLincoln.DebugSymbols.SymDump.SymDumpSymbolCollection syms)
        {
            foreach (CFunction func in syms.functions)
            {
                GetDataFromCElement(func);
            }
            foreach (CLabel l in syms.labels)
            {
                GetDataFromCElement(l);
            }
            foreach (CEnum e in syms.enums)
            {
                GetDataFromCElement(e);
            }
            foreach (CStruct s in syms.structs)
            {
                GetDataFromCElement(s);
            }
            foreach (CUnion u in syms.unions)
            {
                GetDataFromCElement(u);
            }
            foreach (CTypeDef t in syms.typedefs)
            {
                GetDataFromCElement(t);
            }
            foreach (CExternalDeclaration x in syms.external_declarations)
            {
                GetDataFromCElement(x);
            }
        }

        public void GetAllDataFromOverlay(BenLincoln.DebugSymbols.SymDump.PsyQOverlayDescription ol)
        {
            foreach (CFunction func in ol.functions)
            {
               GetDataFromCElement(func);
            }
            foreach (CLabel l in ol.labels)
            {
                GetDataFromCElement(l);
            }
            foreach (CEnum e in ol.enums)
            {
                GetDataFromCElement(e);
            }
            foreach (CStruct s in ol.structs)
            {
                GetDataFromCElement(s);
            }
            foreach (CUnion u in ol.unions)
            {
                GetDataFromCElement(u);
            }
            foreach (CTypeDef t in ol.typedefs)
            {
                GetDataFromCElement(t);
            }
            foreach (CExternalDeclaration x in ol.external_declarations)
            {
                GetDataFromCElement(x);
            }
        }

        public void SetBestSourceFileMatch(BenLincoln.DebugSymbols.SymDump.CElement e, string elementName, string defaultSourceFileName, StringBuilder manualChangesBuilder)
        {
            if ((e.source_file != null) && (e.source_file != ""))
            {
                return;
            }
            string result = defaultSourceFileName;
            int highestCount = 0;
            List<string> sourceFileNames = new List<string>();
            Dictionary<string, int> sourceFileRefCounts = new Dictionary<string, int>();
            foreach (string objectFile in e.object_files)
            {
                if (ObjectFileToSourceFileMapEntries.ContainsKey(objectFile))
                {
                    ObjectFileToSourceFileMap oMap = ObjectFileToSourceFileMapEntries[objectFile];

                    foreach (string sourceFileName in oMap.SourceFilesWithCounts.Keys)
                    {
                        if (!sourceFileNames.Contains(sourceFileName))
                        {
                            sourceFileNames.Add(sourceFileName);
                        }
                        int count = oMap.SourceFilesWithCounts[sourceFileName];
                        if (!sourceFileRefCounts.Keys.Contains(sourceFileName))
                        {
                            sourceFileRefCounts.Add(sourceFileName, oMap.SourceFilesWithCounts[sourceFileName]);
                        }
                        else
                        {
                            count = sourceFileRefCounts[sourceFileName] + oMap.SourceFilesWithCounts[sourceFileName];
                            sourceFileRefCounts[sourceFileName] = count;
                        }
                        if (count > highestCount)
                        {
                            highestCount = count;
                        }
                    }
                }
                sourceFileNames.Sort();
                for (int i = 0; i < sourceFileNames.Count; i++)
                {
                    if (sourceFileRefCounts[sourceFileNames[i]] == highestCount)
                    {
                        result = sourceFileNames[i];
                        break;
                    }
                }

                if (sourceFileNames.Count > 1)
                {
                    string objectFiles = BenLincoln.DebugSymbols.Utilities.GetListAsString(e.object_files);
                    manualChangesBuilder.Append(string.Format("The code element '{0}' was mapped to the source file '{1}' based on its association with object file(s) '{2}'. However, multiple source files mapped to that object file, with the following counts of elements:", elementName, result, objectFiles));
                    manualChangesBuilder.Append(Environment.NewLine);
                    foreach (string sfn in sourceFileNames)
                    {
                        manualChangesBuilder.Append("\t");
                        manualChangesBuilder.Append(string.Format("{0}: {1}", sfn, sourceFileRefCounts[sfn]));
                        manualChangesBuilder.Append(Environment.NewLine);
                    }
                    manualChangesBuilder.Append("If you would rather use one of the other destinations, update the JSON file manually before passing it to PopulateSkeleton.");
                    manualChangesBuilder.Append(Environment.NewLine);
                }

            }
            e.source_file = result;
        }
    }
}
