﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

namespace BenLincoln.DebugSymbols
{
    public class InputValidator
    {
        public const string DefaultReplacementForInvalidChars = "";

        public enum EscapingMode
        {
            Hex,
            Octal
        }

        public static string EscapeNonPrintableCharactersInASCIICString(string text, EscapingMode escapeMode)
        {
            string result = text;
            if ((escapeMode != EscapingMode.Hex) && (escapeMode != EscapingMode.Octal))
            {
                throw new Exception(string.Format("Unsupported escape mode: {0}", escapeMode));
            }

            result = result.Replace("\n", "\\n");
            result = result.Replace("\r", "\\r");
            result = result.Replace("\t", "\\t");

            StringBuilder builder = new StringBuilder();

            byte[] charBytes = ASCIIEncoding.ASCII.GetBytes(result);

            for (int i = 0; i < charBytes.Length; i++)
            {
                bool handled = false;
                if ((charBytes[i] < 32) || (charBytes[i] > 126))
                {
                    if (escapeMode == EscapingMode.Hex)
                    {
                        builder.Append(string.Format("\\x{0:X2}", charBytes[i]));
                    }
                    if (escapeMode == EscapingMode.Octal)
                    {
                        builder.Append(string.Format("\\0{0}", Convert.ToString(charBytes[i], 8)));
                    }
                    handled = true;
                }
                if (!handled)
                {
                    builder.Append((char)charBytes[i]);
                }
            }

            return builder.ToString();
        }

        // formats the content of a C string so that it shouldn't be able to escape out of quotes 
        // when placed inside double quotes
        public static string GetReasonablySafeCString(string text)
        {
            string result = text;

            Regex quoteRex = new Regex("([^\\])\"");
            MatchCollection quoteMatches = quoteRex.Matches(result);
            foreach (Match m in quoteMatches)
            {
                string replacementText = string.Format("{0}\\\"", m.Value.Substring(0, 1));
                result = result.Replace(m.Value, replacementText);
            }

            result = EscapeNonPrintableCharactersInASCIICString(result, EscapingMode.Hex);

            return result;
        }

        // formats the content of a Java string so that it shouldn't be able to escape out of quotes 
        // when placed inside double quotes
        public static string GetReasonablySafeJavaString(string text)
        {
            string result = text;

            Regex quoteRex = new Regex("([^\\\\])\"");
            MatchCollection quoteMatches = quoteRex.Matches(result);
            foreach (Match m in quoteMatches)
            {
                string replacementText = string.Format("{0}\\\"", m.Value.Substring(0, 1));
                result = result.Replace(m.Value, replacementText);
            }

            result = EscapeNonPrintableCharactersInASCIICString(result, EscapingMode.Octal);

            return result;
        }

        // replaces characters which are not valid in the name of a C item
        public static string GetReasonablySafeCName(string name, string replacement)
        {
            string result = name;

            StringBuilder builder = new StringBuilder();

            byte[] charBytes = ASCIIEncoding.ASCII.GetBytes(result);

            byte[] invalidCharNums = ASCIIEncoding.ASCII.GetBytes("!\"#$%&'()*+,/:;<=>?@[\\]^`{|}~");

            for (int i = 0; i < charBytes.Length; i++)
            {
                bool handled = false;
                if (charBytes[i] < 32)
                {
                    builder.Append(replacement);
                    handled = true;
                }
                if (!handled && (charBytes[i] > 126))
                {
                    builder.Append(replacement);
                    handled = true;
                }
                if (!handled && (invalidCharNums.Contains(charBytes[i])))
                {
                    builder.Append(replacement);
                    handled = true;
                }
                if (!handled)
                {
                    builder.Append((char)charBytes[i]);
                }
            }

            return builder.ToString();
        }

        public static string GetReasonablySafeCNameWithDefaultReplacement(string name)
        {
            return GetReasonablySafeCName(name, DefaultReplacementForInvalidChars);
        }
    }
}
