﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

using BenLincoln.DebugSymbols.SymDump;

namespace BenLincoln.DebugSymbols
{
    public class DupeCounter
    {
        protected Hashtable _Counts;

        public DupeCounter()
        {
            _Counts = new Hashtable();
        }

        public void HandleName(string name)
        {
            if (_Counts.ContainsKey(name))
            {
                _Counts[name] = ((int)_Counts[name]) + 1;
            }
            else
            {
                _Counts.Add(name, 1);
            }
        }

        public List<string> GetDuplicateNames()
        {
            List<string> result = new List<string>();
            foreach (string k in _Counts.Keys)
            {
                int count = (int)_Counts[k];
                if (count > 1)
                {
                    result.Add(k);
                }
            }
            return result;
        }

        public string GetDupeList()
        {
            StringBuilder builder = new StringBuilder();
            List<string> keyNames = new List<string>();
            foreach (string k in _Counts.Keys)
            {
                keyNames.Add(k);
            }
            keyNames.Sort();
            foreach (string k in keyNames)
            {
                int count = (int)_Counts[k];
                if (count > 1)
                {
                    builder.Append(string.Format("\t{0} - {1} definitions with this name{2}", k, count, Environment.NewLine));
                }
            }

            return builder.ToString();
        }

        public string GetResultOutput(string objectType)
        {
            StringBuilder builder = new StringBuilder();
            string result = GetDupeList();
            if (result != "")
            {
                builder.Append(string.Format("Warning: at least one {0} definition with a duplicate name was detected. You will need to manually resolve any conflicts by renaming the additional definitions in the JSON file, as well as any references to them in the same file, then re-running CreateSkeleton.", objectType));
                builder.Append(Environment.NewLine);
                builder.Append(result);
                builder.Append(Environment.NewLine);
            }
            return builder.ToString();
        }
    }

    public class SourceCodeGenerationResult
    {
        public string SourceCode = "";
        public StringBuilder ManualChangesBuilder = new StringBuilder();
    }

    public class MonolithicHeaderGenerationResult : SourceCodeGenerationResult
    {
        
    }

    public class IndividualSourceFileContent
    {
        public string HeaderContent;
        public string BodyContent;
        public string FooterContent;

        public IndividualSourceFileContent()
        {
            HeaderContent = "";
            BodyContent = "";
            FooterContent = "";
        }

        public string GetAllContent()
        {
            return string.Format("{0}{1}{2}{3}{4}",
                    HeaderContent,
                    Environment.NewLine,
                    BodyContent,
                    Environment.NewLine,
                    FooterContent);
        }
    }

    public class IndividualSourceFileGenerationResult : SourceCodeGenerationResult
    {
        public Dictionary<string, IndividualSourceFileContent> OutputFiles = new Dictionary<string, IndividualSourceFileContent>();
    }

    public class SourceCodeGenerator
    {
        public const string DefaultOrUnmappedFilename = "THISDUST";

        public static MonolithicHeaderGenerationResult GenerateMonolithicHeader(List<CEnum> enums, 
            List<CStruct> structs, List<CUnion> unions, List<CTypeDef> typedefs, List<CTypeDef> generated_typedefs,
            List<CExternalDeclaration> externs, string projectName,
            StringBuilder manualChangesBuilder)
        {
            // create monolithic header file
            MonolithicHeaderGenerationResult result = new MonolithicHeaderGenerationResult();
            StringBuilder mHeaderFileBuilder = new StringBuilder();

            mHeaderFileBuilder.Append(String.Format("#ifndef {0}_MONOLITHIC_HEADER", projectName));
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(String.Format("#define {0}_MONOLITHIC_HEADER", projectName));
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);

            mHeaderFileBuilder.Append("/* Begin enums */");
            mHeaderFileBuilder.Append(Environment.NewLine);

            foreach (CEnum e in enums)
            {
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append(e.GetHeaderCode());
            }

            mHeaderFileBuilder.Append("/* End enums */");
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);


            mHeaderFileBuilder.Append("/* Begin structs */");
            mHeaderFileBuilder.Append(Environment.NewLine);

            foreach (CStruct s in structs)
            {
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append(s.GetHeaderCode());
            }

            mHeaderFileBuilder.Append("/* End structs */");
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);

            mHeaderFileBuilder.Append("/* Begin unions */");
            mHeaderFileBuilder.Append(Environment.NewLine);

            foreach (CUnion u in unions)
            {
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append(u.GetHeaderCode());
            }

            mHeaderFileBuilder.Append("/* End unions */");
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);

            mHeaderFileBuilder.Append("/* Begin typedefs */");
            mHeaderFileBuilder.Append(Environment.NewLine);

            foreach (CTypeDef t in typedefs)
            {
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append(t.GetHeaderCode());
            }

            if (generated_typedefs.Count > 0)
            {
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append("/* Begin typedefs which were autogenerated by TDR */");
                mHeaderFileBuilder.Append(Environment.NewLine);

                generated_typedefs.Sort();

                foreach (CTypeDef t in generated_typedefs)
                {
                    mHeaderFileBuilder.Append(Environment.NewLine);
                    mHeaderFileBuilder.Append(t.GetHeaderCode());
                }
            }

            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append("/* End typedefs which were autogenerated by TDR */");
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);

            mHeaderFileBuilder.Append("/* End typedefs */");
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);


            mHeaderFileBuilder.Append("/* Begin external declarations */");
            mHeaderFileBuilder.Append(Environment.NewLine);

            foreach (CExternalDeclaration e in externs)
            {
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append(e.GetHeaderCode());
                mHeaderFileBuilder.Append(Environment.NewLine);
            }

            mHeaderFileBuilder.Append("/* End external declarations */");
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);

            mHeaderFileBuilder.Append("#endif");

            result.ManualChangesBuilder = manualChangesBuilder;
            result.SourceCode = mHeaderFileBuilder.ToString();

            return result;
        }

        public static StringBuilder CheckDuplicateNames(StringBuilder builder, List<string> dupeList, string objectType, string fileType, string filePath, string fileContent, string conflictType)
        {
            foreach (string dn in dupeList)
            {
                Regex rex = new Regex(string.Format("{0}\\s+[0-9A-Za-z_]", dn));
                if (rex.IsMatch(fileContent))
                {
                    builder.Append(string.Format("{0} file '{1}' appears to refer to the {2} '{3}', {4}. This may help you manually resolve the conflict.", fileType, filePath, objectType, dn, conflictType));
                    builder.Append(Environment.NewLine);
                }
            }
            return builder;
        }

        public static string GetRelativeSourceFilePath(string sourceFileName)
        {
            string result = sourceFileName;
            // remove any leading drive letter if the code was compiled on Windows
            Regex driveLetterRex = new Regex(@"^[A-Za-z]:");
            if (driveLetterRex.IsMatch(result))
            {
                //result = result.Substring(2);
                result = result.Substring(0, 1) + result.Substring(2); ;
            }
            // replace forward and backslashes with the path separator for the current platform
            result = result.Replace("/", Path.DirectorySeparatorChar.ToString());
            result = result.Replace(@"\", Path.DirectorySeparatorChar.ToString());
            // make path relative to output directory
            while (result.StartsWith(Path.DirectorySeparatorChar.ToString()))
            {
                result = result.Substring(1);
            }
            return result;
        }

        public static string GetUnmappedSourceCodeFileName(string subsetName)
        {
            //return String.Format("{0}.C", subsetName);
            //return "THISDUST.C";
            return String.Format("{0}.C", DefaultOrUnmappedFilename);
        }

        public static string GetUnmappedHeaderFileName(string subsetName)
        {
            return Path.ChangeExtension(GetUnmappedSourceCodeFileName(subsetName), "H");
        }

        public static IndividualSourceFileGenerationResult GenerateSourceCodeFiles(MonolithicHeaderGenerationResult mhgr, GenerateGhidraScriptsResult parentResult, string projectName, 
            string baseOutputDirectoryPath, string sourceCodeSubDirectoryName, string sourceFileName, string unmappedHeaderContentFileName, List<CFunction> functions,
            List<CElement> otherElements)
        {
            if (sourceFileName == "")
            {
                throw new Exception("Attempt to create a source file with no name.");
            }

            IndividualSourceFileGenerationResult result = new IndividualSourceFileGenerationResult();
            // create stubs for individual source files
            string stubSourceFilePath = GetRelativeSourceFilePath(sourceFileName);

            string stubTemp = stubSourceFilePath;
            stubSourceFilePath = Path.Combine(baseOutputDirectoryPath, sourceCodeSubDirectoryName, stubTemp);

            // get a list of unique line numbers for methods from the file, and sort by that
            List<long> lineNumbers = new List<long>();
            foreach (CFunction f in functions)
            {
                if (!lineNumbers.Contains(f.source_file_line_number))
                {
                    lineNumbers.Add(f.source_file_line_number);
                }
            }
            lineNumbers.Sort();

            StringBuilder headerFileHeaderBuilder = new StringBuilder();
            StringBuilder headerFileBodyBuilder = new StringBuilder();
            StringBuilder headerFileFooterBuilder = new StringBuilder();

            StringBuilder sourceFileHeaderBuilder = new StringBuilder();
            StringBuilder sourceFileBodyBuilder = new StringBuilder();
            StringBuilder sourceFileFooterBuilder = new StringBuilder();

            bool wasNotCompiledFromC = false;
            string fileExt = Path.GetExtension(stubSourceFilePath).ToUpper();
            if ((fileExt != ".C") && (fileExt != ".CPP") && (fileExt != ".H"))
            {
                wasNotCompiledFromC = true;
            }

            string headerFilePath = Path.ChangeExtension(stubSourceFilePath, "H");
            string headerFileName = Path.GetFileName(headerFilePath);
            string headerFileDefString = headerFileName.Replace(".", "_");

            bool headerAndSourceAreSameFile = false;
            if (Path.GetFileName(stubSourceFilePath) == headerFileName)
            {
                headerAndSourceAreSameFile = true;
            }

            headerFileHeaderBuilder.Append(String.Format("#ifndef {0}", headerFileDefString));
            headerFileHeaderBuilder.Append(Environment.NewLine);
            headerFileHeaderBuilder.Append(String.Format("#define {0}", headerFileDefString));
            headerFileHeaderBuilder.Append(Environment.NewLine);
            headerFileHeaderBuilder.Append(Environment.NewLine);

            if (wasNotCompiledFromC)
            {
                sourceFileHeaderBuilder.Append(String.Format("// WARNING: This file's extension ('{0}') indicates that it may not have been a C or C++ file.", fileExt));
                sourceFileHeaderBuilder.Append(Environment.NewLine);
                sourceFileHeaderBuilder.Append(String.Format("// It should contain Ghidra's best effort at decompiling the machine language to C, "));
                sourceFileHeaderBuilder.Append(Environment.NewLine);
                sourceFileHeaderBuilder.Append(String.Format("// but TDR does not currently attempt to handle assembly language or other non-C/C++ code differently, "));
                sourceFileHeaderBuilder.Append(Environment.NewLine);
                sourceFileHeaderBuilder.Append(String.Format("// so any decompiled code below may not reflect the original project source code for this file."));
                sourceFileHeaderBuilder.Append(Environment.NewLine);
                sourceFileHeaderBuilder.Append(String.Format("// If the source file is from a library, you should replace it with the original library version of the file."));
                sourceFileHeaderBuilder.Append(Environment.NewLine);
                sourceFileHeaderBuilder.Append(String.Format("// Otherwise, manual reconstruction of the code may be required, "));
                sourceFileHeaderBuilder.Append(Environment.NewLine);
                sourceFileHeaderBuilder.Append(String.Format("// Especially if the original source code was assembly language which used hardware features inaccessible to C."));
                sourceFileHeaderBuilder.Append(Environment.NewLine);
            }

            // handle edge case: source file IS header file
            // i.e. if functions are defined in a .H file instead of just declared in it.
            if (!headerAndSourceAreSameFile)
            {
                sourceFileHeaderBuilder.Append(String.Format("#include \"{0}\"", unmappedHeaderContentFileName));
                sourceFileHeaderBuilder.Append(Environment.NewLine);
                if (unmappedHeaderContentFileName != headerFileName)
                {
                    sourceFileHeaderBuilder.Append(String.Format("#include \"{0}\"", headerFileName));
                    sourceFileHeaderBuilder.Append(Environment.NewLine);
                }
            }
            sourceFileBodyBuilder.Append(Environment.NewLine);

            // sort based on SYM file sequence number, so that hopefully dependencies are written in the correct order
            otherElements.Sort();

            foreach (CElement e in otherElements)
            {
                string hc = e.GetHeaderCode();
                if (hc != "")
                {
                    headerFileBodyBuilder.Append(hc);
                    headerFileBodyBuilder.Append(Environment.NewLine);
                }

                string cc = e.GetSourceCode();
                if (cc != "")
                {
                    sourceFileBodyBuilder.Append(cc);
                    sourceFileBodyBuilder.Append(Environment.NewLine);
                }
            }

            foreach (int lineNumber in lineNumbers)
            {
                foreach (CFunction f in functions)
                {
                    if (f.source_file_line_number == lineNumber)
                    {
                        headerFileBodyBuilder.Append(f.GetHeaderCode());
                        headerFileBodyBuilder.Append(Environment.NewLine);
                        headerFileBodyBuilder.Append(Environment.NewLine);

                        sourceFileBodyBuilder.Append(f.GetSourceCode());
                        sourceFileBodyBuilder.Append(Environment.NewLine);
                        sourceFileBodyBuilder.Append(Environment.NewLine);
                    }
                }
            }

            if (headerAndSourceAreSameFile)
            {
                headerFileBodyBuilder.Append("%STUBCONTENT%");
                headerFileBodyBuilder.Append(Environment.NewLine);
            }
            headerFileFooterBuilder.Append(String.Format("#endif", headerFileDefString));

            string headerFileBodyContent = headerFileBodyBuilder.ToString();
            string sourceFileBodyContent = sourceFileBodyBuilder.ToString();
            if (headerAndSourceAreSameFile)
            {
                headerFileBodyContent = headerFileBodyContent.Replace("%STUBCONTENT%", sourceFileBodyContent);
                IndividualSourceFileContent headerFile = new IndividualSourceFileContent();
                headerFile.HeaderContent = headerFileHeaderBuilder.ToString();
                headerFile.BodyContent = headerFileBodyContent;
                headerFile.FooterContent = headerFileFooterBuilder.ToString();
                
                result.OutputFiles.Add(headerFilePath, headerFile);
                //WriteToOutput(headerFilePath, headerFileContent, FileMode.Create);
            }
            else
            {
                IndividualSourceFileContent headerFile = new IndividualSourceFileContent();
                headerFile.HeaderContent = headerFileHeaderBuilder.ToString();
                headerFile.BodyContent = headerFileBodyContent;
                headerFile.FooterContent = headerFileFooterBuilder.ToString();

                IndividualSourceFileContent sourceFile = new IndividualSourceFileContent();
                sourceFile.HeaderContent = sourceFileHeaderBuilder.ToString();
                sourceFile.BodyContent = sourceFileBodyContent;
                sourceFile.FooterContent = sourceFileFooterBuilder.ToString();

                result.OutputFiles.Add(headerFilePath, headerFile);
                result.OutputFiles.Add(stubSourceFilePath, sourceFile);
                //WriteToOutput(headerFilePath, headerFileContent, FileMode.Create);
                //WriteToOutput(stubSourceFilePath, sourceFileStubContent, FileMode.Create);
            }

            if (parentResult != null)
            {
                result.ManualChangesBuilder = CheckDuplicateNames(result.ManualChangesBuilder, parentResult.DuplicateEnumNames, "enum", "Header", headerFilePath, headerFileBodyContent, "which is defined multiple times");
                result.ManualChangesBuilder = CheckDuplicateNames(result.ManualChangesBuilder, parentResult.DuplicateStructNames, "struct", "Header", headerFilePath, headerFileBodyContent, "which is defined multiple times");
                result.ManualChangesBuilder = CheckDuplicateNames(result.ManualChangesBuilder, parentResult.DuplicateUnionNames, "union", "Header", headerFilePath, headerFileBodyContent, "which is defined multiple times");

                result.ManualChangesBuilder = CheckDuplicateNames(result.ManualChangesBuilder, parentResult.EnumStructDupeNames, "enum or struct", "Header", headerFilePath, headerFileBodyContent, "which is used as a name for both an enum and a struct");
                result.ManualChangesBuilder = CheckDuplicateNames(result.ManualChangesBuilder, parentResult.EnumUnionDupeNames, "enum or union", "Header", headerFilePath, headerFileBodyContent, "which is used as a name for both an enum and a union");
                result.ManualChangesBuilder = CheckDuplicateNames(result.ManualChangesBuilder, parentResult.StructUnionDupeNames, "struct or union", "Header", headerFilePath, headerFileBodyContent, "which is used as a name for both a struct and a union");

                result.ManualChangesBuilder = CheckDuplicateNames(result.ManualChangesBuilder, parentResult.DuplicateEnumNames, "enum", "Source", stubSourceFilePath, sourceFileBodyContent, "which is defined multiple times");
                result.ManualChangesBuilder = CheckDuplicateNames(result.ManualChangesBuilder, parentResult.DuplicateStructNames, "struct", "Source", stubSourceFilePath, sourceFileBodyContent, "which is defined multiple times");
                result.ManualChangesBuilder = CheckDuplicateNames(result.ManualChangesBuilder, parentResult.DuplicateUnionNames, "union", "Source", stubSourceFilePath, sourceFileBodyContent, "which is defined multiple times");

                result.ManualChangesBuilder = CheckDuplicateNames(result.ManualChangesBuilder, parentResult.EnumStructDupeNames, "enum or struct", "Header", stubSourceFilePath, sourceFileBodyContent, "which is used as a name for both an enum and a struct");
                result.ManualChangesBuilder = CheckDuplicateNames(result.ManualChangesBuilder, parentResult.EnumUnionDupeNames, "enum or union", "Header", stubSourceFilePath, sourceFileBodyContent, "which is used as a name for both an enum and a union");
                result.ManualChangesBuilder = CheckDuplicateNames(result.ManualChangesBuilder, parentResult.StructUnionDupeNames, "struct or union", "Header", stubSourceFilePath, sourceFileBodyContent, "which is used as a name for both a struct and a union");
            }

            return result;
        }

        public static IndividualSourceFileGenerationResult GetPlaceholderFile(string sourceFileName, string baseOutputDirectoryPath, string sourceCodeSubDirectoryName)
        {
            IndividualSourceFileGenerationResult result = new IndividualSourceFileGenerationResult();
            // create stubs for individual source files
            string stubSourceFilePath = GetRelativeSourceFilePath(sourceFileName);
            stubSourceFilePath = Path.Combine(baseOutputDirectoryPath, sourceCodeSubDirectoryName, stubSourceFilePath);

            StringBuilder builder = new StringBuilder();
            builder.Append("// This is a placeholder file");
            builder.Append(Environment.NewLine);
            builder.Append(string.Format("// The debug symbols for this project reference a file named '{0}',", sourceFileName));
            builder.Append(Environment.NewLine);
            builder.Append("// but the TDR toolchain does not currently support exporting any content into the file.");
            builder.Append(Environment.NewLine);
            IndividualSourceFileContent placeHolderFile = new IndividualSourceFileContent();
            placeHolderFile.HeaderContent = builder.ToString();
            result.OutputFiles.Add(stubSourceFilePath, placeHolderFile);

            return result;
        }

        public static List<PsyQOverlayDescription> GetOverlayList(SymDumpSymbolCollection symbolCollection, List<long> userSpecifiedOverlayIDs, Assembly asm, string programName, string copyright, string infoURL, string[] thanksList)
        {
            List<PsyQOverlayDescription> result = new List<PsyQOverlayDescription>();
            // if the user specified any overlays, make sure they're valid and don't overlap
            Dictionary<long, PsyQOverlayGroup> overlayGroups = new Dictionary<long, PsyQOverlayGroup>();
            List<long> allOverlayIDs = new List<long>();
            foreach (PsyQOverlayDescription overlay in symbolCollection.overlays)
            {
                PsyQOverlayGroup olGroup = new PsyQOverlayGroup();
                if (overlayGroups.Keys.Contains(overlay.offset))
                {
                    olGroup = overlayGroups[overlay.offset];
                }
                olGroup.overlays.Add(overlay);
                allOverlayIDs.Add(overlay.overlay_id);
                if (overlayGroups.Keys.Contains(overlay.offset))
                {
                    overlayGroups[overlay.offset] = olGroup;
                }
                else
                {
                    overlayGroups.Add(overlay.offset, olGroup);
                }
            }
            // first, make sure the user didn't request any invalid overlay IDs
            List<long> badOverlayIDs = new List<long>();
            foreach (long overlayID in userSpecifiedOverlayIDs)
            {
                if (!allOverlayIDs.Contains(overlayID))
                {
                    badOverlayIDs.Add(overlayID);
                }
            }
            if (badOverlayIDs.Count > 0)
            {
                StringBuilder builder = new StringBuilder();
                builder.Append(badOverlayIDs.Count.ToString());
                if (badOverlayIDs.Count == 1)
                {
                    builder.Append(" invalid overlay ID was specified: ");
                    builder.Append(badOverlayIDs[0].ToString());
                }
                else
                {
                    builder.Append(" invalid overlay IDs were specified: ");
                    for (int i = 0; i < badOverlayIDs.Count; i++)
                    {
                        builder.Append(badOverlayIDs[i].ToString());
                        if (i < (badOverlayIDs.Count - 1))
                        {
                            builder.Append(", ");
                        }
                    }
                }
                builder.Append(Environment.NewLine);
                builder.Append(BenLincoln.DebugSymbols.Utilities.GetOverlayList(symbolCollection));
                BenLincoln.DebugSymbols.Utilities.ExitWithError(builder.ToString(), 1, asm, programName, copyright, infoURL, thanksList);
            }
            // second, make sure none of the overlays overlap with each other
            Dictionary<long, List<long>> overlappingIDs = new Dictionary<long, List<long>>();
            foreach (long overlayID in userSpecifiedOverlayIDs)
            {
                PsyQOverlayDescription ol = null;
                foreach (PsyQOverlayDescription overlay in symbolCollection.overlays)
                {
                    if (overlay.overlay_id == overlayID)
                    {
                        ol = overlay;
                        break;
                    }
                }
                bool hasOverlaps = false;
                foreach (PsyQOverlayDescription overlay in symbolCollection.overlays)
                {
                    
                    if (overlay.overlay_id != overlayID)
                    {
                        if (userSpecifiedOverlayIDs.Contains(overlay.overlay_id))
                        {
                            if (ol.OverlapsWithOverlay(overlay))
                            {
                                hasOverlaps = true;
                                if (!overlappingIDs.Keys.Contains(ol.overlay_id))
                                {
                                    overlappingIDs.Add(ol.overlay_id, new List<long>());
                                }
                                overlappingIDs[ol.overlay_id].Add(overlay.overlay_id);
                            }
                        }
                    }
                }
                if (!hasOverlaps)
                {
                    result.Add(ol);
                }
            }
            if (overlappingIDs.Count > 0)
            {
                StringBuilder builder = new StringBuilder();
                builder.Append(overlappingIDs.Count.ToString());
                if (overlappingIDs.Count == 1)
                {
                    builder.Append(" overlapping overlay was specified: ");
                }
                else
                {
                    builder.Append(" overlapping overlays were specified: ");
                }
                builder.Append(Environment.NewLine);
                foreach (long overlappingBaseID in overlappingIDs.Keys)
                {
                    builder.Append("\tOverlay ");
                    builder.Append(overlappingBaseID);
                    builder.Append(" overlaps with overlay");
                    if (overlappingIDs[overlappingBaseID].Count == 1)
                    {
                        builder.Append(" ");
                        builder.Append(overlappingIDs[overlappingBaseID][0]);
                    }
                    else
                    {
                        builder.Append("s ");
                        for (int i = 0; i < overlappingIDs[overlappingBaseID].Count; i++)
                        {
                            builder.Append(overlappingIDs[overlappingBaseID][i]);
                            if (i < (overlappingIDs[overlappingBaseID].Count - 1))
                            {
                                builder.Append(", ");
                            }
                        }
                    }
                    builder.Append(Environment.NewLine);
                }
                builder.Append("Specifying overlapping overlays is not permitted, because they occupy the same memory space, and this will corrupt the results. Refer to the TDR documentation for examples of how to handle this type of project.");
                builder.Append(Environment.NewLine);
                builder.Append(BenLincoln.DebugSymbols.Utilities.GetOverlayList(symbolCollection));
                BenLincoln.DebugSymbols.Utilities.ExitWithError(builder.ToString(), 1, asm, programName, copyright, infoURL, thanksList);
            }

            return result;
        }

    }
}
