﻿// This file is part of BenLincoln.DebugSymbols.

// BenLincoln.DebugSymbols is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DebugSymbols is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DebugSymbols (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

using BenLincoln.DebugSymbols.SymDump;

namespace BenLincoln.DebugSymbols
{
    public class SourceCodeParser
    {
        public static string StripOutFunctionSignatureJunk(string inputString)
        {
            string result = inputString;
            result = result.Replace("\r", "").Replace("\n", " ").Trim();
            while (result.Contains("\t"))
            {
                result = result.Replace("\t", " ");
            }
            while (result.Contains("  "))
            {
                result = result.Replace("  ", " ");
            }
            return result;
        }

        public enum ParserState
        {
            LookingForStartOfNextFunctionSignature,
            LookingForEndOfFunctionSignature,
            IgnoringComments,
            ReadingSource
        }

        // this revised function is ONLY capable of parsing Ghidra output generated using the custom export script 
        // generated by CreateSkeleton. OTOH, it's *much* more reliable.
        public static List<CFunction> GetAllFunctionsFromSource(string sourceCode)
        {
            List<CFunction> result = new List<CFunction>();

            string functionStartTag = "// TDR: BEGIN FUNCTION";
            string functionEndTag = "// TDR: END FUNCTION";
            Regex methodAddressRex = new Regex("// TDR: Address: 0x(?<functionAddressHex>[0-9a-fA-F]+)($|[^0-9a-fA-F])", RegexOptions.Multiline);
            Regex methodNameRex = new Regex("// TDR: Name: (?<functionName>.+)($|[^0-9a-fA-F])", RegexOptions.Multiline);
            Regex methodSignatureRex = new Regex("// TDR: Signature: (?<functionSignature>.+)($|[^0-9a-fA-F])", RegexOptions.Multiline);

            string remainingSource = sourceCode;
            bool done = false;
            while (!done)
            {
                int nextFunctionBegin = remainingSource.IndexOf(functionStartTag);
                if (nextFunctionBegin < 0)
                {
                    done = true;
                }
                else
                {
                    nextFunctionBegin += functionStartTag.Length;
                    int nextFunctionEnd = remainingSource.IndexOf(functionEndTag, nextFunctionBegin);
                    string currentFunction = "";
                    if (nextFunctionEnd < 0)
                    {
                        currentFunction = remainingSource.Substring(nextFunctionBegin);
                        remainingSource = "";
                        done = true;
                    }
                    else
                    {
                        currentFunction = remainingSource.Substring(nextFunctionBegin, nextFunctionEnd - nextFunctionBegin);
                        remainingSource = remainingSource.Substring(nextFunctionEnd + functionEndTag.Length);
                    }
                    try
                    {
                        string methodAddressHex = "";
                        long methodAddress = 0;
                        string methodSignature = "";
                        string methodName = "";
                        MatchCollection methodAddressMatches = methodAddressRex.Matches(currentFunction);
                        if (methodAddressMatches.Count > 0)
                        {
                            if (methodAddressMatches[0].Groups["functionAddressHex"] != null)
                            {
                                methodAddressHex = methodAddressMatches[0].Groups["functionAddressHex"].Value.Trim();
                            }
                        }
                        MatchCollection methodNameMatches = methodNameRex.Matches(currentFunction);
                        if (methodNameMatches.Count > 0)
                        {
                            if (methodNameMatches[0].Groups["functionName"] != null)
                            {
                                methodName = methodNameMatches[0].Groups["functionName"].Value.Trim();
                            }
                        }
                        MatchCollection methodSignatureMatches = methodSignatureRex.Matches(currentFunction);
                        if (methodSignatureMatches.Count > 0)
                        {
                            if (methodSignatureMatches[0].Groups["functionSignature"] != null)
                            {
                                methodSignature = methodSignatureMatches[0].Groups["functionSignature"].Value.Trim();
                            }
                        }
                        bool processMethod = true;
                        if (methodAddressHex == "")
                        {
                            processMethod = false;
                            Console.WriteLine(string.Format("Error: could not find a function address in the following code: '{0}'", currentFunction));
                        }
                        else
                        {
                            methodAddress = long.Parse(methodAddressHex, System.Globalization.NumberStyles.HexNumber);
                        }
                        if (methodSignature == "")
                        {
                            processMethod = false;
                            Console.WriteLine(string.Format("Error: could not find a function signature in the following code: '{0}'", currentFunction));
                        }
                        if (processMethod)
                        {
                            CFunction newFunc = new CFunction();
                            newFunc.method_signature = methodSignature;
                            newFunc.offset_in_memory = methodAddress;
                            newFunc.name = methodName;
                            currentFunction = methodAddressRex.Replace(currentFunction, "");
                            currentFunction = methodSignatureRex.Replace(currentFunction, "");
                            currentFunction = methodNameRex.Replace(currentFunction, "");
                            currentFunction = currentFunction.Trim();
                            newFunc.source_code = currentFunction;
                            result.Add(newFunc);
                        }
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(string.Format("Error: could not parse the following code as a function: '{0}': {1}", currentFunction, ex.Message));
                    }
                }
            }

            return result;
        }

        public static List<CExternalDeclaration> GetExportedData(string exportedData)
        {
            List<CExternalDeclaration> result = new List<CExternalDeclaration>();

            string dataStartTag = "// TDR: BEGIN GLOBAL VARIABLE";
            string dataEndTag = "// TDR: END GLOBAL VARIABLE";
            string headerStartTag = "// TDR: BEGIN C HEADER CONTENT";
            string headerEndTag = "// TDR: END C HEADER CONTENT";
            string codeStartTag = "// TDR: BEGIN C CODE";
            string codeEndTag = "// TDR: END C CODE";
            Regex dataAddressRex = new Regex("// TDR: Address: 0x(?<dataAddressHex>[0-9a-fA-F]+)($|[^0-9a-fA-F])", RegexOptions.Multiline);
            Regex dataNameRex = new Regex("// TDR: Name in Code: (?<dataName>.+)($|[^0-9a-fA-F])", RegexOptions.Multiline);

            string remainingExportedData = exportedData;
            bool done = false;
            while (!done)
            {
                int nextCodeBegin = remainingExportedData.IndexOf(dataStartTag);
                if (nextCodeBegin < 0)
                {
                    done = true;
                }
                else
                {
                    nextCodeBegin += dataStartTag.Length;
                    int nextFunctionEnd = remainingExportedData.IndexOf(dataEndTag, nextCodeBegin);
                    string currentDataDefinition = "";
                    if (nextFunctionEnd < 0)
                    {
                        currentDataDefinition = remainingExportedData.Substring(nextCodeBegin);
                        remainingExportedData = "";
                        done = true;
                    }
                    else
                    {
                        currentDataDefinition = remainingExportedData.Substring(nextCodeBegin, nextFunctionEnd - nextCodeBegin);
                        remainingExportedData = remainingExportedData.Substring(nextFunctionEnd + dataEndTag.Length);
                    }
                    try
                    {
                        string dataAddressHex = "";
                        long dataAddress = 0;
                        string dataName = "";
                        string dataHeader = "";
                        string dataSourceCode = "";
                        MatchCollection methodAddressMatches = dataAddressRex.Matches(currentDataDefinition);
                        if (methodAddressMatches.Count > 0)
                        {
                            if (methodAddressMatches[0].Groups["dataAddressHex"] != null)
                            {
                                dataAddressHex = methodAddressMatches[0].Groups["dataAddressHex"].Value.Trim();
                            }
                        }
                        MatchCollection dataNameMatches = dataNameRex.Matches(currentDataDefinition);
                        if (dataNameMatches.Count > 0)
                        {
                            if (dataNameMatches[0].Groups["dataName"] != null)
                            {
                                dataName = dataNameMatches[0].Groups["dataName"].Value.Trim();
                            }
                        }
                        bool processData = true;
                        if (dataAddressHex == "")
                        {
                            processData = false;
                            Console.WriteLine(string.Format("Error: could not find an address in the following code: '{0}'", currentDataDefinition));
                        }
                        else
                        {
                            dataAddress = long.Parse(dataAddressHex, System.Globalization.NumberStyles.HexNumber);
                        }
                        if (dataName == "")
                        {
                            processData = false;
                            Console.WriteLine(string.Format("Warning: could not find an extern name in the following data: '{0}'", currentDataDefinition));
                        }
                        if ((!currentDataDefinition.Contains(headerStartTag)) || (!currentDataDefinition.Contains(headerEndTag)) 
                            || (!currentDataDefinition.Contains(codeStartTag)) || (!currentDataDefinition.Contains(codeEndTag)))
                        {
                            processData = false;
                            Console.WriteLine(string.Format("Warning: could not find one or more required delimiters in the following data: '{0}'", currentDataDefinition));
                        }
                        if (processData)
                        {
                            dataHeader = currentDataDefinition.Substring(currentDataDefinition.IndexOf(headerStartTag) + headerStartTag.Length + Environment.NewLine.Length);
                            dataHeader = dataHeader.Substring(0, dataHeader.IndexOf(headerEndTag)).Trim();

                            dataSourceCode = currentDataDefinition.Substring(currentDataDefinition.IndexOf(codeStartTag) + codeStartTag.Length + Environment.NewLine.Length);
                            dataSourceCode = dataSourceCode.Substring(0, dataSourceCode.IndexOf(codeEndTag)).Trim();

                            CExternalDeclaration newExtern = new CExternalDeclaration();
                            newExtern.name = dataName;
                            newExtern.offset_in_memory = dataAddress;
                            newExtern.declaration = dataHeader;
                            newExtern.source_code = dataSourceCode;
                            //Console.WriteLine(string.Format("Debug: added exported extern: '{0}' @0x{1:X8}, '{2}'", newExtern.name, newExtern.offset_in_memory, newExtern.declaration));
                            //Console.WriteLine(newExtern.source_code);
                            result.Add(newExtern);
                        }
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(string.Format("Error: could not parse the following code as exported data: '{0}': {1}", currentDataDefinition, ex.Message));
                    }
                }
            }

            return result;
        }

        //public static List<CFunction> GetAllFunctionsFromSource(string sourceCode)
        //{
        //    List<CFunction> result = new List<CFunction>();

        //    // regex string for source code identifiers
        //    // "void", "int", "PLAYER_SetLookAround", etc.
        //    string identifier_regex_fragment = "[^ #();\\`'\"]+";

        //    // not using string.Format in most of this function to avoid side-effects due to regex characters
        //    //Regex methodStartRex = new Regex("^(?!//)(" + identifier_regex_fragment + " )+" + identifier_regex_fragment + "([^)]*)", RegexOptions.Multiline | RegexOptions.IgnoreCase);
        //    //Regex methodStartRex = new Regex("(^|\\n)\\s*(" + identifier_regex_fragment + "\\s+)+" + identifier_regex_fragment + "\\s*\\(", RegexOptions.Singleline | RegexOptions.IgnoreCase);
        //    Regex methodStartRex = new Regex("(^|\\n)(" + identifier_regex_fragment + "\\s+)+" + identifier_regex_fragment + "\\s*\\(", RegexOptions.Singleline | RegexOptions.IgnoreCase);
        //    //Regex methodStartRex = new Regex("^\\s*(" + identifier_regex_fragment + "\\s*)+" + identifier_regex_fragment + "\\s*\\(", RegexOptions.Singleline | RegexOptions.IgnoreCase);
        //    //Regex methodStartPartialRex = new Regex("^\\s*(" + identifier_regex_fragment + " )+" + identifier_regex_fragment, RegexOptions.Multiline | RegexOptions.IgnoreCase);
        //    Regex fixMethodStartGapWhitespace = new Regex("\\s+\\(");

        //    // strip out block comments
        //    string source = sourceCode;
        //    Regex stripBlockCommentRex = new Regex("/\\*(?!\\*/).*\\*/", RegexOptions.Singleline | RegexOptions.IgnoreCase);
        //    source = stripBlockCommentRex.Replace(source, "");
        //    Regex stripInlineCommentsRex = new Regex("//.*$", RegexOptions.Multiline | RegexOptions.IgnoreCase);
        //    source = stripInlineCommentsRex.Replace(source, "");

        //    string sourceCodeUnixLB = source.Replace("\r", "");

        //    ParserState ps = ParserState.LookingForStartOfNextFunctionSignature;
        //    bool doneParsing = false;

        //    string[] sourceLines = sourceCodeUnixLB.Split(new string[] { "\n" }, StringSplitOptions.None);
        //    StringBuilder currentFunctionSourceBuilder = new StringBuilder();
        //    string currentFunctionSignature = "";
        //    string currentFunctionPreParameterSection = "";
        //    string currentFunctionParameterSection = "";
        //    int lineNumber = 0;
        //    int currentBracketLevel = 0;
        //    string currentLineBuffer = sourceLines[0];
        //    int noMatchFoundLoopCount = 0;

        //    while (!doneParsing)
        //    {
        //        switch(ps)
        //        {
        //            case ParserState.IgnoringComments:
        //                // not used in the current model
        //                break;
        //            case ParserState.ReadingSource:
        //                int bufferLength = currentLineBuffer.Length;
        //                for (int charNum = 0; charNum < bufferLength; charNum++)
        //                {
        //                    string currentChar = currentLineBuffer.Substring(0, 1);
        //                    currentLineBuffer = currentLineBuffer.Substring(1);
        //                    if (currentChar == "{")
        //                    {
        //                        currentBracketLevel++;
        //                    }
        //                    if (currentChar == "}")
        //                    {
        //                        currentBracketLevel--;
        //                    }
        //                    if (currentBracketLevel == 0)
        //                    {
        //                        break;
        //                    }
        //                    // this is at the very end so that the trailing } isn't included in the source
        //                    currentFunctionSourceBuilder.Append(currentChar);
        //                }
        //                currentFunctionSourceBuilder.Append(Environment.NewLine);
        //                if (currentBracketLevel == 0)
        //                {
        //                    CFunction newFunc = new CFunction();
        //                    newFunc.guid = Guid.NewGuid();
        //                    newFunc.source_code = currentFunctionSourceBuilder.ToString();
        //                    newFunc.method_signature = StripOutFunctionSignatureJunk(currentFunctionSignature);
        //                    currentFunctionParameterSection = StripOutFunctionSignatureJunk(currentFunctionParameterSection);
        //                    currentFunctionPreParameterSection = StripOutFunctionSignatureJunk(currentFunctionPreParameterSection);
        //                    CParameter[] returnType = CElement.GetCParametersFromString(currentFunctionPreParameterSection, false);
        //                    if (returnType.Length == 0)
        //                    {
        //                        Console.WriteLine(string.Format("Error: could not get a return type from the string '{0}'", currentFunctionPreParameterSection));
        //                    }
        //                    if (returnType.Length > 1)
        //                    {
        //                        Console.WriteLine(string.Format("Warning: unexpected result - got multiple return types from the string '{0}'", currentFunctionPreParameterSection));
        //                    }
        //                    newFunc.return_value = returnType[0];
        //                    newFunc.name = newFunc.return_value.name;
        //                    CParameter[] paramArray = CElement.GetCParametersFromString(currentFunctionParameterSection, false);
        //                    newFunc.parameters = new List<CParameter>();
        //                    for (int pNum = 0; pNum < paramArray.Length; pNum++)
        //                    {
        //                        newFunc.parameters.Add(paramArray[pNum]);
        //                    }

        //                    result.Add(newFunc);

        //                    // reset everything for next function
        //                    currentFunctionSourceBuilder = new StringBuilder();
        //                    currentFunctionSignature = "";
        //                    currentFunctionPreParameterSection = "";
        //                    currentFunctionParameterSection = "";
        //                    ps = ParserState.LookingForStartOfNextFunctionSignature;
        //                }
        //                break;
        //            case ParserState.LookingForStartOfNextFunctionSignature:
        //                if (currentLineBuffer.Length > 1)
        //                {
        //                    while (currentLineBuffer.StartsWith("\n"))
        //                    {
        //                        currentLineBuffer = currentLineBuffer.Substring(1);
        //                    }
        //                }
        //                MatchCollection matches = methodStartRex.Matches(currentLineBuffer);
        //                if (matches.Count == 0)
        //                {
        //                    noMatchFoundLoopCount++;
        //                    // try to handle the annoying case where Ghidra hard-wraps lines like this:
        //                    // void CriticalDampPosition(long dampMode, _Position* position, _Position* targetPos, _SVector* vel, _SVector* accl, int smooth)
        //                    // ...into this:
        //                    // void CriticalDampPosition
        //                    //                    (long dampMode, _Position* position, _Position* targetPos, _SVector* vel, _SVector* accl,
        //                    //                    int smooth)
        //                    //
        //                    // ... or this:
        //                    //_Instance * INSTANCE_IntroduceSavedInstanceWithIntro(_SavedIntroWithIntro * savedIntro, _StreamUnit * streamUnit)
        //                    // ...into this:
        //                    // _Instance *
        //                    // INSTANCE_IntroduceSavedInstanceWithIntro(_SavedIntroWithIntro * savedIntro, _StreamUnit * streamUnit)
        //                    //
        //                    // ...or this:
        //                    // _TFace * CAMERA_SphereToSphereWithLines(Camera * camera, CameraCollisionInfo * colInfo, int secondcheck_flag)
        //                    // ...into this:
        //                    // _TFace * CAMERA_SphereToSphereWithLines
        //                    //                    (Camera * camera, CameraCollisionInfo * colInfo, int secondcheck_flag)
        //                    //if (currentLineBuffer.Trim() != "")
        //                    //{
        //                    //    Console.WriteLine("Debug: something interesting to look at");
        //                    //}
        //                    if (lineNumber < (sourceLines.Length - 2))
        //                    {
        //                        string nextLine = sourceLines[lineNumber + 1];
        //                        currentLineBuffer += "\n";
        //                        currentLineBuffer += nextLine;
        //                        lineNumber++;
        //                    }
        //                    else
        //                    {
        //                        currentLineBuffer = "";
        //                    }

        //                    //if (lineNumber < (sourceLines.Length - 2))
        //                    //{
        //                    //    matches = methodStartPartialRex.Matches(currentLineBuffer);
        //                    //    if (matches.Count == 0)
        //                    //    {
        //                    //        currentLineBuffer = "";
        //                    //    }
        //                    //    else
        //                    //    {
        //                    //        string nextLineTrim = sourceLines[lineNumber + 1].Trim();
        //                    //        if (nextLineTrim.StartsWith("("))
        //                    //        {
        //                    //            currentLineBuffer += nextLineTrim;
        //                    //            lineNumber++;
        //                    //        }
        //                    //    }
        //                    //}
        //                    //else
        //                    //{
        //                    //    currentLineBuffer = "";
        //                    //}
        //                }
        //                if (matches.Count > 0)
        //                {
        //                    noMatchFoundLoopCount = 0;
        //                    currentFunctionSignature = matches[0].Value;
        //                    while (fixMethodStartGapWhitespace.IsMatch(currentFunctionSignature))
        //                    {
        //                        currentFunctionSignature = fixMethodStartGapWhitespace.Replace(currentFunctionSignature, "(");
        //                    }
        //                    currentFunctionPreParameterSection = currentFunctionSignature;
        //                    int startPosition = currentLineBuffer.IndexOf(currentFunctionSignature);
        //                    int remainderPosition = startPosition + currentFunctionSignature.Length;
        //                    currentLineBuffer = currentLineBuffer.Substring(remainderPosition);
        //                    ps = ParserState.LookingForEndOfFunctionSignature;
        //                }
        //                if (matches.Count > 1)
        //                {
        //                    Console.WriteLine(string.Format("Warning: unexpected condition - multiple apparent function beginnings on a single line at line {0} in source code: '{1}'", lineNumber + 1, currentLineBuffer));
        //                }
        //                break;
        //            case ParserState.LookingForEndOfFunctionSignature:
        //                int obPos = currentLineBuffer.IndexOf("{");
        //                if (obPos > -1)
        //                {
        //                    if (obPos > 0)
        //                    {
        //                        string newText = currentLineBuffer.Substring(0, obPos - 1);
        //                        currentFunctionSignature += newText;
        //                        currentFunctionParameterSection += newText;
        //                    }
        //                    if (currentLineBuffer.Length > 1)
        //                    {
        //                        currentLineBuffer = currentLineBuffer.Substring(obPos + 1);
        //                    }
        //                    else
        //                    {
        //                        currentLineBuffer = "";
        //                    }
        //                    currentBracketLevel++;
        //                    ps = ParserState.ReadingSource;
        //                }
        //                else
        //                {
        //                    currentFunctionSignature += currentLineBuffer;
        //                    currentFunctionParameterSection += currentLineBuffer;
        //                    currentLineBuffer = "";
        //                }
        //                break;
        //        }
        //        if (currentLineBuffer == "")
        //        {
        //            lineNumber++;
        //            if (lineNumber >= sourceLines.Length)
        //            {
        //                doneParsing = true;
        //            }
        //            else
        //            {
        //                currentLineBuffer = sourceLines[lineNumber];
        //            }
        //        }
        //    }

        //    return result;
        //}
    }
}
