﻿// This file is part of BenLincoln.DebugSymbols.

// BenLincoln.DebugSymbols is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DebugSymbols is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DebugSymbols (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using BenLincoln.DebugSymbols.SymDump;

namespace BenLincoln.DebugSymbols
{
    public class SourceCodeReference
    {
        public long Offset { get; set; }
        public long LineNumber { get; set; }

        public SourceCodeReference()
        {
            Offset = -1;
            LineNumber = -1;
        }

        public virtual string GetCode()
        {
            string result = "";
            return result;
        }
    }

    public class FunctionSourceCodeReference : SourceCodeReference
    {
        public CFunction Function;

        public FunctionSourceCodeReference()
            : base()
        {

        }

        public override string GetCode()
        {
            return base.GetCode();
        }
    }

    public class SourceCodeSourceFileTracker
    {
        public Dictionary<string, List<CFunction>> FunctionSourceFiles;
        public Dictionary<string, List<CElement>> OtherElementSourceFiles;
        public List<string> AllSourceFiles;

        public SourceCodeSourceFileTracker()
        {
            FunctionSourceFiles = new Dictionary<string, List<CFunction>>();
            OtherElementSourceFiles = new Dictionary<string, List<CElement>>();
            AllSourceFiles = new List<string>();
        }

        public void AddOrUpdateEntry(CFunction func)
        {
            if (FunctionSourceFiles.ContainsKey(func.source_file))
            {
                FunctionSourceFiles[func.source_file].Add(func);
            }
            else
            {
                List<CFunction> newList = new List<CFunction>();
                newList.Add(func);
                FunctionSourceFiles.Add(func.source_file, newList);
            }
            if (!AllSourceFiles.Contains(func.source_file))
            {
                AllSourceFiles.Add(func.source_file);
            }
        }

        public void AddOrUpdateEntry(CElement e)
        {
            if (OtherElementSourceFiles.ContainsKey(e.source_file))
            {
                OtherElementSourceFiles[e.source_file].Add(e);
            }
            else
            {
                List<CElement> newList = new List<CElement>();
                newList.Add(e);
                OtherElementSourceFiles.Add(e.source_file, newList);
            }
            if (!AllSourceFiles.Contains(e.source_file))
            {
                AllSourceFiles.Add(e.source_file);
            }
        }

        public void AddOrUpdateEntryIfNotAlreadyDefined(CFunction func)
        {
            if (FunctionSourceFiles.ContainsKey(func.source_file))
            {
                bool addEntry = true;
                foreach (CFunction existingFunc in FunctionSourceFiles[func.source_file])
                {
                    if (existingFunc.offset_in_memory == func.offset_in_memory)
                    {
                        addEntry = false;
                        if (existingFunc.name == func.name)
                        {
                            Console.WriteLine(string.Format("Info: a function named '{0}' with the address 0x{1:X8} already exists in the list of debug symbols. The additional definition will be ignored.", func.name, func.offset_in_memory));
                        }
                        else
                        {
                            Console.WriteLine(string.Format("Warning: a function definition with the address 0x{0:X8} already exists in the list of debug symbols. The additional definition will be ignored.", func.offset_in_memory));
                            Console.WriteLine(string.Format("\tWarning: existing function is named '{0}', and additional definition is named '{1}'", existingFunc.name, func.name));
                        }
                    }
                }
                if (addEntry)
                {
                    Console.WriteLine(string.Format("Info: found a function named '{0}' in the SLD data for this project at address 0x{1:X8} with source file name '{2}' and line number {3} which would not otherwise be recognized.", func.name, func.offset_in_memory, func.source_file, func.source_file_line_number));
                    FunctionSourceFiles[func.source_file].Add(func);
                }
            }
            else
            {
                List<CFunction> newList = new List<CFunction>();
                newList.Add(func);
                FunctionSourceFiles.Add(func.source_file, newList);
            }
            if (!AllSourceFiles.Contains(func.source_file))
            {
                AllSourceFiles.Add(func.source_file);
            }
        }

        public CFunction GetFunctionAtAddress(string sourceFile, long address)
        {
            if (FunctionSourceFiles.Keys.Contains(sourceFile))
            {
                foreach (CFunction existingFunc in FunctionSourceFiles[sourceFile])
                {
                    if (existingFunc.offset_in_memory == address)
                    {
                        return existingFunc;
                    }
                }
            }
            return null;
        }

        public CElement GetElementAtAddress(long address)
        {
            foreach (string sourceFile in FunctionSourceFiles.Keys)
            {
                foreach (CFunction existingFunc in FunctionSourceFiles[sourceFile])
                {
                    if (existingFunc.offset_in_memory == address)
                    {
                        return existingFunc;
                    }
                }
            }
            foreach (string sourceFile in OtherElementSourceFiles.Keys)
            {
                foreach (CElement existingElem in OtherElementSourceFiles[sourceFile])
                {
                    if (existingElem.offset_in_memory == address)
                    {
                        return existingElem;
                    }
                }
            }

            return null;
        }
    }
}
