﻿// This file is part of BenLincoln.DebugSymbols.

// BenLincoln.DebugSymbols is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DebugSymbols is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DebugSymbols (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Threading;

namespace BenLincoln.DebugSymbols
{
    [StructLayout(LayoutKind.Explicit)]
    public struct UIntIntOverlay
    {
        [FieldOffset(0)] public uint UInt;
        [FieldOffset(0)] public int Int;
    }

    public class Utilities
    {
        public static string IndentString(string inputString, string indent)
        {
            StringBuilder resultBuilder = new StringBuilder();

            // handle cross-platform data correctly
            string lineBreak = "\r\n";

            if (!inputString.Contains("\r"))
            {
                lineBreak = "\n";
            }

            string input = inputString.Replace("\r", "");

            foreach (string line in inputString.Split(new char[] { '\n' }))
            {
                resultBuilder.Append(indent);
                resultBuilder.Append(line);
                resultBuilder.Append(lineBreak);
            }

            return resultBuilder.ToString();
        }

        public static string GetOverlayList(SymDump.SymDumpSymbolCollection symbolCollection)
        {
            StringBuilder builder = new StringBuilder();
            builder.Append("The following overlays were found in this set of debug symbols:");
            builder.Append(Environment.NewLine);
            foreach (BenLincoln.DebugSymbols.SymDump.PsyQOverlayDescription ol in symbolCollection.overlays)
            {
                builder.Append(string.Format("\tOverlay ID: 0x{0:X2} (decimal: {0}), address 0x{1:X8}, length 0x{2:X8} (decimal: {2})", ol.overlay_id, ol.offset, ol.length));
                builder.Append(Environment.NewLine);
            }
            return builder.ToString();
        }

        public static void HandleWarningForOverlays(SymDump.SymDumpSymbolCollection symbolCollection)
        {
            if (symbolCollection.overlays.Count > 0)
            {
                Console.WriteLine(string.Format("Warning: the debug symbols for this project reference {0} PsyQ overlays. Reverse-engineering this type of project requires additional manual effort. Please consult the TDR documentation - for example, the walkthrough of decompiling the OVERLAYS example binary.", symbolCollection.overlays.Count));
                Console.WriteLine(GetOverlayList(symbolCollection));
            }
        }

        public static uint CastIntToUInt(int i)
        {
            UIntIntOverlay overlay = new UIntIntOverlay();
            overlay.Int = i;
            return overlay.UInt;
        }

        public static void DisplayBanner(Assembly asm, string progName, string copyright, string infoURL, string[] thanksList)
        {
            AssemblyName assemblyName = asm.GetName();

            int versionMajor = assemblyName.Version.Major;
            int versionMinor = assemblyName.Version.Minor;

            Console.WriteLine(string.Format("{0} {1}.{2}", progName, versionMajor, versionMinor));
            Console.WriteLine(copyright);
            Console.WriteLine(infoURL);
            if (thanksList.Length > 0)
            {
                Console.WriteLine("Thank you:");
                foreach (string ty in thanksList)
                {
                    Console.WriteLine(ty);
                }
            }
            Console.WriteLine(Environment.NewLine);
        }

        public static void WriteToOutput(string outputPath, string text, FileMode mode)
        {
            FileStream oStream;
            StreamWriter oWriter;
            int maxTries = 3;
            int retryDelay = 1000;

            CreateParentPath(outputPath);

            bool done = false;
            int currentAttempt = 0;

            while (!done)
            {
                try
                {
                    oStream = new FileStream(outputPath, mode, FileAccess.Write);
                    oWriter = new StreamWriter(oStream);

                    oWriter.WriteLine(text);

                    oWriter.Close();
                    oStream.Close();
                    done = true;
                }
                catch (Exception ex)
                {
                    currentAttempt++;
                    if (currentAttempt > maxTries)
                    {
                        done = true;
                        Console.WriteLine(string.Format("Error writing to '{0}': {1}", outputPath, ex.Message));
                        Environment.Exit(1);
                    }
                    else
                    {
                        Thread.Sleep(retryDelay);
                    }
                }
            }
        }

        public static string GetFileAsString(string inputPath)
        {
            string result = "";
            FileStream inStream;
            StreamReader inReader;

            inStream = new FileStream(inputPath, FileMode.Open, FileAccess.Read);
            inReader = new StreamReader(inStream);

            result = inReader.ReadToEnd();

            inReader.Close();
            inStream.Close();
            return result;
        }

        public static string[] GetLinesAsStringArray(string inputString)
        {
            // for easier cross-platform use, this handles Unix and DOS-style newlines the same way
            string inString = inputString.Replace("\r", "");

            string[] result = inString.Split(new string[] { "\n" }, StringSplitOptions.RemoveEmptyEntries);

            return result;
        }

        public static string[] GetFileLinesAsStringArray(string inputPath)
        {
            return GetLinesAsStringArray(GetFileAsString(inputPath));
        }

        public static void ExitWithError(string message, int exitCode, Assembly asm, string programName, string copyright, string infoURL, string[] thanksList)
        {
            BenLincoln.DebugSymbols.Utilities.DisplayBanner(asm, programName, copyright, infoURL, thanksList);
            Console.WriteLine(message);
            System.Environment.ExitCode = exitCode;
            Environment.Exit(exitCode);
        }

        public static string GetEmbeddedTextFile(Assembly asm, string fileName)
        {
            Stream inStream = asm.GetManifestResourceStream(fileName);
            StreamReader inReader = new StreamReader(inStream);
            return inReader.ReadToEnd();
        }

        public static void DisplayHelpMessage(Assembly asm, string progName, string copyright, string infoURL, string[] thanksList, string helpMessageFile)
        {
            BenLincoln.DebugSymbols.Utilities.DisplayBanner(asm, progName, copyright, infoURL, thanksList);

            Console.Write(BenLincoln.DebugSymbols.Utilities.GetEmbeddedTextFile(asm, helpMessageFile));
            Console.WriteLine();
        }

        public static void CreateOutputDirectories(string baseOutputDirectoryPath, string[] outputDirs, Assembly asm, string programName, string copyright, string infoURL, string[] thanksList)
        {
            foreach (string od in outputDirs)
            {
                string fullPath = Path.Combine(baseOutputDirectoryPath, od);
                try
                {
                    Directory.CreateDirectory(fullPath);
                }
                catch (Exception ex)
                {
                    BenLincoln.DebugSymbols.Utilities.ExitWithError(String.Format("Could not create the output directory '{0}': {1}", fullPath, ex.Message), 2, asm, programName, copyright, infoURL, thanksList);
                }
            }
        }

        public static void CreateParentPath(string filePath)
        {
            string parentPath = Path.GetDirectoryName(Path.GetFullPath(filePath));
            if (!Directory.Exists(parentPath))
            {
                try
                {
                    Directory.CreateDirectory(parentPath);
                }
                catch (Exception ex)
                {
                    Console.WriteLine(String.Format("Error: Could not create a required output path ('{0}'): {1}", parentPath, ex.Message));
                    Environment.Exit(5);
                }
            }
        }

        public static long GetHexOrDecimalLong(string text)
        {
            long result = -1;
            if (text.ToLower().StartsWith("0x"))
            {
                try
                {
                    result = long.Parse(text.Substring(2), System.Globalization.NumberStyles.HexNumber);
                    return result;
                }
                catch (Exception ex)
                {
                    // proceed to the next approach, or fall back to throwing an exception if this is the last approach
                }
            }
            try
            {
                result = long.Parse(text);
                return result;
            }
            catch (Exception ex)
            {
                // proceed to the next approach, or fall back to throwing an exception if this is the last approach
            }
            throw new FormatException(string.Format("Could not parse '{0}' as a hexadecimal or decimal integer. For hex, use a '0x' prefix.", text));
        }

        public static string GetStringAsSafeDirectoryName(string text)
        {
            string result = text;
            Regex rex = new Regex("[^A-Za-z0-9\\-_\\.]");
            result = rex.Replace(result, "");
            return result;
        }

        // https://stackoverflow.com/questions/8613187/an-elegant-way-to-consume-all-bytes-of-a-binaryreader/8613300
        public static byte[] ReadAllBytes(BinaryReader reader)
        {
            const int bufferSize = 4096;
            using (var ms = new MemoryStream())
            {
                byte[] buffer = new byte[bufferSize];
                int count;
                while ((count = reader.Read(buffer, 0, buffer.Length)) != 0)
                {
                    ms.Write(buffer, 0, count);
                }
                return ms.ToArray();
            }
        }

        public static string GetObjectFileNameForRenamingFakes(string objectFileName)
        {
            string result = objectFileName.ToUpper();
            if (result.Contains("\\"))
            {
                int lastSlash = result.LastIndexOf("\\");
                result = result.Substring(lastSlash + 1);
            }
            int firstDot = result.IndexOf('.');
            if (firstDot > 0)
            {
                result = result.Substring(0, firstDot);
            }

            // C names can't begin with numbers
            Regex leadingDigitRex = new Regex("^[0-9]");
            if (leadingDigitRex.IsMatch(result))
            {
                result = string.Format("_{0}", result);
            }

            // C names can only consist of alphanumeric characters and underscores
            Regex invalidCharRex = new Regex("[^0-9A-Za-z_]");
            result = invalidCharRex.Replace(result, "_");

            return result;
        }

        public static string GetNewFakeName(string callerName, string fakeName, List<string> objectFileNames, string currentSLDSourceFile, string lastObjectFileName, string lastSLDSourceFile)
        {
            bool ignoreEOS = true;
            if (ignoreEOS)
            {
                if (fakeName == ".eos")
                {
                    return fakeName;
                }
            }
            string result = fakeName.Replace(".", "_");
            if (!result.StartsWith("_"))
            {
                return result;
            }
            string objectFileName = "";
            if (objectFileNames.Count > 0)
            {
                objectFileName = objectFileNames[0];
                if (objectFileNames.Count > 1)
                {
                    Console.WriteLine(string.Format("Warning: multiple object file names found. Using '{0}' from {1}", objectFileName, GetListAsString(objectFileNames)));
                }
            }
            string currentFileName = objectFileName;
            //string currentFileName = "";
            if (currentFileName == "")
            {
                currentFileName = currentSLDSourceFile;
            }
            if (currentFileName == "")
            {
                currentFileName = lastSLDSourceFile;
            }
            //if (currentFileName == "")
            //{
            //    currentFileName = objectFileName;
            //}
            if (currentFileName == "")
            {
                currentFileName = lastObjectFileName;
            }
            

            result = string.Format("{0}{1}", GetObjectFileNameForRenamingFakes(currentFileName), result);
            if (result == fakeName)
            {
                Console.WriteLine(string.Format("Debug: did not rename '{0}', based on object file name '{1}', SLD source file name '{2}', last object file name '{3}', last SLD source file name '{4}' (in '{5}')",
                    result, objectFileName, currentSLDSourceFile, lastObjectFileName, lastSLDSourceFile, callerName));
            }
            else
            {
                Console.WriteLine(string.Format("Debug: renamed '{0}' to '{1}', based on object file name '{2}', SLD source file name '{3}', last object file name '{4}', last SLD source file name '{5}' (in '{6}')", 
                    fakeName, result, objectFileName, currentSLDSourceFile, lastObjectFileName, lastSLDSourceFile, callerName));
            }
            return result;
        }

        public static string GetNewFakeName(string callerName, string fakeName, string objectFileName, string currentSLDSourceFile, string lastObjectFileName, string lastSLDSourceFile)
        {
            List<string> listOfOne = new List<string>();
            listOfOne.Add(objectFileName);
            return GetNewFakeName(callerName, fakeName, listOfOne, currentSLDSourceFile, lastObjectFileName, lastSLDSourceFile);
        }

        public static string GetListAsString(List<string> list)
        {
            StringBuilder builder = new StringBuilder();

            for (int i = 0; i < list.Count; i++)
            {
                builder.Append(list[i]);
                if (i < (list.Count - 1))
                {
                    builder.Append(", ");
                }
            }

            return builder.ToString();
        }
    }
}
