﻿// CreateSkeleton
// Creates a skeleton PlayStation dev project based on the output of SymDumpTE
// or another source of the appropriate JSON data
// Copyright 2019 Ben Lincoln
// https://www.beneaththewaves.net/

// This file is part of CreateSkeleton.

// CreateSkeleton is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// CreateSkeleton is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with CreateSkeleton (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using Newtonsoft.Json;
using BenLincoln.DebugSymbols;
using BenLincoln.DebugSymbols.SymDump;

namespace CreateSkeleton
{

    class Program
    {
        protected const string ProgramName = "CreateSkeleton";
        protected static string Copyright = "Copyright 2019 Ben Lincoln";
        protected const string InfoURL = "https://www.beneaththewaves.net/Software/This_Dust_Remembers_What_It_Once_Was.html";
        protected static string[] ThanksList = new string[] {
            "The United States National Security Agency for Ghidra (https://www.ghidra-sre.org/)",
            "Vladimir Kononovich (https://github.com/lab313ru) for https://github.com/lab313ru/ghidra_psx_ldr, which was the source for the segment-defining code in one of the generated Ghidra Java scripts",
            "Gh0stBlade (https://github.com/Gh0stBlade)",
            "Andrew Fradley (https://github.com/afradley)"
            };
        protected const string HelpMessageFile = "CreateSkeleton.HelpMessage.txt";

        protected const string ProjectNameValidationRegexString = @"^[A-Za-z0-9_]{1,16}$";
        protected const string ProjectNameExplanation = "alphanumeric characters and underscore (_)";
        //protected const uint MinAddressForLabels = 0x80000000;

        static void Main(string[] args)
        {
            if ((args.Length < 5))
            {
                BenLincoln.DebugSymbols.Utilities.DisplayHelpMessage(Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList, HelpMessageFile);
                Environment.Exit(1);
            }

            string projectName = "";
            string baseOutputDirectoryPath = "";
            string inputFilePath = "";
            bool forceOverwrite = false;
            bool ignoreLabels = false;
            bool ignoreNonASCIILabels = false;
            bool replaceNonASCIILabels = false;
            bool convertExternalDefinitionsToLabels = false;
            bool addStackPointerOffset = false;
            bool assignReturnAddressStorage = false;
            bool assignParameterStorage = false;
            bool createPlayStationMemory = false;
            bool assumeSNGPBase = false;
            bool mapSLDFunctions = false;
            bool explicitGPBase = false;
            long GPBase = 0;
            List<long> overlayIDs = new List<long>();
            string updatedJSONSymbolFilePath = "";

            ArrayList arguments = new ArrayList();
            for (int i = 0; i < args.Length; i++)
            {
                // if the argument is one with a second component, catch that now
                string nextArg = "";
                switch (args[i])
                {
                    case "--name":
                        nextArg = args[i + 1];
                        projectName = nextArg;
                        Regex pnRex = new Regex(ProjectNameValidationRegexString);
                        if (!pnRex.IsMatch(projectName))
                        {
                            BenLincoln.DebugSymbols.Utilities.ExitWithError(String.Format("The specified project name is invalid. The name must consist only of {0}.", ProjectNameExplanation), 1, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
                        }
                        i++;
                        break;
                    case "--output":
                        nextArg = args[i + 1];
                        baseOutputDirectoryPath = Path.GetFullPath(nextArg);
                        i++;
                        break;
                    case "--use-gp-base":
                        nextArg = args[i + 1];
                        explicitGPBase = true;
                        try
                        {
                            GPBase = BenLincoln.DebugSymbols.Utilities.GetHexOrDecimalLong(nextArg);
                        }
                        catch (Exception ex)
                        {
                            BenLincoln.DebugSymbols.Utilities.ExitWithError(ex.Message, 1, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
                        }
                        i++;
                        break;
                    case "--overlay":
                        nextArg = args[i + 1];
                        try
                        {
                            long newOverlayID = BenLincoln.DebugSymbols.Utilities.GetHexOrDecimalLong(nextArg);
                            if (!overlayIDs.Contains(newOverlayID))
                            {
                                overlayIDs.Add(newOverlayID);
                            }
                        }
                        catch (Exception ex)
                        {
                            BenLincoln.DebugSymbols.Utilities.ExitWithError(ex.Message, 1, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
                        }
                        i++;
                        break;
                    case "--output-updated-json":
                        nextArg = args[i + 1];
                        try
                        {
                            updatedJSONSymbolFilePath = Path.GetFullPath(nextArg);
                        }
                        catch (Exception ex)
                        {
                            BenLincoln.DebugSymbols.Utilities.ExitWithError(ex.Message, 1, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
                        }
                        i++;
                        break;
                    default:
                        arguments.Add(args[i]);
                        break;
                }
            }

            ArrayList positionalArgs = new ArrayList();

            foreach (string arg in arguments)
            {
                if ((arg == "-?") || (arg == "-h") || (arg == "--help"))
                {
                    BenLincoln.DebugSymbols.Utilities.DisplayHelpMessage(Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList, HelpMessageFile);
                    Environment.Exit(0);
                }
                bool handled = false;
                if (arg == "--overwrite")
                {
                    forceOverwrite = true;
                    handled = true;
                }
                if (arg == "--ignore-labels")
                {
                    ignoreLabels = true;
                    handled = true;
                }
                if (arg == "--ignore-non-ascii-labels")
                {
                    ignoreNonASCIILabels = true;
                    handled = true;
                }
                if (arg == "--replace-non-ascii-labels")
                {
                    replaceNonASCIILabels = true;
                    handled = true;
                }
                if (arg == "--externs-to-labels")
                {
                    convertExternalDefinitionsToLabels = true;
                    handled = true;
                }
                if (arg == "--add-stack-pointer-offset")
                {
                    addStackPointerOffset = true;
                    handled = true;
                }
                if (arg == "--assign-ra-storage")
                {
                    assignReturnAddressStorage = true;
                    handled = true;
                }
                if (arg == "--assign-parameter-storage")
                {
                    assignParameterStorage = true;
                    handled = true;
                }
                if (arg == "--create-playstation-memory")
                {
                    createPlayStationMemory = true;
                    handled = true;
                }
                if (arg == "--assume-sn-gp-base")
                {
                    assumeSNGPBase = true;
                    handled = true;
                }
                if (arg == "--map-sld-functions")
                {
                    mapSLDFunctions = true;
                    handled = true;
                }
                // right now, all positional arguments are file paths
                if (!handled)
                {
                    positionalArgs.Add(arg);
                }
            }

            if (positionalArgs.Count > 0)
            {
                inputFilePath = (string)positionalArgs[0];
            }

            if (inputFilePath == "")
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("No input SymDump file path was specified.", 1, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
            }

            if (!File.Exists(inputFilePath))
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("The specified input SymDump file does not exist.", 1, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
            }
            if (inputFilePath == updatedJSONSymbolFilePath)
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("The input SymDump file and the output (updated) SymDump JSON file cannot be the same file.", 1, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
            }

            if (baseOutputDirectoryPath == "")
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("No output directory was specified.", 1, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
            }
            if (forceOverwrite)
            {
                if (Directory.Exists(baseOutputDirectoryPath))
                {
                    try
                    {
                        Directory.Delete(baseOutputDirectoryPath, true);
                    }
                    catch (Exception ex)
                    {
                        BenLincoln.DebugSymbols.Utilities.ExitWithError(String.Format("The output directory '{0}' already exists, and could not be deleted: {1}", baseOutputDirectoryPath, ex.Message), 2, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
                    }
                }
            }
            /*
            else
            {
                if (Directory.Exists(baseOutputDirectoryPath))
                {
                    ExitWithError("The specified output directory already exists. Use the --overwrite option to replace it.", 1);
                }
            }
            */

            StringBuilder manualChangesBuilder = new StringBuilder();

            // get SymDumpTE output
            string symDumpRaw = "";
            try
            {
                symDumpRaw = BenLincoln.DebugSymbols.Utilities.GetFileAsString(inputFilePath);
            }
            catch (Exception ex)
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError(String.Format("Could not read the input file '{0}': {1}", inputFilePath, ex.Message), 2, 
                    Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
            }
            RenameJSONEntriesForGhidraResult rnjr = BenLincoln.DebugSymbols.GhidraScriptGenerator.RenameJSONEntriesForGhidra(symDumpRaw, manualChangesBuilder);
            manualChangesBuilder = rnjr.ManualChangesBuilder;
            symDumpRaw = rnjr.JSONText;
            SymDumpSymbolCollection symbolCollection = JsonConvert.DeserializeObject<SymDumpSymbolCollection>(symDumpRaw);

            List<PsyQOverlayDescription> includedOverlays = BenLincoln.DebugSymbols.SourceCodeGenerator.GetOverlayList(symbolCollection, overlayIDs, 
                Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);

            Assembly asm = Assembly.GetExecutingAssembly();

            string manualChangesPath = Path.Combine(baseOutputDirectoryPath, String.Format("README-{0}-CreateSkeleton-Manual_Changes_Required.txt", projectName));

            CodeElementFileMapper elemMapperMain = new CodeElementFileMapper();
            elemMapperMain.GetAllDataFromSymbolCollection(symbolCollection);

            // Ghidra scripts apply to the base code plus all overlays at once
            List<CFunction> allFunctions = new List<CFunction>();
            List<PsyQSLDDataSet> allSLDData = new List<PsyQSLDDataSet>();
            List<CLabel> allLabels = new List<CLabel>();
            List<CEnum> allEnums = new List<CEnum>();
            List<CStruct> allStructs = new List<CStruct>();
            List<CUnion> allUnions = new List<CUnion>();
            List <CTypeDef> allTypedefs = new List<CTypeDef>();
            List <CExternalDeclaration> allExterns = new List<CExternalDeclaration>();
            foreach (CFunction func in symbolCollection.functions)
            {
                allFunctions.Add(func);
            }
            foreach (PsyQSLDDataSet ds in symbolCollection.sld_data)
            {
                allSLDData.Add(ds);
            }
            foreach (CLabel l in symbolCollection.labels)
            {
                allLabels.Add(l);
            }
            foreach (CEnum e in symbolCollection.enums)
            {
                allEnums.Add(e);
            }
            foreach (CStruct s in symbolCollection.structs)
            {
                allStructs.Add(s);
            }
            foreach (CUnion u in symbolCollection.unions)
            {
                allUnions.Add(u);
            }
            foreach (CTypeDef t in symbolCollection.typedefs)
            {
                allTypedefs.Add(t);
            }
            foreach (CExternalDeclaration x in symbolCollection.external_declarations)
            {
                allExterns.Add(x);
            }

            foreach (CFunction func in symbolCollection.functions)
            {
                elemMapperMain.SetBestSourceFileMatch(func, func.name, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedSourceCodeFileName(BenLincoln.DebugSymbols.MetaGenerator.SubsetNameMainBinary), manualChangesBuilder);
            }
            foreach (CLabel l in symbolCollection.labels)
            {
                elemMapperMain.SetBestSourceFileMatch(l, l.name, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedSourceCodeFileName(BenLincoln.DebugSymbols.MetaGenerator.SubsetNameMainBinary), manualChangesBuilder);
            }
            // TKTK: re-enable this after improving the feature
            //foreach (CEnum e in symbolCollection.enums)
            //{
            //    elemMapperMain.SetBestSourceFileMatch(e, e.name, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedHeaderFileName(BenLincoln.DebugSymbols.MetaGenerator.SubsetNameMainBinary), manualChangesBuilder);
            //}
            //foreach (CStruct s in symbolCollection.structs)
            //{
            //    elemMapperMain.SetBestSourceFileMatch(s, s.name, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedHeaderFileName(BenLincoln.DebugSymbols.MetaGenerator.SubsetNameMainBinary), manualChangesBuilder);
            //}
            //foreach (CUnion u in symbolCollection.unions)
            //{
            //    elemMapperMain.SetBestSourceFileMatch(u, u.name, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedHeaderFileName(BenLincoln.DebugSymbols.MetaGenerator.SubsetNameMainBinary), manualChangesBuilder);
            //}
            //foreach (CTypeDef t in symbolCollection.typedefs)
            //{
            //    elemMapperMain.SetBestSourceFileMatch(t, t.typedef_signature, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedHeaderFileName(BenLincoln.DebugSymbols.MetaGenerator.SubsetNameMainBinary), manualChangesBuilder);
            //}
            //foreach (CExternalDeclaration x in symbolCollection.external_declarations)
            //{
            //    elemMapperMain.SetBestSourceFileMatch(x, x.declaration, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedHeaderFileName(BenLincoln.DebugSymbols.MetaGenerator.SubsetNameMainBinary), manualChangesBuilder);
            //}

            foreach (PsyQOverlayDescription ol in includedOverlays)
            {
                CodeElementFileMapper elemMapperOverlay = new CodeElementFileMapper();
                elemMapperOverlay.GetAllDataFromOverlay(ol);
                foreach (CFunction func in ol.functions)
                {
                    allFunctions.Add(func);
                }
                foreach (PsyQSLDDataSet ds in ol.sld_data)
                {
                    allSLDData.Add(ds);
                }
                foreach (CLabel l in ol.labels)
                {
                    allLabels.Add(l);
                }
                foreach (CEnum e in ol.enums)
                {
                    allEnums.Add(e);
                }
                foreach (CStruct s in ol.structs)
                {
                    allStructs.Add(s);
                }
                foreach (CUnion u in ol.unions)
                {
                    allUnions.Add(u);
                }
                foreach (CTypeDef t in ol.typedefs)
                {
                    allTypedefs.Add(t);
                }
                foreach (CExternalDeclaration x in ol.external_declarations)
                {
                    allExterns.Add(x);
                }

                string shortOverlayName = BenLincoln.DebugSymbols.MetaGenerator.GetShortOverlayName(ol.overlay_id);

                foreach (CFunction func in ol.functions)
                {
                    elemMapperOverlay.SetBestSourceFileMatch(func, func.name, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedSourceCodeFileName(shortOverlayName), manualChangesBuilder);
                }
                foreach (CLabel l in ol.labels)
                {
                    elemMapperOverlay.SetBestSourceFileMatch(l, l.name, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedSourceCodeFileName(shortOverlayName), manualChangesBuilder);
                }
                foreach (CEnum e in ol.enums)
                {
                    elemMapperOverlay.SetBestSourceFileMatch(e, e.name, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedHeaderFileName(shortOverlayName), manualChangesBuilder);
                }
                foreach (CStruct s in ol.structs)
                {
                    elemMapperOverlay.SetBestSourceFileMatch(s, s.name, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedHeaderFileName(shortOverlayName), manualChangesBuilder);
                }
                foreach (CUnion u in ol.unions)
                {
                    elemMapperOverlay.SetBestSourceFileMatch(u, u.name, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedHeaderFileName(shortOverlayName), manualChangesBuilder);
                }
                foreach (CTypeDef t in ol.typedefs)
                {
                    elemMapperOverlay.SetBestSourceFileMatch(t, t.typedef_signature, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedHeaderFileName(shortOverlayName), manualChangesBuilder);
                }
                foreach (CExternalDeclaration x in ol.external_declarations)
                {
                    elemMapperOverlay.SetBestSourceFileMatch(x, x.declaration, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedHeaderFileName(shortOverlayName), manualChangesBuilder);
                }
            }

            //symbolCollection.SortAllElements();

            BenLincoln.DebugSymbols.GenerateGhidraScriptsResult gsfr = BenLincoln.DebugSymbols.MetaGenerator.GenerateGhidraScripts(
                symbolCollection.program_counter, symbolCollection.stack_pointer_base, symbolCollection.stack_pointer_size_or_offset,
                symbolCollection.global_pointer, 
                symbolCollection.memory_segments,
                allFunctions, allSLDData, allLabels, allEnums, allStructs, allUnions, allTypedefs, allExterns, 
                includedOverlays, projectName, baseOutputDirectoryPath, assumeSNGPBase, explicitGPBase, GPBase, addStackPointerOffset, 
                createPlayStationMemory, ignoreLabels, ignoreNonASCIILabels, replaceNonASCIILabels, convertExternalDefinitionsToLabels, mapSLDFunctions,
                assignReturnAddressStorage, assignParameterStorage,
                manualChangesBuilder,
                symbolCollection.ToGhidraXML(String.Format("{0}.SYM", projectName), String.Format("{0}.EXE", projectName)),
                asm, ProgramName, Copyright, InfoURL, ThanksList);
            manualChangesBuilder = gsfr.ManualChangesBuilder;

            // add any generated typedefs to the symbol collection
            CodeElementFileMapper elemMapperMain2 = new CodeElementFileMapper();
            elemMapperMain2.GetAllDataFromSymbolCollection(symbolCollection);
            foreach (long overlayNum in gsfr.FunctionPointerTypeDefs.Keys)
            {
                CodeElementFileMapper elemMapperOverlay2 = new CodeElementFileMapper();
                string shortOverlayName = "";
                if (overlayNum > -1)
                {
                    elemMapperOverlay2.GetAllDataFromOverlay(symbolCollection.GetOverlayByID(overlayNum));
                    shortOverlayName = BenLincoln.DebugSymbols.MetaGenerator.GetShortOverlayName(overlayNum);
                }                

                for (int i = 0; i < gsfr.FunctionPointerTypeDefs[overlayNum].Count; i++)
                {
                    if (overlayNum == -1)
                    {
                        
                        elemMapperMain2.SetBestSourceFileMatch(gsfr.FunctionPointerTypeDefs[overlayNum][i], gsfr.FunctionPointerTypeDefs[overlayNum][i].typedef_signature, 
                            BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedHeaderFileName(BenLincoln.DebugSymbols.MetaGenerator.SubsetNameMainBinary), manualChangesBuilder);
                        if (!symbolCollection.generated_typedefs.Contains(gsfr.FunctionPointerTypeDefs[overlayNum][i]))
                        {
                            symbolCollection.generated_typedefs.Add(gsfr.FunctionPointerTypeDefs[overlayNum][i]);
                        }
                    }
                    else
                    {
                        elemMapperOverlay2.SetBestSourceFileMatch(gsfr.FunctionPointerTypeDefs[overlayNum][i], gsfr.FunctionPointerTypeDefs[overlayNum][i].typedef_signature, BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedHeaderFileName(shortOverlayName), manualChangesBuilder);
                        if (!symbolCollection.GetOverlayByID(overlayNum).generated_typedefs.Contains(gsfr.FunctionPointerTypeDefs[overlayNum][i]))
                        {
                            symbolCollection.GetOverlayByID(overlayNum).generated_typedefs.Add(gsfr.FunctionPointerTypeDefs[overlayNum][i]);
                        }
                    }
                }
            }

            //symbolCollection.SortAllElements();

            // everything else is handled in a segmented manner
            // base program code
            string baseProgramDirectoryName = BenLincoln.DebugSymbols.MetaGenerator.SubsetNameMainBinary;
            //if (includedOverlays.Count == 0)
            //{
            //    baseProgramDirectoryName = "";
            //}
            GenerateSkeletonFilesForSubsetResult gsffs = BenLincoln.DebugSymbols.MetaGenerator.GenerateSkeletonFilesForSubset(projectName, -1, baseProgramDirectoryName, baseOutputDirectoryPath, 
                gsfr, mapSLDFunctions, symbolCollection.functions, symbolCollection.sld_data, symbolCollection.labels, symbolCollection.enums,
                symbolCollection.structs, symbolCollection.unions, symbolCollection.typedefs, symbolCollection.generated_typedefs, 
                symbolCollection.external_declarations, "CreateSkeleton", BenLincoln.DebugSymbols.MetaGenerator.DirectoryNameSourceStubs,
                asm, ProgramName, Copyright, InfoURL, ThanksList);
            symbolCollection.AddFunctions(gsffs.AddedFunctions);
            // overlays
            foreach (PsyQOverlayDescription ol in includedOverlays)
            {
                GenerateSkeletonFilesForSubsetResult gsffso = BenLincoln.DebugSymbols.MetaGenerator.GenerateSkeletonFilesForSubset(projectName, ol.overlay_id, ol.GetDirectoryNameForDecompilation(), 
                    baseOutputDirectoryPath, gsfr, mapSLDFunctions, ol.functions, ol.sld_data, ol.labels, ol.enums,
                    ol.structs, ol.unions, ol.typedefs, ol.generated_typedefs, ol.external_declarations,
                    "CreateSkeleton", BenLincoln.DebugSymbols.MetaGenerator.DirectoryNameSourceStubs,
                    asm, ProgramName, Copyright, InfoURL, ThanksList);
                ol.AddFunctions(gsffso.AddedFunctions);
            }

            if (includedOverlays.Count == 0)
            {
                BenLincoln.DebugSymbols.Utilities.HandleWarningForOverlays(symbolCollection);
            }

            BenLincoln.DebugSymbols.Utilities.WriteToOutput(manualChangesPath, manualChangesBuilder.ToString(), FileMode.Create);

            // if specified, output an updated version of the symbol file for use with PopulateSkeleton
            if (updatedJSONSymbolFilePath != "")
            {
                string result = symbolCollection.ToJSON();
                try
                {
                    BenLincoln.DebugSymbols.Utilities.WriteToOutput(updatedJSONSymbolFilePath, result, FileMode.Create);
                }
                catch (Exception ex)
                {
                    BenLincoln.DebugSymbols.Utilities.ExitWithError(String.Format("Could not write to the output file '{0}': {1}", updatedJSONSymbolFilePath, ex.Message), 1, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
                }
            }

            Console.WriteLine("");
        }


    }
}
