﻿// PlayStation ELF Converter
// Converts between ELF and PSX-EXE executable formats
// Copyright 2019 Ben Lincoln
// https://www.beneaththewaves.net/
// Based in part on information documented in LameGuy64's elf2x utility
// (http://www.psxdev.net/forum/viewtopic.php?f=75&t=1613)

// This file is part of PlayStation ELF Converter.

// PlayStation ELF Converter is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// PlayStation ELF Converter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with PlayStation ELF Converter (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace BenLincoln.ExecutableFiles
{
    public class ELFProgramHeader
    {
        public enum ELFProgramType
        {
            SegmentTypeNull = 0x00000000,
            SegmentTypeLoadable = 0x00000001,
            SegmentTypeDynamicLinkingInfo = 0x00000002,
            SegmentTypeInterpreterInfo = 0x00000003,
            SegmentTypeAuxiliaryInfo = 0x00000004,
            SegmentTypeReserved = 0x00000005,
            SegmentTypeProgramHeaderTable = 0x00000006,
            SegmentTypeOSReservedLow = 0x60000000,
            SegmentTypeOSReservedHigh = 0x6FFFFFFF,
            SegmentTypeProcessReservedLow = 0x70000000,
            SegmentTypeProcessReservedHigh = 0x7FFFFFFF
        }

        public uint SegmentType = 0;                // bytes 0-3
        public uint Offset = 0;              // bytes 4-7    - segment offset in file
        public uint VirtualAddress = 0;      // bytes 8-11   - segment base virtual address in memory
        public uint PhysicalAddress = 0;     // bytes 12-15  - not used by the PlayStation
        public uint FileSize = 0;            // bytes 16-19  - size (in bytes) of the segment in the file
        public uint MemSize = 0;             // bytes 20-23  - size (in bytes) of the segment in memory
        public uint Flags = 0;                      // bytes 24-27
        public uint Alignment = 0;                  // bytes 28-31

        public ELFProgramHeader()
        {
        }

        public string GetSegmentTypeName()
        {
            string result = "Undefined";

            if (SegmentType != null)
            {
                switch (SegmentType)
                {
                    case (uint)ELFProgramType.SegmentTypeNull:
                        result = "Null";
                        break;
                    case (uint)ELFProgramType.SegmentTypeLoadable:
                        result = "Loadable";
                        break;
                    case (uint)ELFProgramType.SegmentTypeDynamicLinkingInfo:
                        result = "Dynamic Linking Information";
                        break;
                    case (uint)ELFProgramType.SegmentTypeInterpreterInfo:
                        result = "Interpreter Information";
                        break;
                    case (uint)ELFProgramType.SegmentTypeAuxiliaryInfo:
                        result = "Auxiliary Information";
                        break;
                    case (uint)ELFProgramType.SegmentTypeReserved:
                        result = "Reserved";
                        break;
                    case (uint)ELFProgramType.SegmentTypeProgramHeaderTable:
                        result = "Program Header Table";
                        break;
                    default:
                        result = "Unknown";
                        if ((SegmentType >= (uint)ELFProgramType.SegmentTypeOSReservedLow) && (SegmentType <= (uint)ELFProgramType.SegmentTypeOSReservedHigh))
                        {
                            result = "OS Reserved";
                        }
                        if ((SegmentType >= (uint)ELFProgramType.SegmentTypeProcessReservedLow) && (SegmentType <= (uint)ELFProgramType.SegmentTypeProcessReservedHigh))
                        {
                            result = "Process Reserved";
                        }
                        break;
                }
                result = String.Format("{0} (0x{1:X8})", result, SegmentType);
            }
            return result;
        }

    }

    public class ELFSectionHeader
    {
        public enum ELFSectionType
        {
            SectionTypeNull = 0,
            SectionTypeProgramData = 1,
            SectionTypeSymbolTable = 2,
            SectionTypeStringTable = 3,
            SectionTypeRelocationEntriesWithAddends = 4,
            SectionTypeSymbolHashTable = 5,
            SectionTypeDynamicLinkingInfo = 6,
            SectionTypeNotes = 7,
            SectionTypeBSS = 8,
            SectionTypeRelocationEntriesWithoutAddends = 9,
            SectionTypeReserved = 0x0A,
            SectionTypeDynamicLinkerSymbolTable = 0x0B,
            SectionTypeConstructorArray = 0x0E,
            SectionTypeDestructorArray = 0x0F,
            SectionTypePreconstructorArray = 0x10,
            SectionTypeSectionGroup = 0x11,
            SectionTypeExtendedSectionIndices = 0x12,
            SectionTypeNumberOfDefinedTypes = 0x13,
            SectionrTypeOSSpecificLow = 0x60000000
        }

        public uint NameTableOffset = 0;
        public uint SectionType = 0;
        public uint Flags = 0;
        public uint VirtualAddress = 0;
        public uint Offset = 0;
        public uint FileSize = 0;
        public uint LinkIndex = 0;
        public uint Info = 0;
        public uint Alignment = 0;
        public uint EntrySize = 0;

        public ELFSectionHeader()
        {
        }

        public string GetSectionTypeName()
        {
            string result = "Undefined";

            if (SectionType != null)
            {
                switch (SectionType)
                {
                    case (uint)ELFSectionType.SectionTypeNull:
                        result = "Null";
                        break;
                    case (uint)ELFSectionType.SectionTypeProgramData:
                        result = "Program Data";
                        break;
                    case (uint)ELFSectionType.SectionTypeSymbolTable:
                        result = "Symbol Table";
                        break;
                    case (uint)ELFSectionType.SectionTypeStringTable:
                        result = "String Table";
                        break;
                    case (uint)ELFSectionType.SectionTypeRelocationEntriesWithAddends:
                        result = "Relocation Entries With Addends";
                        break;
                    case (uint)ELFSectionType.SectionTypeSymbolHashTable:
                        result = "Symbol Hash Table";
                        break;
                    case (uint)ELFSectionType.SectionTypeDynamicLinkingInfo:
                        result = "Dynamic Linking Info";
                        break;
                    case (uint)ELFSectionType.SectionTypeNotes:
                        result = "Notes";
                        break;
                    case (uint)ELFSectionType.SectionTypeBSS:
                        result = "BSS";
                        break;
                    case (uint)ELFSectionType.SectionTypeRelocationEntriesWithoutAddends:
                        result = "Relocation Entries Without Addends";
                        break;
                    case (uint)ELFSectionType.SectionTypeReserved:
                        result = "Reserved";
                        break;
                    case (uint)ELFSectionType.SectionTypeDynamicLinkerSymbolTable:
                        result = "Dynamic Linker Symbol Table";
                        break;
                    case (uint)ELFSectionType.SectionTypeConstructorArray:
                        result = "Constructor Array";
                        break;
                    case (uint)ELFSectionType.SectionTypeDestructorArray:
                        result = "Destructor Array";
                        break;
                    case (uint)ELFSectionType.SectionTypePreconstructorArray:
                        result = "Preconstructor Array";
                        break;
                    case (uint)ELFSectionType.SectionTypeSectionGroup:
                        result = "Section Group";
                        break;
                    case (uint)ELFSectionType.SectionTypeExtendedSectionIndices:
                        result = "Extended Section Indices";
                        break;
                    case (uint)ELFSectionType.SectionTypeNumberOfDefinedTypes:
                        result = "Number Of Defined Types";
                        break;
                    default:
                        result = "Unknown";
                        if (SectionType >= (uint)ELFSectionType.SectionrTypeOSSpecificLow)
                        {
                            result = "OS Reserved";
                        }
                        break;
                }
                result = String.Format("{0} (0x{1:X8})", result, SectionType);
            }
            return result;
        }

    }

    public struct ELFProgram
    {
        public byte[] ProgramData;
    }

    public struct ELFSection
    {
        public byte[] SectionData;
    }

    public class ELF : BinaryFile
    {
        // magic number for ELF files
        public const uint ELFMagic = 0x464C457F;
        // minimum valid ELF file size
        protected const int MinSize = 82;

        public uint Magic;                  // bytes 0-3
        public byte WordSize;               // byte 4
        public byte Endianness;             // byte 5
        public byte ELFVersion;             // byte 6
        public byte OSABI;                  // byte 7
        public uint Unused1;                // bytes 8-11
        public uint Unused2;                // bytes 12-15

        public ushort ELFType;                 // bytes 16-17
        public ushort InstructionSet;             // bytes 18-19
        public uint ELFVersion2;            // bytes 20-23

        public uint ProgramEntryAddress;    // bytes 24-27
        public uint ProgramHeaderTableOffset;    // bytes 28-31
        public uint ProgramSectionHeaderTableOffset;     // bytes 32-35
        public uint Flags;                  // bytes 36-39
        public ushort HeaderSize;           // bytes 40-41
        public ushort ProgramEntrySize;     // bytes 42-43
        public ushort ProgramEntryCount;    // bytes 44-45
        public ushort SectionEntrySize;         // bytes 46-47
        public ushort SectionEntryCount;        // bytes 48-49
        public ushort SectionNamesIndex;        // bytes 50-51

        public ELFProgramHeader[] ProgramHeaders;
        public ELFProgram[] ProgramDataArray;
        public ELFSectionHeader[] SectionHeaders;
        //public string[] SectionNames;
        //public byte[] SectionNames;
        public ELFSection[] SectionDataArray;

        protected const uint DefaultAlignment = 0x00001000;
        protected const uint DefaultPlayStationSectionAttributes = 0x00000007;  // read/write/execute

        public ELF()
        {
            _FileTypeName = "ELF";
        }

        public override void ReadFile(BinaryReader binaryReader)
        {
            if (binaryReader.BaseStream.Length < MinSize)
            {
                throw new Exception(String.Format("This file is {0} bytes long, but at least {1} bytes are required.", binaryReader.BaseStream.Length, MinSize));
            }
            binaryReader.BaseStream.Seek(0, SeekOrigin.Begin);
            Magic = binaryReader.ReadUInt32();
            if (Magic != ELFMagic)
            {
                throw new Exception(String.Format("This file is not an ELF. Expected {0:X8} as the magic number, but found {1:X8}", ELFMagic, Magic));
            }
            WordSize = binaryReader.ReadByte();
            if (WordSize != 1)
            {
                throw new Exception(String.Format("This file is an ELF, but it does not have a 32-bit word size. Expected word size 1, found word size {0}", WordSize));
            }
            Endianness = binaryReader.ReadByte();
            if (Endianness != 1)
            {
                throw new Exception(String.Format("This file is an ELF, but it is not little-endian. Expected endianness value 1, found endianness value {0}", Endianness));
            }
            ELFVersion = binaryReader.ReadByte();
            OSABI = binaryReader.ReadByte();
            Unused1 = binaryReader.ReadUInt32();
            Unused2 = binaryReader.ReadUInt32();

            ELFType = binaryReader.ReadUInt16();
            if (ELFType != 2)
            {
                throw new Exception(String.Format("This file is an ELF, but its type is not executable. Expected type 2, found type {0}", ELFType));
            }
            InstructionSet = binaryReader.ReadUInt16();
            if (InstructionSet != 8)
            {
                throw new Exception(String.Format("This file is an ELF, but its instruction set is not MIPS. Expected instruction set 8, found instruction set {0}", InstructionSet));
            }
            ELFVersion2 = binaryReader.ReadUInt32();

            ProgramEntryAddress = binaryReader.ReadUInt32();
            ProgramHeaderTableOffset = binaryReader.ReadUInt32();
            ProgramSectionHeaderTableOffset = binaryReader.ReadUInt32();
            Flags = binaryReader.ReadUInt32();
            HeaderSize = binaryReader.ReadUInt16();
            ProgramEntrySize = binaryReader.ReadUInt16();
            ProgramEntryCount = binaryReader.ReadUInt16();
            SectionEntrySize = binaryReader.ReadUInt16();
            SectionEntryCount = binaryReader.ReadUInt16();
            SectionNamesIndex = binaryReader.ReadUInt16();

            ProgramHeaders = new ELFProgramHeader[ProgramEntryCount];
            ProgramDataArray = new ELFProgram[ProgramEntryCount];

            // get program headers and associated data
            // this is USUALLY right after the ELF header, but not required to be
            binaryReader.BaseStream.Seek(ProgramHeaderTableOffset, SeekOrigin.Begin);
            try
            {
                for (int i = 0; i < ProgramEntryCount; i++)
                {
                    ELFProgramHeader newHeader = new ELFProgramHeader();
                    newHeader.SegmentType = binaryReader.ReadUInt32();
                    newHeader.Offset = binaryReader.ReadUInt32();
                    newHeader.VirtualAddress = binaryReader.ReadUInt32();
                    newHeader.PhysicalAddress = binaryReader.ReadUInt32();
                    newHeader.FileSize = binaryReader.ReadUInt32();
                    newHeader.MemSize = binaryReader.ReadUInt32();
                    newHeader.Flags = binaryReader.ReadUInt32();
                    newHeader.Alignment = binaryReader.ReadUInt32();
                    ProgramHeaders[i] = newHeader;
                }
            }
            catch (Exception ex)
            {
                throw new Exception(String.Format("Could not parse the ELF file due to the following error while reading its segment headers: {0}", ex.Message));
            }

            try
            {
                for (int i = 0; i < ProgramEntryCount; i++)
                {
                    ELFProgram newSegment = new ELFProgram();
                    binaryReader.BaseStream.Seek(ProgramHeaders[i].Offset, SeekOrigin.Begin);
                    newSegment.ProgramData = binaryReader.ReadBytes((int)ProgramHeaders[i].FileSize);
                    ProgramDataArray[i] = newSegment;
                }
            }
            catch (Exception ex)
            {
                throw new Exception(String.Format("Could not parse the ELF file due to the following error while reading its programs: {0}", ex.Message));
            }

            // get section headers and associated data
            SectionHeaders = new ELFSectionHeader[SectionEntryCount];
            SectionDataArray = new ELFSection[SectionEntryCount];
            try
            {
                binaryReader.BaseStream.Seek(ProgramSectionHeaderTableOffset, SeekOrigin.Begin);
                for (int i = 0; i < SectionEntryCount; i++)
                {
                    ELFSectionHeader newHeader = new ELFSectionHeader();
                    newHeader.NameTableOffset = binaryReader.ReadUInt32();
                    newHeader.SectionType = binaryReader.ReadUInt32();
                    newHeader.Flags = binaryReader.ReadUInt32();
                    newHeader.VirtualAddress = binaryReader.ReadUInt32();
                    newHeader.Offset = binaryReader.ReadUInt32();
                    newHeader.FileSize = binaryReader.ReadUInt32();
                    newHeader.LinkIndex = binaryReader.ReadUInt32();
                    newHeader.Info = binaryReader.ReadUInt32();
                    newHeader.Alignment = binaryReader.ReadUInt32();
                    newHeader.EntrySize = binaryReader.ReadUInt32();
                    SectionHeaders[i] = newHeader;
                }
            }
            catch (Exception ex)
            {
                throw new Exception(String.Format("Could not parse the ELF file due to the following error while reading its section headers: {0}", ex.Message));
            }

            try
            {
                for (int i = 0; i < SectionEntryCount; i++)
                {
                    ELFSection newSection = new ELFSection();
                    binaryReader.BaseStream.Seek(SectionHeaders[i].Offset, SeekOrigin.Begin);
                    newSection.SectionData = binaryReader.ReadBytes((int)SectionHeaders[i].FileSize);
                    SectionDataArray[i] = newSection;
                }
            }
            catch (Exception ex)
            {
                throw new Exception(String.Format("Could not parse the ELF file due to the following error while reading its sections: {0}", ex.Message));
            }
        }

        protected int GetOffsetOfSectionName(Dictionary<int, string> sectionNames, string name)
        {
            int offset = 0;
            List<int> sectionNums = new List<int>();
            foreach (int sN in sectionNames.Keys)
            {
                sectionNums.Add(sN);
            }
            sectionNums.Sort();

            foreach (int sN in sectionNums)
            {
                if (sectionNames[sN] == name)
                {
                    return offset + 1;
                }
                offset += sectionNames[sN].Length;
                offset++;   // null byte delimiter
            }
            throw new Exception(string.Format("Could not find a section named '{0}' in the list of section names.", name));
        }

        /* 
         So I spent all this time writing a better version of this function that could handle tacking an arbitrary 
         number of segments on, only to discover that PsyQ collapses everything into one giant segment even though
         it shows individual segments for the overlaps in its MAP files.

            Maybe it will come in handy someday anyway
             */
        public void SetFromPSXEXE(PSXEXE psxExe, MemorySegment[] additionalSegments)
        {
            Magic = ELFMagic;
            WordSize = 1;
            Endianness = 1;
            ELFVersion = 1;
            OSABI = 0; 
            Unused1 = 0;
            Unused2 = 0;
            ELFType = 2;
            InstructionSet = 8;
            ELFVersion2 = 1;

            ProgramEntryAddress = psxExe.PSXExec.ProgramCounter;
            ProgramHeaderTableOffset = 0x00000034;
            Flags = 0x00000001; // this is based on mipsgcc example output
            HeaderSize = 0x0034;
            ProgramEntryCount = (ushort)(1 + additionalSegments.Length);
            ProgramEntrySize = 0x0020;
            SectionEntryCount = (ushort)(7 + additionalSegments.Length);
            SectionEntrySize = 0x0028;
            //SectionNamesIndex = (ushort)(6 + additionalSegments.Length);
            SectionNamesIndex = 6;

            ProgramHeaders = new ELFProgramHeader[ProgramEntryCount];

            ELFProgramHeader newHeader = new ELFProgramHeader();
            newHeader.SegmentType = (uint)ELFProgramHeader.ELFProgramType.SegmentTypeLoadable;
            newHeader.Offset = 0x1000; // this is based on mipsgcc example output
            newHeader.VirtualAddress = psxExe.PSXExec.TextSegmentAddress;
            newHeader.PhysicalAddress = psxExe.PSXExec.TextSegmentAddress;
            newHeader.FileSize = (uint)psxExe.ProgramCode.Length;
            newHeader.MemSize = (uint)psxExe.PSXExec.TextSegmentSize;
            newHeader.Flags = 0x00000007; // this is based on mipsgcc example output
            newHeader.Alignment = DefaultAlignment; // 0x00001000 is based on mipsgcc example output
            ProgramHeaders[0] = newHeader;

            ProgramDataArray = new ELFProgram[ProgramEntryCount];
            ProgramDataArray[0].ProgramData = psxExe.ProgramCode;

            int phNum = 1;
            uint currentOffset = newHeader.Offset + newHeader.FileSize;
            foreach (MemorySegment aSeg in additionalSegments)
            {
                ELFProgramHeader segHeader = new ELFProgramHeader();
                segHeader.SegmentType = (uint)ELFProgramHeader.ELFProgramType.SegmentTypeLoadable;
                while (currentOffset % DefaultAlignment > 0)
                {
                    currentOffset++;
                }
                segHeader.Offset = currentOffset; // this is based on mipsgcc example output
                segHeader.VirtualAddress = aSeg.Offset;
                segHeader.PhysicalAddress = aSeg.Offset;
                segHeader.FileSize = (uint)aSeg.Length;
                segHeader.MemSize = (uint)aSeg.Length;
                segHeader.Flags = 0x00000007; // this is based on mipsgcc example output
                segHeader.Alignment = 0x00001000; // this is based on mipsgcc example output
                ProgramHeaders[phNum] = segHeader;
                ProgramDataArray[phNum].ProgramData = aSeg.Data;
                phNum++;
                currentOffset += (uint)aSeg.Length;
            }

            // raw byte version of the following null-byte-delimited strings:
            // .symtab
            // .strtab
            // .shstrtab
            // .text
            // .rodata
            // .data
            // .bss
            // .reginfo
            // byte[] sectionNameData = new byte[] { 0x00, 0x2E, 0x73, 0x79, 0x6D, 0x74, 0x61, 0x62, 0x00, 0x2E, 0x73, 0x74, 0x72, 0x74, 0x61, 0x62, 0x00, 0x2E, 0x73, 0x68, 0x73, 0x74, 0x72, 0x74, 0x61, 0x62, 0x00, 0x2E, 0x74, 0x65, 0x78, 0x74, 0x00, 0x2E, 0x72, 0x6F, 0x64, 0x61, 0x74, 0x61, 0x00, 0x2E, 0x64, 0x61, 0x74, 0x61, 0x00, 0x2E, 0x62, 0x73, 0x73, 0x00, 0x2E, 0x72, 0x65, 0x67, 0x69, 0x6E, 0x66, 0x6F };

            Dictionary<int, string> sectionNames = new Dictionary<int, string>();
            sectionNames.Add(0, ".symtab");
            sectionNames.Add(1, ".strtab");
            sectionNames.Add(2, ".shstrtab");
            sectionNames.Add(3, ".text");
            sectionNames.Add(4, ".rodata");
            sectionNames.Add(5, ".data");
            sectionNames.Add(6, ".bss");
            sectionNames.Add(7, ".reginfo");
            int sN = 8;
            foreach (MemorySegment aSeg in additionalSegments)
            {
                sectionNames.Add(sN, aSeg.Name);
                sN++;
            }
            List<byte> sectionNameBytes = new List<byte>();
            //sectionNameBytes.Add(0);
            for (int i = 0; i < sN; i++)
            {
                sectionNameBytes.Add(0);
                foreach (byte b in Encoding.ASCII.GetBytes(sectionNames[i]))
                {
                    sectionNameBytes.Add(b);
                }                
            }
            byte[] sectionNameData = sectionNameBytes.ToArray();

            // more-or-less hardcoded construction of a valid set of sections for a stripped ELF
            SectionHeaders = new ELFSectionHeader[SectionEntryCount];
            SectionDataArray = new ELFSection[SectionEntryCount];

            // null section
            SectionHeaders[0] = new ELFSectionHeader();
            SectionDataArray[0] = new ELFSection();
            SectionDataArray[0].SectionData = new byte[0];

            // program code - basically matches what's in the program entry
            // in original GCC-compiled code, this would only contain the text section
            SectionHeaders[1] = new ELFSectionHeader();
            SectionDataArray[1] = new ELFSection();
            SectionDataArray[1].SectionData = new byte[0];
            SectionHeaders[1].SectionType = (uint)ELFSectionHeader.ELFSectionType.SectionTypeProgramData;
            SectionHeaders[1].VirtualAddress = psxExe.PSXExec.TextSegmentAddress;
            //SectionHeaders[1].NameTableOffset = 0x1B;
            SectionHeaders[1].NameTableOffset = (uint)GetOffsetOfSectionName(sectionNames, ".text");
            SectionHeaders[1].Flags = DefaultPlayStationSectionAttributes;
            SectionHeaders[1].Offset = ProgramHeaders[0].Offset;
            SectionHeaders[1].FileSize = ProgramHeaders[0].FileSize;
            SectionHeaders[1].Alignment = 4;

            // where does the program code end in the file?
            uint postCodeOffset = ProgramHeaders[ProgramEntryCount - 1].Offset + ProgramHeaders[ProgramEntryCount - 1].FileSize;
            uint postCodeVirtualAddress = psxExe.PSXExec.TextSegmentAddress + psxExe.PSXExec.TextSegmentSize;

            // read-only data - unused currently, would contain strings in code from GCC, but those go in the 
            // unified text section in this model
            SectionHeaders[2] = new ELFSectionHeader();
            SectionDataArray[2] = new ELFSection();
            SectionDataArray[2].SectionData = new byte[0];
            SectionHeaders[2].SectionType = (uint)ELFSectionHeader.ELFSectionType.SectionTypeProgramData;
            //SectionHeaders[2].NameTableOffset = 0x21;
            SectionHeaders[2].NameTableOffset = (uint)GetOffsetOfSectionName(sectionNames, ".rodata");
            SectionHeaders[2].Flags = 0x00000002;
            SectionHeaders[2].Offset = postCodeOffset;
            SectionHeaders[2].VirtualAddress = postCodeVirtualAddress;
            SectionHeaders[2].FileSize = 0;
            SectionHeaders[2].Alignment = 1;

            // data - unused currently
            SectionHeaders[3] = new ELFSectionHeader();
            SectionDataArray[3] = new ELFSection();
            SectionDataArray[3].SectionData = new byte[0];
            SectionHeaders[3].SectionType = (uint)ELFSectionHeader.ELFSectionType.SectionTypeProgramData;
            //SectionHeaders[3].NameTableOffset = 0x29;
            SectionHeaders[3].NameTableOffset = (uint)GetOffsetOfSectionName(sectionNames, ".data");
            SectionHeaders[3].Flags = 0x00000003;
            SectionHeaders[3].Offset = postCodeOffset;
            SectionHeaders[3].VirtualAddress = postCodeVirtualAddress;
            SectionHeaders[3].FileSize = 0;
            SectionHeaders[3].Alignment = 1;

            // bss - unused currently
            SectionHeaders[4] = new ELFSectionHeader();
            SectionDataArray[4] = new ELFSection();
            SectionDataArray[4].SectionData = new byte[0];
            SectionHeaders[4].SectionType = (uint)ELFSectionHeader.ELFSectionType.SectionTypeBSS;
            //SectionHeaders[4].NameTableOffset = 0x2F;
            SectionHeaders[4].NameTableOffset = (uint)GetOffsetOfSectionName(sectionNames, ".bss");
            SectionHeaders[4].Flags = 0x00000003;
            SectionHeaders[4].Offset = postCodeOffset;
            SectionHeaders[4].VirtualAddress = postCodeVirtualAddress;
            SectionHeaders[4].FileSize = 0;
            SectionHeaders[4].Alignment = 1;

            // reginfo (register state information) - not guaranteed to be accurate
            byte[] regInfoData = new byte[24];
            // first four bytes are supposed to represent the bit-mask of general-purpose registers used by the binary
            // this is what a very basic ELF I compiled using mipsgcc looked like:
            //regInfoData[0] = 0x1C;
            //regInfoData[1] = 0x03;
            //regInfoData[2] = 0x00;
            //regInfoData[3] = 0xB0;
            // I'll just set all of the bits - what could possibly go wrong?
            // Seems better than missing any, though, and getting an accurate list would require decompiling the binary
            regInfoData[0] = 0xFF;
            regInfoData[1] = 0xFF;
            regInfoData[2] = 0xFF;
            regInfoData[3] = 0xFF;
            // next four bytes represent the bit-mask of co-processor registers used by the binary
            // hope it's OK for them to all be zeroes

            // the next three sets of four bytes are for three additional co-processors, but use of them is disallowed for ABI-compliant programs
            // again, hope it's OK for them to all be zeroes

            // last four bytes are the GP register value
            // I think this is always zero in actual PSX-EXE files, but try to handle it correctly
            regInfoData[20] = (byte)(psxExe.PSXExec.SavedGlobalPointer & 0xFF);
            regInfoData[21] = (byte)((psxExe.PSXExec.SavedGlobalPointer >> 8) & 0xFF);
            regInfoData[22] = (byte)((psxExe.PSXExec.SavedGlobalPointer >> 16) & 0xFF);
            regInfoData[23] = (byte)((psxExe.PSXExec.SavedGlobalPointer >> 24) & 0xFF);

            // make sure the reginfo data is aligned correctly
            uint reginfoOffset = postCodeOffset;
            while (reginfoOffset % 4 != 0)
            {
                reginfoOffset++;
            }
            uint registerInfoVirtualAddress = postCodeVirtualAddress;
            while (registerInfoVirtualAddress % 4 != 0)
            {
                registerInfoVirtualAddress++;
            }

            SectionHeaders[5] = new ELFSectionHeader();
            SectionDataArray[5] = new ELFSection();
            SectionDataArray[5].SectionData = regInfoData;
            SectionHeaders[5].SectionType = (uint)0x70000006;
            //SectionHeaders[5].NameTableOffset = 0x34;
            SectionHeaders[5].NameTableOffset = (uint)GetOffsetOfSectionName(sectionNames, ".reginfo");
            SectionHeaders[5].Flags = 0x00000000;
            SectionHeaders[5].Offset = reginfoOffset;
            SectionHeaders[5].VirtualAddress = registerInfoVirtualAddress;
            SectionHeaders[5].FileSize = (uint)regInfoData.Length;
            SectionHeaders[5].Alignment = 4;
            SectionHeaders[5].EntrySize = 1;

            
            SectionHeaders[6] = new ELFSectionHeader();
            SectionDataArray[6] = new ELFSection();
            SectionDataArray[6].SectionData = sectionNameData;
            SectionHeaders[6].SectionType = (uint)ELFSectionHeader.ELFSectionType.SectionTypeStringTable;
            //SectionHeaders[6].NameTableOffset = 0x11;
            SectionHeaders[6].NameTableOffset = (uint)GetOffsetOfSectionName(sectionNames, ".shstrtab");
            SectionHeaders[6].Flags = 0x00000000;
            SectionHeaders[6].Offset = reginfoOffset + (uint)regInfoData.Length;
            SectionHeaders[6].FileSize = (uint)sectionNameData.Length;
            SectionHeaders[6].Alignment = 1;
            SectionHeaders[6].EntrySize = 0;

            int maxHeader = 6;
            if (additionalSegments.Length > 0)
            {
                foreach (MemorySegment seg in additionalSegments)
                {
                    maxHeader++;
                    int programHeaderNum = maxHeader - 6;
                    SectionHeaders[maxHeader] = new ELFSectionHeader();
                    SectionDataArray[maxHeader] = new ELFSection();
                    SectionDataArray[maxHeader].SectionData = new byte[0];
                    SectionHeaders[maxHeader].SectionType = (uint)ELFSectionHeader.ELFSectionType.SectionTypeProgramData;
                    SectionHeaders[maxHeader].VirtualAddress = seg.Offset;
                    SectionHeaders[maxHeader].NameTableOffset = (uint)GetOffsetOfSectionName(sectionNames, seg.Name);
                    SectionHeaders[maxHeader].Flags = DefaultPlayStationSectionAttributes;
                    SectionHeaders[maxHeader].Offset = ProgramHeaders[programHeaderNum].Offset;
                    SectionHeaders[maxHeader].FileSize = ProgramHeaders[programHeaderNum].FileSize;
                    SectionHeaders[maxHeader].Alignment = 4;
                }
            }

            // +4 is a hack, figure that out later
            //ProgramSectionHeaderTableOffset = SectionHeaders[maxHeader].Offset + SectionHeaders[maxHeader].FileSize + 4;
            ProgramSectionHeaderTableOffset = SectionHeaders[6].Offset + SectionHeaders[6].FileSize + 4;
        }

        public override void WriteFile(BinaryWriter outputWriter)
        {
            outputWriter.Write(Magic);
            outputWriter.Write(WordSize);
            outputWriter.Write(Endianness);
            outputWriter.Write(ELFVersion);
            outputWriter.Write(OSABI);
            outputWriter.Write(Unused1);
            outputWriter.Write(Unused2);
            outputWriter.Write(ELFType);
            outputWriter.Write(InstructionSet);
            outputWriter.Write(ELFVersion2);
            outputWriter.Write(ProgramEntryAddress);
            outputWriter.Write(ProgramHeaderTableOffset);
            outputWriter.Write(ProgramSectionHeaderTableOffset);
            outputWriter.Write(Flags);
            outputWriter.Write(HeaderSize);
            outputWriter.Write(ProgramEntrySize);
            outputWriter.Write(ProgramEntryCount);
            outputWriter.Write(SectionEntrySize);
            outputWriter.Write(SectionEntryCount);
            outputWriter.Write(SectionNamesIndex);
            while (outputWriter.BaseStream.Position < ProgramHeaderTableOffset)
            {
                outputWriter.Write((byte)0x00);
            }
            // program headers
            for (int i = 0; i < ProgramHeaders.Length; i++)
            {
                outputWriter.Write(ProgramHeaders[i].SegmentType);
                outputWriter.Write(ProgramHeaders[i].Offset);
                outputWriter.Write(ProgramHeaders[i].VirtualAddress);
                outputWriter.Write(ProgramHeaders[i].PhysicalAddress);
                outputWriter.Write(ProgramHeaders[i].FileSize);
                outputWriter.Write(ProgramHeaders[i].MemSize);
                outputWriter.Write(ProgramHeaders[i].Flags);
                outputWriter.Write(ProgramHeaders[i].Alignment);
            }
            // program data
            for (int i = 0; i < ProgramHeaders.Length; i++)
            {
                while (outputWriter.BaseStream.Position < ProgramHeaders[i].Offset)
                {
                    outputWriter.Write((byte)0x00);
                }
                outputWriter.Write(ProgramDataArray[i].ProgramData);
            }
            // section headers
            // fill zeroes where some of the section data will go later
            while (outputWriter.BaseStream.Position < ProgramSectionHeaderTableOffset)
            {
                outputWriter.Write((byte)0x00);
            }
            for (int i = 0; i < SectionHeaders.Length; i++)
            {
                outputWriter.Write(SectionHeaders[i].NameTableOffset);
                outputWriter.Write(SectionHeaders[i].SectionType);
                outputWriter.Write(SectionHeaders[i].Flags);
                outputWriter.Write(SectionHeaders[i].VirtualAddress);
                outputWriter.Write(SectionHeaders[i].Offset);
                outputWriter.Write(SectionHeaders[i].FileSize);
                outputWriter.Write(SectionHeaders[i].LinkIndex);
                outputWriter.Write(SectionHeaders[i].Info);
                outputWriter.Write(SectionHeaders[i].Alignment);
                outputWriter.Write(SectionHeaders[i].EntrySize);
            }

            // write any section data
            for (int i = 0; i < SectionHeaders.Length; i++)
            {
                if (SectionHeaders[i].FileSize > 0)
                {
                    if (SectionHeaders[i].Offset > outputWriter.BaseStream.Length)
                    {
                        outputWriter.BaseStream.Seek(outputWriter.BaseStream.Length, SeekOrigin.Begin);
                        int numBytes = (int)SectionHeaders[i].Offset - (int)outputWriter.BaseStream.Length;
                        for (int j = 0; j < numBytes; j++)
                        {
                            outputWriter.Write((byte)0x00);
                        }
                    }
                    outputWriter.BaseStream.Seek(SectionHeaders[i].Offset, SeekOrigin.Begin);
                    outputWriter.Write(SectionDataArray[i].SectionData);
                }
            }

        }

        public override string GetFileInfoString()
        {
            StringBuilder builder = new StringBuilder(base.GetFileInfoString());
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Magic: 0x{0:X8}", Magic));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Word Size: 0x{0:X2}", WordSize));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Endianness: 0x{0:X2}", Endianness));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("ELF Version: 0x{0:X2}", ELFVersion));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("OSABI: 0x{0:X2}", OSABI));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Unused 1: 0x{0:X8}", Unused1));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Unused 2: 0x{0:X8}", Unused2));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("ELF Type: 0x{0:X4}", ELFType));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Instruction Set: 0x{0:X4}", InstructionSet));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("ELFVersion 2: 0x{0:X8}", ELFVersion2));
            builder.Append(Environment.NewLine);

            builder.Append(String.Format("Program Entry Address: 0x{0:X8}", ProgramEntryAddress));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Program Header Table Offset: 0x{0:X8}", ProgramHeaderTableOffset));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Program Section Header Table Offset: 0x{0:X8}", ProgramSectionHeaderTableOffset));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Flags: 0x{0:X8}", Flags));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Header Size: 0x{0:X4}", HeaderSize));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Program Entry Size: 0x{0:X4}", ProgramEntrySize));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Program Entry Count: 0x{0:X4}", ProgramEntryCount));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Section Entry Size: 0x{0:X4}", SectionEntrySize));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Section Entry Count: 0x{0:X4}", SectionEntryCount));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Section Names Index: 0x{0:X4}", SectionNamesIndex));
            builder.Append(Environment.NewLine);

            for (int i = 0; i < ProgramEntryCount; i++)
            {
                builder.Append(String.Format("Program {0}", i));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tProgram Type: {0}", ProgramHeaders[i].GetSegmentTypeName()));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tProgram Offset in ELF: 0x{0:X8}", ProgramHeaders[i].Offset));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tProgram Virtual Address: 0x{0:X8}", ProgramHeaders[i].VirtualAddress));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tProgram Physical Address: 0x{0:X8}", ProgramHeaders[i].PhysicalAddress));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tProgram File Size (in ELF): 0x{0:X8}", ProgramHeaders[i].FileSize));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tProgram Memory Size: 0x{0:X8}", ProgramHeaders[i].MemSize));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tFlags: 0x{0:X8}", ProgramHeaders[i].Flags));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tAlignment: 0x{0:X8}", ProgramHeaders[i].Alignment));
                builder.Append(Environment.NewLine);
            }

            for (int i = 0; i < SectionEntryCount; i++)
            {
                builder.Append(String.Format("Section {0}", i));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tSection Type: {0}", SectionHeaders[i].GetSectionTypeName()));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tName Table Offset: 0x{0:X8}", SectionHeaders[i].NameTableOffset));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tName: '{0}'", GetSectionNameByOffset((int)SectionHeaders[i].NameTableOffset)));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tFlags: 0x{0:X8}", SectionHeaders[i].Flags));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tOffset: 0x{0:X8}", SectionHeaders[i].Offset));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tVirtual Address: 0x{0:X8}", SectionHeaders[i].VirtualAddress));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tFile Size: 0x{0:X8}", SectionHeaders[i].FileSize));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tLink Index: 0x{0:X8}", SectionHeaders[i].LinkIndex));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tInformation: 0x{0:X8}", SectionHeaders[i].Info));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tAlignment: 0x{0:X8}", SectionHeaders[i].Alignment));
                builder.Append(Environment.NewLine);
                builder.Append(String.Format("\tEntry Size: 0x{0:X8}", SectionHeaders[i].EntrySize));
                builder.Append(Environment.NewLine);
            }

            return builder.ToString();
        }

        protected string GetSectionNameByOffset(int offset)
        {
            if (SectionNamesIndex >= SectionDataArray.Length)
            {
                return "[Invalid Section Index]";
            }
            if (offset >= SectionDataArray[SectionNamesIndex].SectionData.Length)
            {
                return "[Invalid Offset in Section Name Table]";
            }
            StringBuilder snBuilder = new StringBuilder();
            for (int i = offset; i < SectionDataArray[SectionNamesIndex].SectionData.Length; i++)
            {
                if (SectionDataArray[SectionNamesIndex].SectionData[i] == 0x00)
                {
                    break;
                }
                snBuilder.Append((char)SectionDataArray[SectionNamesIndex].SectionData[i]);
            }
            return snBuilder.ToString();
        }
    }
}
