﻿// PlayStation ELF Converter
// Converts between ELF and PSX-EXE executable formats
// Copyright 2019 Ben Lincoln
// https://www.beneaththewaves.net/
// Based in part on information documented in LameGuy64's elf2x utility
// (http://www.psxdev.net/forum/viewtopic.php?f=75&t=1613)

// This file is part of PlayStation ELF Converter.

// PlayStation ELF Converter is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// PlayStation ELF Converter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with PlayStation ELF Converter (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace BenLincoln.ExecutableFiles
{
    public struct PSXEXEC
    {
        public uint ProgramCounter;
        public uint GlobalPointer;             // only populated in SCE-format EXEs
        public uint TextSegmentAddress;
        public uint TextSegmentSize;
        public uint DataSegmentAddress;        // only populated in SCE-format EXEs
        public uint DataSegmentSize;           // only populated in SCE-format EXEs
        public uint BSSSegmentAddress;         // only populated in SCE-format EXEs
        public uint BSSSegmentSize;            // only populated in SCE-format EXEs
        public uint StackSegmentAddress;       // seems to always be set to 0x801FFFF0, even though it actually starts at 0x801F8000
        public uint StackSegmentSize;          // seems to always be 0, even though it's actually 16KiB
        public uint SavedStackPointer;
        public uint SavedFramePointer;
        public uint SavedGlobalPointer;
        public uint SavedReturnAddress;
        public uint SavedBaseAddress;
    }

    public enum LicenseType
    {
        Japan,
        NorthAmerica,
        Europe,
        Custom
    }

    public class PSXEXE : BinaryFile
    {
        protected const int MinSize = 4096;

        public enum PSXEXEType
        {
            Unknown,
            PSXEXE,
            SCE,
            CPE
        }

        public byte[] Header;               // bytes 0-7
        public uint TextSegmentOffset;      // bytes 8-11, only populated in SCE-format EXEs
        public uint DataSegmentOffset;      // bytes 12-15, only populated in SCE-format EXEs
        public PSXEXEC PSXExec;             // 60 bytes - 16-75
        public byte[] License;              // 64 bytes - 76 - 139
        public byte[] Padding2;             // 1908 bytes - 140-2047
        public byte[] ProgramCode;          // bytes 2048 - end of file in a typical PSX-EXE file

        public PSXEXEType EXEType;

        public string CustomLicenseString;

        protected const string _LicenseStringJapan = "Sony Computer Entertainment Inc. for Japan area";
        protected const string _LicenseStringNorthAmerica = "Sony Computer Entertainment Inc. for North America area";
        protected const string _LicenseStringEurope = "Sony Computer Entertainment Inc. for Europe area";

        public PSXEXE()
        {
            _FileTypeName = "PSX-EXE";
        }

        public override void ReadFile(BinaryReader binaryReader)
        {
            if (binaryReader.BaseStream.Length < MinSize)
            {
                throw new Exception(String.Format("This file is {0} bytes long, but at least {1} bytes are required.", binaryReader.BaseStream.Length, MinSize));
            }
            EXEType = PSXEXEType.Unknown;
            binaryReader.BaseStream.Seek(0, SeekOrigin.Begin);
            Header = binaryReader.ReadBytes(8);
            string headerAsASCIIString = ByteArrayToASCIIString(Header);
            if (headerAsASCIIString.Length > 6)
            {
                if (headerAsASCIIString.Substring(0, 7) == "SCE EXE")
                {
                    EXEType = PSXEXEType.SCE;
                }
            }
            if (headerAsASCIIString.Length > 7)
            {
                if (headerAsASCIIString.Substring(0, 8) == "PS-X EXE")
                {
                    EXEType = PSXEXEType.PSXEXE;
                }
            }
            if (EXEType == PSXEXEType.Unknown)
            {
                throw new Exception(String.Format("Unrecognized file header: 0x{0} ('{1}')", ByteArrayToHexString(Header), ByteArrayToASCIIString(Header)));
            }
            TextSegmentOffset = binaryReader.ReadUInt32();
            DataSegmentOffset = binaryReader.ReadUInt32();

            PSXExec = new PSXEXEC();
            PSXExec.ProgramCounter = binaryReader.ReadUInt32();
            PSXExec.GlobalPointer = binaryReader.ReadUInt32();
            PSXExec.TextSegmentAddress = binaryReader.ReadUInt32();
            PSXExec.TextSegmentSize = binaryReader.ReadUInt32();
            PSXExec.DataSegmentAddress = binaryReader.ReadUInt32();
            PSXExec.DataSegmentSize = binaryReader.ReadUInt32();
            PSXExec.BSSSegmentAddress = binaryReader.ReadUInt32();
            PSXExec.BSSSegmentSize = binaryReader.ReadUInt32();
            PSXExec.StackSegmentAddress = binaryReader.ReadUInt32();
            PSXExec.StackSegmentSize = binaryReader.ReadUInt32();
            PSXExec.SavedStackPointer = binaryReader.ReadUInt32();
            PSXExec.SavedFramePointer = binaryReader.ReadUInt32();
            PSXExec.SavedGlobalPointer = binaryReader.ReadUInt32();
            PSXExec.SavedReturnAddress = binaryReader.ReadUInt32();
            PSXExec.SavedBaseAddress = binaryReader.ReadUInt32();

            License = binaryReader.ReadBytes(64);
            Padding2 = binaryReader.ReadBytes(1908);

            // this will need to be handled differently for files that explicitly set the text/data/bss segment sizes
            int remainingBytes = (int)(binaryReader.BaseStream.Length - 2048);

            ProgramCode = binaryReader.ReadBytes(remainingBytes);
        }

        public override void WriteFile(BinaryWriter outputWriter)
        {
        }

        public void AddOverlay(MemorySegment newOverlay)
        {
            long topOffset = newOverlay.Offset + newOverlay.Length;
            if ((topOffset > PSXExec.TextSegmentAddress + PSXExec.TextSegmentSize))
            {
                Console.WriteLine("Warning: the specified overlay is larger than the text segment for this binary, and the overlay will be truncated as a result.");
            }
            byte[] newProgramCode = new byte[PSXExec.TextSegmentSize];
            for (int i = 0; i < newProgramCode.Length; i++)
            {
                newProgramCode[i] = 0;
            }
            Array.Copy(ProgramCode, 0, newProgramCode, 0, ProgramCode.Length);
            long bytesToCopy = newOverlay.Length;
            if (PSXExec.TextSegmentAddress < newOverlay.Offset)
            {
                long baseOffset = newOverlay.Offset - PSXExec.TextSegmentAddress;
                if (baseOffset < 0)
                {
                    Console.WriteLine(string.Format("Calculated a base offset of {0}, which is nonsensical. Please contact a developer for assistance.", baseOffset));
                }
                if ((baseOffset + bytesToCopy) > newProgramCode.LongLength)
                {
                    bytesToCopy = newProgramCode.LongLength - baseOffset;
                }
                Array.Copy(newOverlay.Data, 0, newProgramCode, baseOffset, bytesToCopy);
                ProgramCode = newProgramCode;
            }
            else
            {
                Console.WriteLine(string.Format("Calculated a base offset of less than zero, which is nonsensical. Please contact a developer for assistance."));
            }
        }

        public override string GetFileInfoString()
        {
            StringBuilder builder = new StringBuilder(base.GetFileInfoString());
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Header: '{0}' (0x{1})", ByteArrayToASCIIString(Header), ByteArrayToHexString(Header)));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Text Segment Offset: 0x{0:X8}", TextSegmentOffset));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("ProgramCounter: 0x{0:X8}", PSXExec.ProgramCounter));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Global Pointer: 0x{0:X8}", PSXExec.GlobalPointer));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Text Segment Address: 0x{0:X8}", PSXExec.TextSegmentAddress));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("TextSegmentSize: 0x{0:X8}", PSXExec.TextSegmentSize));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Data Segment Address: 0x{0:X8}", PSXExec.DataSegmentAddress));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Data Segment Size: 0x{0:X8}", PSXExec.DataSegmentSize));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("BSS Segment Address: 0x{0:X8}", PSXExec.BSSSegmentAddress));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("BSS Segment Size: 0x{0:X8}", PSXExec.BSSSegmentSize));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Stack Segment Address: 0x{0:X8}", PSXExec.StackSegmentAddress));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Stack Segment Size: 0x{0:X8}", PSXExec.StackSegmentSize));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Saved Stack Pointer: 0x{0:X8}", PSXExec.SavedStackPointer));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Saved Frame Pointer: 0x{0:X8}", PSXExec.SavedFramePointer));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Saved Global Pointer: 0x{0:X8}", PSXExec.SavedGlobalPointer));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Saved Return Address: 0x{0:X8}", PSXExec.SavedReturnAddress));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("Saved Base Address: 0x{0:X8}", PSXExec.SavedBaseAddress));
            builder.Append(Environment.NewLine);
            builder.Append(String.Format("License String: '{0}' (0x{1})", ByteArrayToASCIIString(License), ByteArrayToHexString(License)));
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }

        public static string ByteArrayToASCIIString(byte[] bytes)
        {
            return System.Text.Encoding.ASCII.GetString(bytes, 0, bytes.Length);
        }

        // Begin: https://stackoverflow.com/questions/311165/how-do-you-convert-a-byte-array-to-a-hexadecimal-string-and-vice-versa
        public static string ByteArrayToHexString(byte[] bytes)
        {
            return BitConverter.ToString(bytes).Replace("-", "");
        }
        // End: https://stackoverflow.com/questions/311165/how-do-you-convert-a-byte-array-to-a-hexadecimal-string-and-vice-versa
    }
}
