﻿// this entire project/solution is a heavily-customized fork of https://github.com/stohrendorf/symdump/

// SymDumpTE
// Exports debug symbol data from PsyQ .SYM files.
// Original code by Steffen Ohrendorf (https://github.com/stohrendorf/symdump/)
// Customizations by Ben Lincoln, 2019
// https://www.beneaththewaves.net/

// This file is part of SymDumpTE.

// SymDumpTE is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// SymDumpTE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with SymDumpTE (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

using symdump.exefile;
using symdump.exefile.util;
using symdump.symfile;

using BenLincoln.DebugSymbols;

namespace SymDumpTE
{
    class Program
    {
        protected const string ProgramName = "SymDumpTE";
        protected static string Copyright = "Original SymDump by Steffen Ohrendorf, extended and customized into SymDumpTE by Ben Lincoln, 2019";
        protected const string InfoURL = "Original: https://github.com/stohrendorf/symdump/, Customized: https://www.beneaththewaves.net/Software/This_Dust_Remembers_What_It_Once_Was.html";
        protected static string[] ThanksList = new string[] {
            "Steffen Ohrendorf for SymDump",
            "Vladimir Kononovich (https://github.com/lab313ru) for https://github.com/lab313ru/dumpsym_src",
            "Gh0stBlade (https://github.com/Gh0stBlade)",
            "Andrew Fradley (https://github.com/afradley)"
            };
        protected const string HelpMessageFile = "SymDumpTE.HelpMessage.txt";

        static void Main(string[] args)
        {
            if ((args.Length < 1))
            {
                BenLincoln.DebugSymbols.Utilities.DisplayHelpMessage(Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList, HelpMessageFile);
                Environment.Exit(1);
            }

            string inputFilePath = "";
            string outputFilePath = "";
            bool jsonOutput = false;
            bool disassemble = false;
            SymDumpProcessingConfiguration config = new SymDumpProcessingConfiguration();

            ArrayList arguments = new ArrayList();
            for (int i = 0; i < args.Length; i++)
            {
                // if the argument is one with a second component, catch that now
                //string nextArg = "";
                switch (args[i])
                {
                    default:
                        arguments.Add(args[i]);
                        break;
                }
            }

            ArrayList positionalArgs = new ArrayList();

            foreach (string arg in arguments)
            {
                if ((arg == "-?") || (arg == "-h") || (arg == "--help"))
                {
                    BenLincoln.DebugSymbols.Utilities.DisplayHelpMessage(Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList, HelpMessageFile);
                    Environment.Exit(0);
                }
                bool handled = false;
                if (arg == "--json")
                {
                    jsonOutput = true;
                    handled = true;
                }
                if (arg == "--disassemble")
                {
                    disassemble = true;
                    handled = true;
                }
                if (arg == "--auto-rename-fakes")
                {
                    config.AutoRenameFakes = true;
                    handled = true;
                }
                if (arg == "--inline-fakes")
                {
                    config.InlineFakes = true;
                    handled = true;
                }
                if (arg == "--rename-for-compatibility")
                {
                    config.RenameForCompatibility = true;
                    handled = true;
                }
                if (arg == "--ignore-duplicate-definitions")
                {
                    config.IgnoreDuplicateDefinitions = true;
                    handled = true;
                }
                if (arg == "--debug")
                {
                    config.ShowDebugMessages = true;
                    handled = true;
                }
                if (!handled)
                {
                    positionalArgs.Add(arg);
                }
            }

            if (positionalArgs.Count > 0)
            {
                inputFilePath = (string)positionalArgs[0];
            }

            if (positionalArgs.Count > 1)
            {
                outputFilePath = (string)positionalArgs[1];
            }

            if (inputFilePath == "")
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("No input symbol file path was specified.", 1, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
            }

            if (!File.Exists(inputFilePath))
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("The specified input symbol file does not exist.", 1, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
            }

            if (config.InlineFakes && (!config.AutoRenameFakes))
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("The --inline-fakes option requires that --auto-rename-fakes also be specified.", 1, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
            }

            SymFile symFile;

            var exeFilename = Path.ChangeExtension(inputFilePath, "EXE");

            using (var fs = new FileStream(inputFilePath, FileMode.Open))
            {
                symFile = new SymFile(new BinaryReader(fs), config);
                string defaultSourceFileBaseName = Path.GetFileNameWithoutExtension(inputFilePath);

                if (File.Exists(exeFilename))
                {
                    try
                    {
                        symFile.GetDataFromPSXEXE(exeFilename);
                    }
                    catch (Exception ex)
                    {
                        BenLincoln.DebugSymbols.Utilities.ExitWithError(String.Format("Could not get data from PSX EXE '{0}': {1}", exeFilename, ex.Message), 9, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
                    }
                }

                if (jsonOutput)
                {
                    string result = symFile.DumpJSON(defaultSourceFileBaseName, config);
                    if (outputFilePath == "")
                    {
                        Console.WriteLine(result);
                    }
                    else
                    {
                        try
                        {
                            BenLincoln.DebugSymbols.Utilities.WriteToOutput(outputFilePath, result, FileMode.Create);
                        }
                        catch (Exception ex)
                        {
                            BenLincoln.DebugSymbols.Utilities.ExitWithError(String.Format("Could not write to the output file '{0}': {1}", outputFilePath, ex.Message), 7, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);
                        }
                    }
                }
                else
                {
                    symFile.Dump(System.Console.Out);
                }
            }

            if (disassemble)
            {
                if (!File.Exists(exeFilename))
                {
                    System.Console.WriteLine(String.Format("// could not find file '{0}' to disassemble", exeFilename));
                    return;
                }

                System.Console.WriteLine();
                System.Console.WriteLine("// disassembly");

                using (var fs = new EndianBinaryReader(new FileStream(exeFilename, FileMode.Open)))
                {
                    var exeFile = new ExeFile(fs, symFile);
                    exeFile.Disassemble();
                    exeFile.Dump();
                }
            }
            Console.WriteLine();
        }
    }
}
