﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using symdump.symfile.util;
using symdump.util;

namespace symdump.symfile
{
    public class EnumDef : IEquatable<EnumDef>
    {
        private readonly Dictionary<string, int> _members = new Dictionary<string, int>();
        //private readonly string _name;
        private string _name;
        private readonly uint _size;
        public readonly int SYMFileSequenceNumber;
        public string SLDSourceFileName;
        public List<string> ObjectFileNames;

        public string Name
        {
            get
            {
                return _name;
            }
            set
            {
                _name = value;
            }
        }

        public EnumDef(BinaryReader stream, string name, SymDumpProcessingConfiguration config)
        {
            _name = name;
            if (config.RenameForCompatibility)
            {
                _name = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(_name, false);
            }
            SYMFileSequenceNumber = config.Issuer.GetNextNumber();
            while (true)
            {
                var typedValue = new TypedValue(stream, config);
                if (typedValue.Type == (0x80 | 20))
                {
                    var ti = stream.ReadTypeInfo(false, config);
                    var memberName = stream.ReadPascalString();

                    if (ti.ClassType == ClassType.EndOfStruct)
                        break;

                    if (ti.ClassType != ClassType.EnumMember)
                        throw new Exception("Unexpected class");

                    _members.Add(memberName, typedValue.Value);
                }
                else if (typedValue.Type == (0x80 | 22))
                {
                    var ti = stream.ReadTypeInfo(true, config);
                    if (ti.TypeDef.BaseType != BaseType.Null)
                        throw new Exception($"Expected baseType={BaseType.Null}, but it's {ti.TypeDef.BaseType}");

                    if (ti.Dims.Length != 0)
                        throw new Exception($"Expected dims=0, but it's {ti.Dims.Length}");

                    if (ti.Tag != _name)
                    {
                        if ((BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(ti.Tag, true) != _name) && (!ti.Tag.EndsWith(_name)))
                        {
                            throw new Exception($"Expected name={_name}, but it's {ti.Tag}");
                        }
                    }

                    var tag = stream.ReadPascalString();
                    if (tag != ".eos")
                        throw new Exception($"Expected tag=.eos, but it's {tag}");

                    if (ti.ClassType != ClassType.EndOfStruct)
                        throw new Exception($"Expected classType={ClassType.EndOfStruct}, but it's {ti.ClassType}");

                    _size = ti.Size;
                    break;
                }
                else
                {
                    throw new Exception("Unexpected entry");
                }
            }
            ObjectFileNames = new List<string>();
        }

        public bool Equals(EnumDef other)
        {
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;
            return _members.SequenceEqual(other._members) && string.Equals(_name, other._name);
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != GetType()) return false;
            return Equals((EnumDef) obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                return (_members.GetHashCode() * 397) ^ _name.GetHashCode();
            }
        }

        public string GetCType()
        {
            string cType;
            switch (_size)
            {
                case 1:
                    cType = "char";
                    break;
                case 2:
                    cType = "short";
                    break;
                case 4:
                    cType = "int";
                    break;
                default:
                    cType = ("unknown /* $Cannot determine primitive type for size {size} */");
                    break;
            }
            return cType;
        }

        public void Dump(IndentedTextWriter writer)
        {
            string cType = GetCType();

            writer.WriteLine($"enum {_name} : {cType} {{");
            ++writer.Indent;
            foreach (var kvp in _members)
                writer.WriteLine($"{kvp.Key} = {kvp.Value},");
            --writer.Indent;
            writer.WriteLine("}; // " + BenLincoln.DebugSymbols.Utilities.GetListAsString(ObjectFileNames));
        }

        public BenLincoln.DebugSymbols.SymDump.CEnum ToBLDSEnumDef(string sourceFile)
        {
            string cType = GetCType();

            BenLincoln.DebugSymbols.SymDump.CEnum result = new BenLincoln.DebugSymbols.SymDump.CEnum();
            result.hashcode = GetHashCode();
            result.primitive_type = cType;
            result.source_file = sourceFile;
            result.hashcode = GetHashCode();
            result.name = _name;
            result.members = new Dictionary<string, int>();
            result.sym_file_sequence_number = SYMFileSequenceNumber;
            result.source_file = SLDSourceFileName;
            result.object_files = ObjectFileNames;
            foreach (string key in _members.Keys)
            {
                result.members.Add(key, _members[key]);
            }

            return result;
        }

        public string DumpJSON(string sourceFile, string lineIndent)
        {
            BenLincoln.DebugSymbols.SymDump.CEnum resultDef = ToBLDSEnumDef(sourceFile);

            return BenLincoln.DebugSymbols.Utilities.IndentString(resultDef.ToJSON(), lineIndent);

            //StringBuilder builder = new StringBuilder();

            //builder.Append(lineIndent);
            //builder.Append("{");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"name\": \"{0}\",", _name));
            //builder.Append(Environment.NewLine);
            //if (sourceFile != "")
            //{
            //    builder.Append(lineIndent);
            //    builder.Append(String.Format("\t\"source_file\": \"{0}\",", sourceFile));
            //    builder.Append(Environment.NewLine);
            //}
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"primitive_type\": \"{0}\",", cType));
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("\t\"members\": ");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("\t{");
            //builder.Append(Environment.NewLine);
            //int elemNum = 1;
            //foreach (var kvp in _members)
            //{
            //    builder.Append(lineIndent);
            //    builder.Append(String.Format("\t\t\"{0}\": \"{1}\"", kvp.Key, kvp.Value));
            //    if (elemNum < _members.Count)
            //    {
            //        builder.Append(",");
            //    }
            //    elemNum++;
            //    builder.Append(Environment.NewLine);
            //}
            //builder.Append(lineIndent);
            //builder.Append("\t}");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("}");

            //return builder.ToString();
        }
    }
}
