using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using symdump.symfile.util;
using symdump.util;

namespace symdump.symfile
{
    public class Function
    {
        public readonly uint Address;
        private readonly List<Block> _blocks = new List<Block>();
        private readonly string _file;
        private readonly uint _lastLine;
        private readonly uint _line;
        private readonly uint _mask;
        private readonly int _maskOffs;
        private string _name;
        public readonly int SYMFileSequenceNumber;
        public string SLDSourceFileName;
        public List<string> ObjectFileNames;

        //private readonly List<string> _parameters = new List<string>();
        private List<BenLincoln.DebugSymbols.SymDump.CParameter> _parameters = new List<BenLincoln.DebugSymbols.SymDump.CParameter>();
        private readonly Register _register;
        private readonly string _returnType;
        private readonly Register _stackBase;
        private readonly uint _stackFrameSize;

        public string Name
        {
            get
            {
                return _name;
            }
            set
            {
                _name = value;
            }
        }

        public Function(BinaryReader reader, uint ofs, IReadOnlyDictionary<string, string> funcTypes, SymDumpProcessingConfiguration config)
        {
            Address = ofs;

            _stackBase = (Register) reader.ReadUInt16();
            _stackFrameSize = reader.ReadUInt32();
            _register = (Register) reader.ReadUInt16();
            _mask = reader.ReadUInt32();
            _maskOffs = reader.ReadInt32();

            _line = reader.ReadUInt32();
            _file = reader.ReadPascalString();
            _name = reader.ReadPascalString();
            SYMFileSequenceNumber = config.Issuer.GetNextNumber();
            ObjectFileNames = new List<string>();

            if (!funcTypes.TryGetValue(_name, out _returnType))
            {
                _returnType = "__UNKNOWN__";
            }

            while (true)
            {
                var typedValue = new TypedValue(reader, config);

                if (reader.SkipSld(typedValue))
                {
                    Console.WriteLine(string.Format("Warning: intentionally-ignored type {0}", typedValue.Type & 0x7f));
                    continue;
                }

                TypeInfo ti;
                string memberName;
                switch (typedValue.Type & 0x7f)
                {
                    case 14: // end of function
                        _lastLine = reader.ReadUInt32();
                        return;
                    case 16: // begin of block
                        _blocks.Add(new Block(reader, (uint) typedValue.Value, reader.ReadUInt32(), this, config));
                        continue;
                    case 20:
                        ti = reader.ReadTypeInfo(false, config);
                        memberName = reader.ReadPascalString();
                        break;
                    case 22:
                        ti = reader.ReadTypeInfo(true, config);
                        memberName = reader.ReadPascalString();
                        break;
                    default:
                        Console.WriteLine(string.Format("Warning: unhandled typed value {0} - will be ignored", typedValue.Type & 0x7f));
                        continue;
                        //throw new Exception("Nope");
                }

                if (ti == null)
                {
                    Console.WriteLine(string.Format("Warning: ti is null for typed value {0}", typedValue.Type & 0x7f));
                    break;
                }

                if (memberName == null)
                {
                    Console.WriteLine(string.Format("Warning: memberName is null for typed value {0}", typedValue.Type & 0x7f));
                    break;
                }

                if (ti.ClassType == ClassType.Argument)
                {
                    //_parameters.Add($"{ti.AsCode(memberName)} /*stack {typedValue.Value}*/");
                    BenLincoln.DebugSymbols.SymDump.CParameter p = new BenLincoln.DebugSymbols.SymDump.CParameter();
                    //p.name = ti.GetName(memberName);
                    //p.type_name = ti.GetCType();
                    //BenLincoln.DebugSymbols.SymDump.CParameter[] pParsed = BenLincoln.DebugSymbols.SymDump.CElement.GetCParametersFromString(ti.AsCode(memberName), true);
                    //p.type_name = pParsed[0].type_name;
                    p.type_name = ti.AsCode("");
                    //p.name = pParsed[0].name;
                    p.name = memberName;
                    //p.is_function_pointer = pParsed[0].is_function_pointer;
                    p.is_function_pointer = false;
                    if (ti.GetDerivedTypes(memberName).Contains("FunctionReturnType"))
                    {
                        p.is_function_pointer = true;
                    }
                    if (p.type_name.Contains("*)("))
                    {
                        p.is_function_pointer = true;
                    }
                    p.stack_offset = typedValue.Value;
                    p.parameter_signature = $"{ti.AsCode(memberName)} /*stack {typedValue.Value}*/";
                    //p.is_varargs = pParsed[0].is_varargs;
                    p.is_varargs = p.parameter_signature.Replace(" ", "").Contains("...)");
                    //p.hashcode = ti.GetHashCode();
                    p.derived_types = ti.GetDerivedTypes("").ToArray();
                    _parameters.Add(p);
                }
                else if (ti.ClassType == ClassType.RegParam)
                {
                    //_parameters.Add($"{ti.AsCode(memberName)} /*${(Register)typedValue.Value}*/");
                    BenLincoln.DebugSymbols.SymDump.CParameter p = new BenLincoln.DebugSymbols.SymDump.CParameter();
                    //p.name = ti.GetName(memberName);
                    //p.type_name = ti.GetCType();
                    //BenLincoln.DebugSymbols.SymDump.CParameter[] pParsed = BenLincoln.DebugSymbols.SymDump.CElement.GetCParametersFromString(ti.AsCode(memberName), true);
                    //p.type_name = pParsed[0].type_name;
                    p.type_name = ti.AsCode("");
                    //p.name = pParsed[0].name;
                    p.name = memberName;
                    p.register_name = ((Register)typedValue.Value).ToString();
                    //p.is_function_pointer = pParsed[0].is_function_pointer;
                    p.is_function_pointer = false;
                    if (ti.GetDerivedTypes(memberName).Contains("FunctionReturnType"))
                    {
                        p.is_function_pointer = true;
                    }
                    if (p.type_name.Contains("*)("))
                    {
                        p.is_function_pointer = true;
                    }
                    //p.is_varargs = pParsed[0].is_varargs;
                    p.parameter_signature = $"{ti.AsCode(memberName)} /*${(Register)typedValue.Value}*/";
                    p.is_varargs = p.parameter_signature.Replace(" ", "").Contains("...)");
                    //p.hashcode = ti.GetHashCode();
                    p.derived_types = ti.GetDerivedTypes("").ToArray();
                    _parameters.Add(p);
                }
                //else
                //{
                //    Console.WriteLine("Debug");
                //}
            }
            ObjectFileNames = new List<string>();
        }

        private IEnumerable<Register> SavedRegisters => Enumerable.Range(0, 32)
            .Where(i => ((1 << i) & _mask) != 0)
            .Select(i => (Register) i);

        public void Dump(IndentedTextWriter writer)
        {
            writer.WriteLine("/*");
            writer.WriteLine($" * Offset 0x{Address:X}");
            writer.WriteLine($" * {_file} (line {_line})");
            if (ObjectFileNames.Count > 0)
            {
                writer.WriteLine(string.Format(" * Object files: {0}", BenLincoln.DebugSymbols.Utilities.GetListAsString(ObjectFileNames)));
            }
            writer.WriteLine($" * Stack frame base ${_stackBase}, size {_stackFrameSize}");
            if (_mask != 0)
            {
                writer.WriteLine($" * Saved registers at offset {_maskOffs}: {string.Join(" ", SavedRegisters)}");
            }
            writer.WriteLine(" */");

            writer.WriteLine(GetSignature());

            _blocks.ForEach(b => b.Dump(writer));

            if (_blocks.Count != 0)
                return;

            writer.WriteLine("{");
            writer.WriteLine("}");
        }

        public BenLincoln.DebugSymbols.SymDump.CFunction ToBLDSCFunction(string defaultSourceFile)
        {
            BenLincoln.DebugSymbols.SymDump.CFunction result = new BenLincoln.DebugSymbols.SymDump.CFunction();
            result.name = _name;
            string sourceFile = _file;
            if (SLDSourceFileName != sourceFile)
            {
                if (((sourceFile == null) || (sourceFile == "")) && ((SLDSourceFileName != null) && (SLDSourceFileName != "")))
                {
                    Console.WriteLine(string.Format("Info: Using SLD source file '{0}' instead of empty sourceFile value '{1}' for function '{2}'", SLDSourceFileName, sourceFile, result.name));
                    sourceFile = SLDSourceFileName;
                }
                else
                {
                    // Console.WriteLine(string.Format("Debug: SLD source file '{0}' != source file '{1}' for function '{2}'", SLDSourceFileName, sourceFile, result.name));
                }
            }
            if (sourceFile.Trim() == "")
            {
                sourceFile = defaultSourceFile;
            }
            result.source_file = sourceFile;
            result.source_file_line_number = _line;
            result.offset_in_memory = Address;
            result.hashcode = GetHashCode();
            result.stack_frame_base = _stackBase.ToString();
            result.stack_frame_size = _stackFrameSize;
            result.mask = _mask;
            result.mask_offset = _maskOffs;
            result.saved_registers = new List<string>();
            result.sym_file_sequence_number = SYMFileSequenceNumber;
            result.object_files = ObjectFileNames;
            foreach (Register sr in SavedRegisters)
            {
                result.saved_registers.Add(sr.ToString());
            }
            result.method_signature = GetSignature();
            result.return_value = new BenLincoln.DebugSymbols.SymDump.CParameter();
            result.return_value.type_name = _returnType;
            result.return_value.parameter_signature = string.Format("{0} {1}", _returnType, _name);
            bool rvIsFunctionPointer = result.return_value.is_function_pointer;
            if (result.return_value.type_name.Contains("*)("))
            {
                rvIsFunctionPointer = true;
            }
            if ((result.return_value.derived_types != null) && (result.return_value.derived_types.Contains("FunctionReturnType")))
            {
                rvIsFunctionPointer = true;
            }
            if (rvIsFunctionPointer)
            {
                result.return_value.is_function_pointer = true;
                //// these need to be assigned a name because Ghidra can't really refer to them otherwise, at least not in any way I know of
                //result.return_value.type_name = result.return_value.type_name.Substring(0, result.return_value.type_name.Length - 6);
                //if (result.return_value.name == "")
                //{
                //    result.return_value.name = string.Format("unnamed_{0}_function_pointer", result.return_value.type_name);
                //}
            }
            result.return_value.register_name = _register.ToString();
            result.return_value.name = _name;
            for (int i = 0; i < _parameters.Count; i++)
            {
                bool pIsFunctionPointer = _parameters[i].is_function_pointer;
                if (_parameters[i].type_name.Contains("*)("))
                {
                    pIsFunctionPointer = true;
                }
                if ((_parameters[i].derived_types != null) && (_parameters[i].derived_types.Contains("FunctionReturnType")))
                {
                    pIsFunctionPointer = true;
                }
                if (pIsFunctionPointer)
                {
                    _parameters[i].is_function_pointer = true;
                }
            }
            result.parameters = _parameters;

            result.blocks = new List<BenLincoln.DebugSymbols.SymDump.CBlock>();
            foreach (Block b in _blocks)
            {
                result.blocks.Add(b.ToBLDSCBlock(sourceFile));
            }
            return result;
        }

        public string DumpJSON(string defaultSourceFile, string lineIndent)
        {
            BenLincoln.DebugSymbols.SymDump.CFunction resultDef = ToBLDSCFunction(defaultSourceFile);

            return BenLincoln.DebugSymbols.Utilities.IndentString(resultDef.ToJSON(), lineIndent);

        }

        public string GetSignature()
        {
            List<string> parameter_signatures = new List<string>();
            foreach (BenLincoln.DebugSymbols.SymDump.CParameter p in _parameters)
            {
                parameter_signatures.Add(p.parameter_signature);
            }
            //return $"{_returnType} /*${_register}*/ {_name}({string.Join(", ", _parameters)})";
            return $"{_returnType} /*${_register}*/ {_name}({string.Join(", ", parameter_signatures)})";
        }
    }
}
