﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using symdump.util;

namespace symdump.symfile
{
    public class StructDef : IEquatable<StructDef>
    {
        private List<StructMember> _members = new List<StructMember>();
        private string _name;
        public readonly int SYMFileSequenceNumber;
        public string SLDSourceFileName;
        public List<string> ObjectFileNames;
        public bool OriginalWasFake;
        public bool HasBeenInlined;
        protected SymDumpProcessingConfiguration ProcessingConfig;
        protected List<string> _TypeInfoTags;
        protected List<string> _TypeInfoObjectFiles;

        public string Name
        {
            get
            {
                return _name;
            }
            set
            {
                _name = value;
            }
        }

        public List<StructMember> Members
        {
            get
            {
                return _members;
            }
            set
            {
                _members = value;
            }
        }

        public StructDef(BinaryReader stream, string name, SymDumpProcessingConfiguration config)
        {
            _name = name;
            OriginalWasFake = IsFake;
            HasBeenInlined = false;
            ProcessingConfig = config;
            _TypeInfoTags = new List<string>();
            _TypeInfoObjectFiles = new List<string>();
            if (config.RenameForCompatibility)
            {
                //_name = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(_name, false);
                _name = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(_name, true);
            }
            SYMFileSequenceNumber = config.Issuer.GetNextNumber();
            while (true)
            {
                var typedValue = new TypedValue(stream, config);
                if (typedValue.Type == (0x80 | 20))
                {
                    var m = new StructMember(typedValue, stream, false, string.Format("struct {0}", this.Name), this.GetHashCode(), config);

                    if (m.TypeInfo.ClassType == ClassType.EndOfStruct)
                    {
                        break;
                    }

                    if ((m.TypeInfo.Tag != null) && (m.TypeInfo.Tag != "") && (!_TypeInfoTags.Contains(m.TypeInfo.Tag)))
                    {
                        _TypeInfoTags.Add(m.TypeInfo.Tag);
                    }

                    foreach (string objectFileName in m.TypeInfo.ObjectFileNames)
                    {
                        if (!_TypeInfoObjectFiles.Contains(objectFileName))
                        {
                            _TypeInfoObjectFiles.Add(objectFileName);
                        }
                    }

                    _members.Add(m);
                }
                else if (typedValue.Type == (0x80 | 22))
                {
                    var m = new StructMember(typedValue, stream, true, string.Format("struct {0}", this.Name), this.GetHashCode(), config);

                    if (m.TypeInfo.ClassType == ClassType.EndOfStruct)
                    {
                        break;
                    }

                    if ((m.TypeInfo.Tag != null) && (m.TypeInfo.Tag != "") && (!_TypeInfoTags.Contains(m.TypeInfo.Tag)))
                    {
                        _TypeInfoTags.Add(m.TypeInfo.Tag);
                    }

                    foreach (string objectFileName in m.TypeInfo.ObjectFileNames)
                    {
                        if (!_TypeInfoObjectFiles.Contains(objectFileName))
                        {
                            _TypeInfoObjectFiles.Add(objectFileName);
                        }
                    }

                    _members.Add(m);
                }
                else
                {
                    Console.WriteLine(string.Format("Unexpected struct type: {0} - data will be ignored", typedValue.Type));
                }
            }
            ObjectFileNames = new List<string>();
            foreach (string tiObjectFile in _TypeInfoObjectFiles)
            {
                ObjectFileNames.Add(tiObjectFile);
            }

            string unmodifiedName = Name;
            if (config.AutoRenameFakes && (IsFake))
            {
                string lastObjectFileName = "";
                if (config.Tracker.ObjectFileDataEntries.Count > 0)
                {
                    lastObjectFileName = config.Tracker.ObjectFileDataEntries[config.Tracker.ObjectFileDataEntries.Count - 1].FileName;
                }
                string lastSLDSourceFileName = "";
                if (config.Tracker.SLDEntries.Count > 0)
                {
                    lastSLDSourceFileName = config.Tracker.SLDEntries[config.Tracker.SLDEntries.Count - 1].SLDFile;
                }
                Name = BenLincoln.DebugSymbols.Utilities.GetNewFakeName("StructDef", Name, ObjectFileNames, config.Tracker.CurrentSLDFile, lastObjectFileName, lastSLDSourceFileName);
            }
            if (OriginalWasFake)
            {
                if (!config.FakeStructs.ContainsKey(Name))
                {
                    config.FakeStructs.Add(Name, this);
                }
                else
                {
                    if (this.Equals(config.FakeStructs[Name]))
                    {
                        SymFile.OutputDebugMessage(string.Format("the list of fake structs already contains '{0}', with an identical definition", Name), config);
                        SymFile.OutputDebugMessage(string.Format("Existing struct: {0}{1}", Environment.NewLine, config.FakeStructs[Name].GetAsString()), config);
                    }
                    else
                    {
                        SymFile.OutputErrorMessage(string.Format("the list of fake structs already contains '{0}', with a different definition", Name), config);
                        SymFile.OutputErrorMessage(string.Format("Existing struct: {0}{1}", Environment.NewLine, config.FakeStructs[Name].GetAsString()), config);
                        SymFile.OutputErrorMessage(string.Format("Conflicting struct: {0}{1}", Environment.NewLine, GetAsString()), config);
                    }
                }
            }
        }

        public bool IsFake => new Regex(@"^(\.|_)\d+fake$").IsMatch(_name);

        public bool Equals(StructDef other)
        {
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;
            //return _members.SequenceEqual(other._members) && string.Equals(_name, other._name) && string.Equals(ObjectFileName, other.ObjectFileName) && string.Equals(SLDSourceFileName, other.SLDSourceFileName);
            bool result = true;

            if (!_members.SequenceEqual(other._members))
            {
                SymFile.OutputDebugMessage(string.Format("Struct members don't match"), ProcessingConfig);
                result = false;
            }
            if ((!OriginalWasFake) || (!other.OriginalWasFake))
            {
                if (_name != other._name)
                {
                    SymFile.OutputDebugMessage(string.Format("StructMember objects don't have matching names ('{0}' vs '{1}')", _name, other._name), ProcessingConfig);
                    result = false;
                }
            }
            return result;
        }

        public void Dump(IndentedTextWriter writer)
        {
            if (!HasBeenInlined)
            {
                writer.WriteLine($"struct {_name} {{");
                ++writer.Indent;
                foreach (var m in _members)
                {
                    writer.WriteLine(m);
                }
                --writer.Indent;
                writer.WriteLine("}; // " + BenLincoln.DebugSymbols.Utilities.GetListAsString(ObjectFileNames));
            }
        }

        // this doesn't work very well in terms of indenting inlined structs/unions
        // I haven't fixed it because it's just for debugging at present
        public string GetAsString()
        {
            StringBuilder builder = new StringBuilder();
            builder.Append($"struct {_name} {{");
            builder.Append(Environment.NewLine);
            foreach (var m in _members)
            {
                builder.Append("\t");
                builder.Append(m);
                builder.Append(Environment.NewLine);
            }
            builder.Append("}; // " + BenLincoln.DebugSymbols.Utilities.GetListAsString(ObjectFileNames));
            builder.Append(Environment.NewLine);
            return builder.ToString();
        }

        public string GetAsInlineStruct(string memberName, TypeInfo ti, TypedValue tv)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append("struct {");
            builder.Append(Environment.NewLine);
            foreach (var m in _members)
            {
                builder.Append("\t");
                builder.Append(m.ToString());
                builder.Append(Environment.NewLine);
            }
            builder.Append("} ");
            builder.Append(memberName);
            builder.Append(string.Format("; // size={0}, offset={1}, found in object files: ", ti.Size, tv.Value));
            builder.Append(BenLincoln.DebugSymbols.Utilities.GetListAsString(ObjectFileNames));
            return builder.ToString();
        }

        public BenLincoln.DebugSymbols.SymDump.CStruct ToBLDSCStruct(int overlayID)
        {
            BenLincoln.DebugSymbols.SymDump.CStruct result = new BenLincoln.DebugSymbols.SymDump.CStruct();
            result.name = _name;
            result.hashcode = GetHashCode();
            result.sym_file_sequence_number = SYMFileSequenceNumber;
            result.source_file = SLDSourceFileName;
            result.object_files = ObjectFileNames;
            result.fake = OriginalWasFake;
            result.members = new List<BenLincoln.DebugSymbols.SymDump.CStructMember>();
            foreach (StructMember m in _members)
            {
                BenLincoln.DebugSymbols.SymDump.CStructMember newMem = m.ToBLDSCStructMember();
                newMem.overlay_id = overlayID;
                result.members.Add(newMem);
            }

            return result;
        }

        //public string DumpJSON(string sourceFile, string lineIndent)
        //{
        //    BenLincoln.DebugSymbols.SymDump.CStruct resultDef = ToBLDSCStruct();

        //    return BenLincoln.DebugSymbols.Utilities.IndentString(resultDef.ToJSON(), lineIndent);
        //}

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != GetType()) return false;
            return Equals((StructDef) obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                return ((_members != null ? _members.GetHashCode() : 0) * 397) ^
                       (_name != null ? _name.GetHashCode() : 0);
            }
        }

        public StructMember ForOffset(uint ofs)
        {
            return _members
                .Where(m => m.TypeInfo.ClassType != ClassType.Bitfield && m.TypedValue.Value <= ofs)
                .OrderBy(m => m.TypedValue.Value)
                .FirstOrDefault();
        }
    }
}
