﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

using BenLincoln.DebugSymbols.SymDump;

namespace BenLincoln.DebugSymbols
{
    public class GenerateGhidraScriptsResult
    {
        public StringBuilder ManualChangesBuilder;

        public Dictionary<long, List<CTypeDef>> FunctionPointerTypeDefs = new Dictionary<long, List<CTypeDef>>();
        public List<CTypeDef> AllFunctionPointerTypeDefs = new List<CTypeDef>();

        public List<string> DuplicateEnumNames = new List<string>();
        public List<string> DuplicateStructNames = new List<string>();
        public List<string> DuplicateUnionNames = new List<string>();

        public List<string> EnumStructDupeNames = new List<string>();
        public List<string> EnumUnionDupeNames = new List<string>();
        public List<string> StructUnionDupeNames = new List<string>();
    }

    public class GenerateSkeletonFilesForSubsetResult
    {
        public BenLincoln.DebugSymbols.SourceCodeSourceFileTracker FSTracker;
        public List<CFunction> AddedFunctions;

        public GenerateSkeletonFilesForSubsetResult()
        {
            FSTracker = new SourceCodeSourceFileTracker();
            AddedFunctions = new List<CFunction>();
        }
    }

    public class MetaGenerator
    {
        public const string DirectoryNameSourceStubs = "source-stubs";
        public const string DirectoryNameSourceDecompiled = "source-decompiled";
        public const string DirectoryNameGhidraFiles = "ghidra_files";

        public const string SubsetNameMainBinary = "PRIMARY";

        public const string FunctionPointerTypeDefPrefix = "TDRFuncPtr_";

        protected const int MaxFunctionDefinitionsPerGhidraScriptSubroutine = 100;
        protected const int MaxLabelDefinitionsPerGhidraScriptSubroutine = 50;
        protected const int MaxDataTypeImportsPerGhidraScriptSubroutine = 200;

        protected static int[] GetValidLabelTypes()
        {
            List<int> resultList = new List<int>();
            for (int i = 2; i < 255; i++)
            {
                resultList.Add(i);
            }
            return resultList.ToArray();
        }

        // labels that overlap with function names, etc. and cause problems
        protected static List<string> GetBlacklistedLabels()
        {
            List<string> result = new List<string>();
            result.Add("_ramsize");
            result.Add("__datalen");
            result.Add("__textlen");
            result.Add("__heapsize");

            result.Add("_bss_obj");
            result.Add("_bss_objend");
            result.Add("_bss_org");
            result.Add("_bss_orgend");
            result.Add("_bss_size");

            result.Add("_data_obj");
            result.Add("_data_objend");
            result.Add("_data_org");
            result.Add("_data_orgend");
            result.Add("_data_size");

            result.Add("_rdata_obj");
            result.Add("_rdata_objend");
            result.Add("_rdata_org");
            result.Add("_rdata_orgend");
            result.Add("_rdata_size");

            result.Add("_sbss_obj");
            result.Add("_sbss_objend");
            result.Add("_sbss_org");
            result.Add("_sbss_orgend");
            result.Add("_sbss_size");

            result.Add("_sdata_obj");
            result.Add("_sdata_objend");
            result.Add("_sdata_org");
            result.Add("_sdata_orgend");
            result.Add("_sdata_size");

            result.Add("_text_obj");
            result.Add("_text_objend");
            result.Add("_text_org");
            result.Add("_text_orgend");
            result.Add("_text_size");

            result.Add("__bss_obj");
            result.Add("__bss_objend");
            result.Add("__bss_org");
            result.Add("__bss_orgend");
            result.Add("__bss_size");
            result.Add("__bsslen");

            result.Add("__data_obj");
            result.Add("__data_objend");
            result.Add("__data_org");
            result.Add("__data_orgend");
            result.Add("__data_size");

            result.Add("__rdata_obj");
            result.Add("__rdata_objend");
            result.Add("__rdata_org");
            result.Add("__rdata_orgend");
            result.Add("__rdata_size");

            result.Add("__sbss_obj");
            result.Add("__sbss_objend");
            result.Add("__sbss_org");
            result.Add("__sbss_orgend");
            result.Add("__sbss_size");

            result.Add("__sdata_obj");
            result.Add("__sdata_objend");
            result.Add("__sdata_org");
            result.Add("__sdata_orgend");
            result.Add("__sdata_size");

            result.Add("__text_obj");
            result.Add("__text_objend");
            result.Add("__text_org");
            result.Add("__text_orgend");
            result.Add("__text_size");

            return result;
        }

        protected static List<string> GetBlacklistedLabelSuffixes()
        {
            List<string> result = new List<string>();
            result.Add("_ctors_obj");
            result.Add("_dtors_obj");

            result.Add("_ctors_org");
            result.Add("_dtors_org");

            result.Add("_ctors_size");
            result.Add("_dtors_size");

            result.Add("_objend");
            result.Add("_orgend");

            result.Add("_text_size");
            result.Add("_bss_size");
            result.Add("_sbss_size");
            result.Add("_data_size");
            result.Add("_sdata_size");

            result.Add("_text_obj");
            result.Add("_text_objend");
            result.Add("_text_org");
            result.Add("_text_orgend");

            result.Add("_data_obj");
            result.Add("_data_objend");
            result.Add("_data_org");
            result.Add("_data_orgend");

            result.Add("_rdata_obj");
            result.Add("_rdata_objend");
            result.Add("_rdata_org");
            result.Add("_rdata_orgend");

            result.Add("_sdata_obj");
            result.Add("_sdata_objend");
            result.Add("_sdata_org");
            result.Add("_sdata_orgend");

            result.Add("_bss_obj");
            result.Add("_bss_objend");
            result.Add("_bss_org");
            result.Add("_bss_orgend");

            result.Add("_sbss_obj");
            result.Add("_sbss_objend");
            result.Add("_sbss_org");
            result.Add("_sbss_orgend");

            return result;
        }

        public static string GetShortOverlayName(long overlayID)
        {
            return string.Format("OL{0:2X}", overlayID);
        }



        public static GenerateSkeletonFilesForSubsetResult GenerateSkeletonFilesForSubset(string projectName, long overlayID, string subsetName, string baseOutputDirectory,
            GenerateGhidraScriptsResult parentResult, bool mapSLDFunctions,
            List<CFunction> functions, List<PsyQSLDDataSet> sldData, List<CLabel> labels,
            List<CEnum> enums, List<CStruct> structs, List<CUnion> unions, List<CTypeDef> typedefs, List<CTypeDef> generated_typedefs,
            List<CExternalDeclaration> externs,
            string utilityName, string codeDirectoryName, 
            Assembly asm, string programName, string copyright, string infoURL, string[] thanksList)
        {
            GenerateSkeletonFilesForSubsetResult result = new GenerateSkeletonFilesForSubsetResult();
            StringBuilder manualChangesBuilderForSubset = new StringBuilder();
            // create base output directory structure
            string baseOutputDirectoryForSubset = Path.Combine(baseOutputDirectory, subsetName);
            string[] outputDirs = new string[] { DirectoryNameSourceStubs, DirectoryNameSourceDecompiled };
            BenLincoln.DebugSymbols.Utilities.CreateOutputDirectories(baseOutputDirectoryForSubset, outputDirs, asm, programName, copyright, infoURL, thanksList);
            string manualChangesPathForSubset = Path.Combine(baseOutputDirectoryForSubset, String.Format("README-{0}-{1}-{2}-Manual_Changes_Required.txt", projectName, subsetName, utilityName));

            //string mhfName = string.Format("{0}-{1}", projectName.ToUpper(), subsetName.ToUpper());
            string mhfName = string.Format("{0}", projectName.ToUpper(), subsetName.ToUpper());
            if (subsetName == SubsetNameMainBinary)
            {
                mhfName = projectName.ToUpper();
            }
            //string monolithicHeaderFilePath = Path.Combine(baseOutputDirectoryForSubset, DirectoryNameSourceStubs, String.Format("{0}.H", mhfName));
            string monolithicHeaderFilePath = Path.Combine(baseOutputDirectoryForSubset, String.Format("{0}.H", mhfName));

            MonolithicHeaderGenerationResult mhgr = SourceCodeGenerator.GenerateMonolithicHeader(enums, structs, unions, typedefs, generated_typedefs, externs, 
                projectName, manualChangesBuilderForSubset);
            manualChangesBuilderForSubset = mhgr.ManualChangesBuilder;

            // help operator locate duplicate names in the monolithic header
            if (parentResult != null)
            {
                manualChangesBuilderForSubset = BenLincoln.DebugSymbols.SourceCodeGenerator.CheckDuplicateNames(manualChangesBuilderForSubset, parentResult.DuplicateEnumNames, "enum", "Header", monolithicHeaderFilePath, mhgr.SourceCode, "which is defined multiple times");
                manualChangesBuilderForSubset = BenLincoln.DebugSymbols.SourceCodeGenerator.CheckDuplicateNames(manualChangesBuilderForSubset, parentResult.DuplicateStructNames, "struct", "Header", monolithicHeaderFilePath, mhgr.SourceCode, "which is defined multiple times");
                manualChangesBuilderForSubset = BenLincoln.DebugSymbols.SourceCodeGenerator.CheckDuplicateNames(manualChangesBuilderForSubset, parentResult.DuplicateUnionNames, "union", "Header", monolithicHeaderFilePath, mhgr.SourceCode, "which is defined multiple times");

                manualChangesBuilderForSubset = BenLincoln.DebugSymbols.SourceCodeGenerator.CheckDuplicateNames(manualChangesBuilderForSubset, parentResult.EnumStructDupeNames, "enum or struct", "Header", monolithicHeaderFilePath, mhgr.SourceCode, "which is used as a name for both an enum and a struct");
                manualChangesBuilderForSubset = BenLincoln.DebugSymbols.SourceCodeGenerator.CheckDuplicateNames(manualChangesBuilderForSubset, parentResult.EnumUnionDupeNames, "enum or union", "Header", monolithicHeaderFilePath, mhgr.SourceCode, "which is used as a name for both an enum and a union");
                manualChangesBuilderForSubset = BenLincoln.DebugSymbols.SourceCodeGenerator.CheckDuplicateNames(manualChangesBuilderForSubset, parentResult.StructUnionDupeNames, "struct or union", "Header", monolithicHeaderFilePath, mhgr.SourceCode, "which is used as a name for both a struct and a union");
            }

            BenLincoln.DebugSymbols.Utilities.WriteToOutput(monolithicHeaderFilePath, mhgr.SourceCode, FileMode.Create);

            foreach (CFunction func in functions)
            {
                result.FSTracker.AddOrUpdateEntry(func);
            }
            foreach (CEnum e in enums)
            {
                result.FSTracker.AddOrUpdateEntry(e);
            }
            foreach (CStruct s in structs)
            {
                result.FSTracker.AddOrUpdateEntry(s);
            }
            foreach (CUnion u in unions)
            {
                result.FSTracker.AddOrUpdateEntry(u);
            }
            foreach (CExternalDeclaration x in externs)
            {
                result.FSTracker.AddOrUpdateEntry(x);
            }
            foreach (CTypeDef t in typedefs)
            {
                result.FSTracker.AddOrUpdateEntry(t);
            }
            foreach (CTypeDef t in generated_typedefs)
            {
                result.FSTracker.AddOrUpdateEntry(t);
            }

            if (mapSLDFunctions)
            {
                SourceCodeSourceFileTracker sldMapTracker = new SourceCodeSourceFileTracker();
                foreach (CFunction f in functions)
                {
                    sldMapTracker.AddOrUpdateEntry(f);
                }
                foreach (PsyQSLDDataSet ds in sldData)
                {
                    foreach (AddressAndLineNumberTuple aaln in ds.data)
                    {
                        CFunction func = new CFunction();
                        CFunction existingFunc = sldMapTracker.GetFunctionAtAddress(ds.source_file, aaln.address);
                        bool addFunc = true;
                        if (existingFunc != null)
                        {
                            func = existingFunc;
                            if (func.blocks == null)
                            {
                                func.blocks = new List<CBlock>();
                            }
                            CBlock newBlock = new CBlock();
                            newBlock.start_line = aaln.line_number;
                            newBlock.end_line = aaln.line_number + 1;
                            newBlock.overlay_id = func.overlay_id;
                            func.blocks.Add(newBlock);
                            addFunc = false;
                        }
                        else
                        {
                            CElement checkElem = sldMapTracker.GetElementAtAddress(aaln.address);
                            if (checkElem != null)
                            {
                                Console.WriteLine(string.Format("Info: ignoring SLD entry @0x{0:X8}, because there is an existing element named '{1}' at that address in the debug symbol data.", aaln.address, checkElem.name));
                                addFunc = false;
                            }
                            else
                            {
                                func.offset_in_memory = aaln.address;
                                func.source_file = ds.source_file;
                                func.source_file_line_number = aaln.line_number;
                                func.overlay_id = ds.overlay_id;
                                List<CLabel> labelsForAddress = SymDumpSymbolCollection.GetLabelsByAddress(labels, func.offset_in_memory, null);
                                if (labelsForAddress.Count > 0)
                                {
                                    string chosenLabelName = labelsForAddress[0].name;
                                    bool gotNameWithConfidence = false;
                                    // TKTK: any additional logic to choose another of the labels
                                    for (int i = 0; i < labelsForAddress.Count; i++)
                                    {
                                        if (labelsForAddress[i].name == "__SN_ENTRY_POINT")
                                        {
                                            chosenLabelName = "__SN_ENTRY_POINT";
                                            gotNameWithConfidence = true;
                                            break;
                                        }
                                    }
                                    if (!gotNameWithConfidence)
                                    {
                                        for (int i = 0; i < labelsForAddress.Count; i++)
                                        {
                                            if (labelsForAddress[i].label_type > 1)
                                            {
                                                chosenLabelName = labelsForAddress[i].name;
                                                gotNameWithConfidence = true;
                                                break;
                                            }
                                        }
                                    }
                                    if (!gotNameWithConfidence)
                                    {
                                        if (chosenLabelName.StartsWith("_"))
                                        {
                                            for (int i = 0; i < labelsForAddress.Count; i++)
                                            {
                                                if (!labelsForAddress[i].name.StartsWith("_"))
                                                {
                                                    chosenLabelName = labelsForAddress[i].name;
                                                    break;
                                                }
                                            }
                                        }
                                    }
                                    if (!gotNameWithConfidence)
                                    {
                                        if (chosenLabelName.StartsWith("_"))
                                        {
                                            // so far, all of the names I've seen with leading underscores in this type of data are 
                                            // NOT function names except for __SN_ENTRY_POINT. If I find any with further testing, I'll disable 
                                            // this part.
                                            func.name = string.Format("function_{0:X8}", aaln.address);
                                        }
                                    }
                                    if (labelsForAddress.Count > 1)
                                    {
                                        StringBuilder mcLabelBuilder = new StringBuilder();
                                        mcLabelBuilder.Append(string.Format("Warning: multiple labels were found in the debug symbols for the address 0x{0:X8}. '{1}' was automatically selected for use. You may wish to manually specify another name for the function in Ghidra.", aaln.address, chosenLabelName));
                                        mcLabelBuilder.Append(Environment.NewLine);
                                        mcLabelBuilder.Append("\tAll labels for this address:");
                                        mcLabelBuilder.Append(Environment.NewLine);
                                        foreach (CLabel l in labelsForAddress)
                                        {
                                            mcLabelBuilder.Append(string.Format("\t\t'{0}'", l.name));
                                            mcLabelBuilder.Append(Environment.NewLine);
                                        }
                                        manualChangesBuilderForSubset.Append(mcLabelBuilder.ToString());
                                        Console.WriteLine(mcLabelBuilder.ToString());
                                    }
                                    func.name = chosenLabelName;
                                }
                                else
                                {
                                    addFunc = false;
                                    func.name = string.Format("function_{0:X8}", aaln.address);
                                }
                                func.method_signature = string.Format("void {0}() /* {0} method signature is not contained in the debug symbol data. This is likely either a library function or the game was compiled without debug symbols. Please refer to the TDR documentation for additional guidance. */", func.name);
                                CParameter rvPlaceHolder = new CParameter();
                                rvPlaceHolder.parameter_signature = "void";
                                func.return_value = rvPlaceHolder;
                            }
                        }
                        if (addFunc)
                        {
                            sldMapTracker.AddOrUpdateEntry(func);
                            if (!result.AddedFunctions.Contains(func))
                            {
                                result.AddedFunctions.Add(func);
                            }
                        }
                    }
                }
                foreach (string sourceFile in sldMapTracker.FunctionSourceFiles.Keys)
                {
                    foreach (CFunction func in sldMapTracker.FunctionSourceFiles[sourceFile])
                    {
                        result.FSTracker.AddOrUpdateEntryIfNotAlreadyDefined(func);
                    }
                    // TKTK: Define these as functions in the Ghidra script too (maybe?)
                }
            }

            // create stubs for individual source files
            IndividualSourceFileGenerationResult allResults = new IndividualSourceFileGenerationResult();

            foreach (string sourceFileName in result.FSTracker.AllSourceFiles)
            {
                List<CFunction> funcs = new List<CFunction>();
                List<CElement> elems = new List<CElement>();
                if (result.FSTracker.FunctionSourceFiles.ContainsKey(sourceFileName))
                {
                    funcs = result.FSTracker.FunctionSourceFiles[sourceFileName];
                }
                if (result.FSTracker.OtherElementSourceFiles.ContainsKey(sourceFileName))
                {
                    elems = result.FSTracker.OtherElementSourceFiles[sourceFileName];
                }
                string actualSourceFileName = sourceFileName;
                if (actualSourceFileName == "")
                {
                    actualSourceFileName = SourceCodeGenerator.GetUnmappedSourceCodeFileName(subsetName);
                }
                IndividualSourceFileGenerationResult isfgr = SourceCodeGenerator.GenerateSourceCodeFiles(mhgr, parentResult, projectName, baseOutputDirectoryForSubset, 
                    codeDirectoryName, actualSourceFileName, SourceCodeGenerator.GetUnmappedHeaderFileName(subsetName), funcs, elems);
                //foreach (string fileName in isfgr.OutputFiles.Keys)
                //{
                //    BenLincoln.DebugSymbols.Utilities.CreateParentPath(fileName);
                //    BenLincoln.DebugSymbols.Utilities.WriteToOutput(fileName, isfgr.OutputFiles[fileName], FileMode.Create);
                //}
                foreach (string fileName in isfgr.OutputFiles.Keys)
                {
                    if (allResults.OutputFiles.ContainsKey(fileName))
                    {
                        allResults.OutputFiles[fileName].BodyContent = string.Format("{0}{1}{2}", 
                            allResults.OutputFiles[fileName].BodyContent,
                            Environment.NewLine,
                            isfgr.OutputFiles[fileName].BodyContent);
                    }
                    else
                    {
                        allResults.OutputFiles.Add(fileName, isfgr.OutputFiles[fileName]);
                    }
                }
            }

            foreach (string fileName in allResults.OutputFiles.Keys)
            {
                BenLincoln.DebugSymbols.Utilities.CreateParentPath(fileName);
                BenLincoln.DebugSymbols.Utilities.WriteToOutput(fileName, allResults.OutputFiles[fileName].GetAllContent(), FileMode.Create);
            }

            // create empty files for source files which were referenced in the SLD information, but not anywhere else
            List<string> createdSLDFiles = new List<string>();
            foreach (PsyQSLDDataSet ds in sldData)
            {
                if (createdSLDFiles.Contains(ds.source_file))
                {
                    Console.WriteLine(string.Format("Debug: ignoring duplicate SLD information reference to the file '{0}'.", ds.source_file));
                }
                else
                {
                    createdSLDFiles.Add(ds.source_file);
                    if (result.FSTracker.FunctionSourceFiles.Keys.Contains(ds.source_file))
                    {
                        Console.WriteLine(string.Format("Info: SLD information for this project contains a reference to the file '{0}', which is already referenced elsewhere. No placeholder will be created.", ds.source_file));
                    }
                    else
                    {
                        IndividualSourceFileGenerationResult isfgr = SourceCodeGenerator.GetPlaceholderFile(ds.source_file, baseOutputDirectoryForSubset, DirectoryNameSourceStubs);

                        foreach (string fileName in isfgr.OutputFiles.Keys)
                        {
                            Console.WriteLine(string.Format("Info: SLD information for this project contains a reference to the file '{0}', which is not referenced elsewhere. A placeholder will be created at '{1}'.", ds.source_file, fileName));
                            BenLincoln.DebugSymbols.Utilities.CreateParentPath(fileName);
                            BenLincoln.DebugSymbols.Utilities.WriteToOutput(fileName, isfgr.OutputFiles[fileName].GetAllContent(), FileMode.Create);
                        }
                    }
                }
            }
            BenLincoln.DebugSymbols.Utilities.WriteToOutput(manualChangesPathForSubset, manualChangesBuilderForSubset.ToString(), FileMode.Create);
            return result;
        }

        public static Dictionary<string, string> ReplacePlaceholderInTemplates(Dictionary<string, string> templates, string placeHolder, string replacement)
        {
            List<string> templateNames = new List<string>();
            Dictionary<string, string> result = new Dictionary<string, string>();

            foreach (string templateName in templates.Keys)
            {
                string updated = templates[templateName];
                updated = updated.Replace(placeHolder, replacement);
                result.Add(templateName, updated);
            }

            return result;
        }

        public static GenerateGhidraScriptsResult GenerateGhidraScripts(long programCounter, long stackPointerBase, long stackPointerSizeOrOffset,
            long globalPointer,
            List<MemorySegment> memSegments,
            List<CFunction> functions, List<PsyQSLDDataSet> sldData, List<CLabel> labels,
            List<CEnum> enums, List<CStruct> structs, List<CUnion> unions, List<CTypeDef> typedefs,
            List<CExternalDeclaration> externs, 
            List<BenLincoln.DebugSymbols.SymDump.PsyQOverlayDescription> includedOverlays, 
            string projectName, string baseOutputDirectory,
            bool assumeSNGPBase, bool explicitGPBase, long GPBase, bool addStackPointerOffset, bool createPlayStationMemory,
            bool ignoreLabels, bool ignoreNonASCIILabels, bool replaceNonASCIILabels, bool convertExternalDefinitionsToLabels, bool mapSLDFunctions,
            bool assignReturnAddressStorage, bool assignParameterStorage,
            StringBuilder manualChangesBuilder,
            string ghidraXMLSymbols,
            Assembly asm, string programName, string copyright, string infoURL, string[] thanksList)
        {
            GenerateGhidraScriptsResult result = new GenerateGhidraScriptsResult();

            if (ignoreNonASCIILabels && replaceNonASCIILabels)
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("ASCII labels can either be ignored or replaced, but not both.", 1, asm, programName, copyright, infoURL, thanksList);
            }

            // handle base/shared code
            string ghidraMapMemoryAndLabelsScriptPath = Path.Combine(baseOutputDirectory, DirectoryNameGhidraFiles, String.Format("{0}TDRMapMemoryAndCreateLabels.java", projectName));
            string ghidraDefineFunctionsScriptPath = Path.Combine(baseOutputDirectory, DirectoryNameGhidraFiles, String.Format("{0}TDRDefineFunctions.java", projectName));
            string ghidraArrayIdentificationScriptPath = Path.Combine(baseOutputDirectory, DirectoryNameGhidraFiles, String.Format("{0}TDRAggressiveArrayIdentification.java", projectName));
            string ghidraCodeExportScriptPath = Path.Combine(baseOutputDirectory, DirectoryNameGhidraFiles, String.Format("{0}TDRDecompile.java", projectName));
            string ghidraDataExportScriptPath = Path.Combine(baseOutputDirectory, DirectoryNameGhidraFiles, String.Format("{0}TDRExportData.java", projectName));
            string ghidraXMLFilePath = Path.Combine(baseOutputDirectory, DirectoryNameGhidraFiles, String.Format("{0}.xml", projectName));
            string ghidraDecompiledCCodeFilePath = Path.Combine(baseOutputDirectory, String.Format("{0}.C", projectName));
            //string ghidraExportedDataCCodeFilePath = Path.Combine(baseOutputDirectory, String.Format("{0}GLOBALVARS.C", projectName));
            string ghidraExportedDataCCodeFilePath = Path.Combine(baseOutputDirectory, String.Format("XPRTDATA.C", projectName));

            // create truly monolithic header file mostly for Ghidra
            string monolithicHeaderFilePath = Path.Combine(baseOutputDirectory, String.Format("{0}.H", projectName));

            // create Ghidra function-definition script
            string ghidraMemoryMapAndLabelScriptTemplate = BenLincoln.DebugSymbols.Utilities.GetEmbeddedTextFile(asm, "CreateSkeleton.GhidraScript-Template-MapMemoryAndCreateLabels.txt");
            string ghidraFunctionDefinitionScriptTemplate = BenLincoln.DebugSymbols.Utilities.GetEmbeddedTextFile(asm, "CreateSkeleton.GhidraScript-Template-DefineFunctions.txt");
            string ghidraDefineFunctionsSubroutineTemplate = BenLincoln.DebugSymbols.Utilities.GetEmbeddedTextFile(asm, "CreateSkeleton.GhidraScript-Template-DefineFunctionsSubroutine.txt");
            string ghidraDefineLabelsSubroutineTemplate = BenLincoln.DebugSymbols.Utilities.GetEmbeddedTextFile(asm, "CreateSkeleton.GhidraScript-Template-CreateLabelsSubroutine.txt");
            string ghidraImportDataTypeSubroutineTemplate = BenLincoln.DebugSymbols.Utilities.GetEmbeddedTextFile(asm, "CreateSkeleton.GhidraScript-Template-ImportDataTypesSubroutine.txt");
            string ghidraCustomDecompileExportTemplate = BenLincoln.DebugSymbols.Utilities.GetEmbeddedTextFile(asm, "CreateSkeleton.GhidraScript-Template-TDRDecompile.txt");
            string ghidraArrayIdentificationTemplate = BenLincoln.DebugSymbols.Utilities.GetEmbeddedTextFile(asm, "CreateSkeleton.GhidraScript-Template-TDRAggressiveArrayIdentification.txt");
            string ghidraDataExportTemplate = BenLincoln.DebugSymbols.Utilities.GetEmbeddedTextFile(asm, "CreateSkeleton.GhidraScript-Template-TDRExportData.txt");

            Dictionary<string, string> templatesForStringReplacement = new Dictionary<string, string>();
            templatesForStringReplacement.Add("ghidraMemoryMapAndLabelScriptTemplate", ghidraMemoryMapAndLabelScriptTemplate);
            templatesForStringReplacement.Add("ghidraFunctionDefinitionScriptTemplate", ghidraFunctionDefinitionScriptTemplate);
            templatesForStringReplacement.Add("ghidraCustomDecompileExportTemplate", ghidraCustomDecompileExportTemplate);
            templatesForStringReplacement.Add("ghidraArrayIdentificationTemplate", ghidraArrayIdentificationTemplate);
            templatesForStringReplacement.Add("ghidraDataExportTemplate", ghidraDataExportTemplate);

            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%DECOMPILED_C_CODE_FILE%", ghidraDecompiledCCodeFilePath.Replace("\\", "\\\\"));

            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%EXPORTED_DATA_C_CODE_FILE%", ghidraExportedDataCCodeFilePath.Replace("\\", "\\\\"));

            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%PROJECT_NAME%", projectName);

            // set register values
            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%ENTRYPOINT_ADDRESS%", string.Format("0x{0:X8}", programCounter));
            long stackPointerValue = stackPointerBase;
            long snGlobalPointerValue = 0;
            CLabel snGLabel = SymDumpSymbolCollection.GetLabelByName(labels, "__SN_GP_BASE");
            if (snGLabel != null)
            {
                snGlobalPointerValue = snGLabel.offset_in_memory;
                Console.WriteLine(string.Format("Info: found __SN_GP_BASE (global pointer value) 0x{0:X8} in the symbol data.", snGlobalPointerValue));
                if (assumeSNGPBase)
                {
                    Console.WriteLine(string.Format("Info: will request that Ghidra explicitly assume that $gp is set to 0x{0:X8} at the beginning of each function except 'entry'.", snGlobalPointerValue));
                }
                else
                {
                    Console.WriteLine("Warning: The __SN_GP_BASE value will be ignored. It is highly recommended that you re-run CreateSkeleton.exe with the --assume-sn-gp-base flag unless doing so is resulting in bad data.");
                }
            }
            else
            {
                if (assumeSNGPBase)
                {
                    BenLincoln.DebugSymbols.Utilities.ExitWithError("Error: Did not find an __SN_GP_BASE value in the debug symbol data. You should re-run CreateSkeleton.exe without the --assume-sn-gp-base flag.", 1, asm, programName, copyright, infoURL, thanksList);
                }
                else
                {
                    Console.WriteLine("Info: Did not find an __SN_GP_BASE value in the debug symbol data.");
                }
            }
            if (explicitGPBase)
            {
                if (snGLabel != null)
                {
                    Console.WriteLine(string.Format("Warning: a global pointer value of 0x{0:X8} has been manually specified. This will override the value of 0x{1:X8} found in the debug symbol data. Be sure you know what you are doing.", GPBase, snGlobalPointerValue));
                }
                else
                {
                    Console.WriteLine(string.Format("Info: a global pointer value of 0x{0:X8} has been manually specified.", GPBase));
                }
                assumeSNGPBase = true;
                snGlobalPointerValue = GPBase;
            }

            if (addStackPointerOffset)
            {
                stackPointerValue += stackPointerSizeOrOffset;
                if (stackPointerSizeOrOffset != 0)
                {
                    Console.WriteLine(string.Format("Info: adding the non-zero stack pointer offset value 0x{0:X8} to the stack pointer value.", stackPointerSizeOrOffset));
                }

            }
            else
            {
                if (stackPointerSizeOrOffset != 0)
                {
                    Console.WriteLine(string.Format("Warning: the stack pointer size/offset value for this binary is not zero (0x{0:X8}). If you notice bad decompilation results, consider re-running CreateSkeleton.exe with the --add-stack-pointer-offset flag.", stackPointerSizeOrOffset));
                }
            }
            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%STACK_POINTER_VALUE%", string.Format("0x{0:X8}", stackPointerValue));
            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%GLOBAL_POINTER_VALUE%", string.Format("0x{0:X8}", globalPointer));

            StringBuilder memoryMapBuilder = new StringBuilder();

            long lowerCodeBound = 0xFFFFFFFF;
            long upperCodeBound = 0x00000000;
            if (memSegments.Count > 0)
            {
                memoryMapBuilder.Append("\t\tMemory mem = currentProgram.getMemory();");
                memoryMapBuilder.Append(Environment.NewLine);
                foreach (MemorySegment ms in memSegments)
                {
                    memoryMapBuilder.Append(string.Format("\t\t// segment {0} - {1}: 0x{2:X8}-0x{3:X8} (length: 0x{4:X8})", ms.segment_number, BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(ms.name), ms.offset, ms.offset + ms.length, ms.length));
                    memoryMapBuilder.Append(Environment.NewLine);

                    bool includeMemory = true;

                    if (ms.segment_number > 1)
                    {
                        if (createPlayStationMemory)
                        {
                            includeMemory = false;
                        }
                        else
                        {
                            memoryMapBuilder.Append(string.Format("\t\tMemoryBlock mb{0} = mem.createUninitializedBlock(\"{1}\", toAddr(0x{2:X8}), 0x{3:X8}, false);", ms.segment_number, BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(ms.name), ms.offset, ms.length));
                            memoryMapBuilder.Append(Environment.NewLine);
                            memoryMapBuilder.Append(string.Format("\t\tmb{0}.setRead(true);", ms.segment_number));
                            memoryMapBuilder.Append(Environment.NewLine);
                            memoryMapBuilder.Append(string.Format("\t\tmb{0}.setWrite(true);", ms.segment_number));
                            memoryMapBuilder.Append(Environment.NewLine);
                            memoryMapBuilder.Append(string.Format("\t\tmb{0}.setExecute(true);", ms.segment_number));
                            memoryMapBuilder.Append(Environment.NewLine);
                        }
                    }

                    if (includeMemory)
                    {
                        if (ms.offset < lowerCodeBound)
                        {
                            lowerCodeBound = ms.offset;
                        }
                        long segmentUpperBound = ms.offset + ms.length;
                        if (segmentUpperBound > upperCodeBound)
                        {
                            upperCodeBound = segmentUpperBound;
                        }
                    }
                }
            }
            if (createPlayStationMemory)
            {
                if (lowerCodeBound > upperCodeBound)
                {
                    Console.WriteLine(string.Format("Error: Text/code segment lower bound was calculated as 0x{0:X8}, but the upper bound was calculated as 0x{1:X8}, which is below the lower bound. Please contact the developer of this tool for guidance. The additional PlayStation memory map data will not be created.", lowerCodeBound, upperCodeBound));
                }
                else
                {
                    memoryMapBuilder.Append("\t\t// create all additional memory map/segment data");
                    memoryMapBuilder.Append(Environment.NewLine);
                    memoryMapBuilder.Append(string.Format("\t\tsetupMemoryMap(currentProgram, 0x{0:X8}L, 0x{1:X8}L);", lowerCodeBound, (upperCodeBound - lowerCodeBound)));
                    memoryMapBuilder.Append(Environment.NewLine);
                }
            }

            long psMemorySegmentBegin = 0x0;
            long psMemorySegmentEnd = 0x1FFFFFFF;
            if ((lowerCodeBound >= 0x80000000) && (lowerCodeBound < 0xA0000000))
            {
                psMemorySegmentBegin = 0x80000000;
                psMemorySegmentEnd = 0x9FFFFFFF;
            }
            if (lowerCodeBound >= 0xA0000000)
            {
                psMemorySegmentBegin = 0xA0000000;
                psMemorySegmentEnd = 0xBFFFFFFF;
            }

            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%PS_PRIMARY_MEMORY_BEGIN%", string.Format("0x{0:X8}", psMemorySegmentBegin));
            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%PS_PRIMARY_MEMORY_END%", string.Format("0x{0:X8}", psMemorySegmentEnd));

            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%MEMORY_MAP_POPULATION%", memoryMapBuilder.ToString());

            StringBuilder labelImportBuilder = new StringBuilder();
            StringBuilder labelSubImportBuilder = new StringBuilder();
            StringBuilder labelImportCallBuilder = new StringBuilder();
            int labelNum = 0;
            int labelsInSubCount = 0;
            int labelSubRoutineNumber = 0;
            List<string> blacklistedLabels = GetBlacklistedLabels();
            List<string> blacklistedLabelSuffixes = GetBlacklistedLabelSuffixes();

            if (!ignoreLabels)
            {
                int lNum = 0;
                foreach (CLabel l in labels)
                {
                    bool includeLabel = true;
                    // not functions and whatnot
                    if (l.label_type < 2)
                    {
                        includeLabel = false;
                        Console.WriteLine(string.Format("Warning: Ignoring type {0} label '{1}' at 0x{2:X8}", l.label_type, l.name, l.offset_in_memory));
                    }
                    //if ((includeLabel) && (l.name.StartsWith("_")))
                    //{
                    //    if (blacklistedLabels.Contains(l.name))
                    //    {
                    //        includeLabel = false;
                    //    }
                    //    if (includeLabel)
                    //    {
                    //        foreach (string blacklisted in blacklistedLabelSuffixes)
                    //        {
                    //            if (l.name.EndsWith(blacklisted))
                    //            {
                    //                includeLabel = false;
                    //                break;
                    //            }
                    //        }
                    //    }
                    //}
                    //if (l.offset < MinAddressForLabels)
                    //{
                    //    includeLabel = false;
                    //}
                    if (includeLabel && ignoreNonASCIILabels)
                    {
                        if (l.name.Trim() == "")
                        {
                            includeLabel = false;
                            Console.WriteLine(string.Format("Warning: Ignoring blank label at 0x{0:X8}", l.offset_in_memory));
                        }
                        if (includeLabel)
                        {
                            char[] labelArray = l.name.ToCharArray();
                            for (int charNum = 0; charNum < labelArray.Length; charNum++)
                            {
                                if ((labelArray[charNum] < 20) || (labelArray[charNum] > 127))
                                {
                                    includeLabel = false;
                                    Console.WriteLine(string.Format("Warning: Ignoring label with non-ASCII characters at 0x{0:X8}", l.offset_in_memory));
                                    break;
                                }
                            }
                        }
                    }
                    if (includeLabel)
                    {
                        string labelName = BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(l.name);
                        labelSubImportBuilder.Append(String.Format("\t\tSymbol sym{0} = symbolTable.createLabel(toAddr(0x{1:X8}), \"{2}\", SourceType.USER_DEFINED);", labelNum, l.offset_in_memory, labelName));
                        labelSubImportBuilder.Append(Environment.NewLine);
                        labelSubImportBuilder.Append(String.Format("\t\tsym{0}.setPrimary();", labelNum));
                        labelSubImportBuilder.Append(Environment.NewLine);
                        labelNum++;
                        labelsInSubCount++;
                    }
                    lNum++;
                    if ((labelsInSubCount > MaxLabelDefinitionsPerGhidraScriptSubroutine) || (lNum >= labels.Count))
                    {
                        string subRoutineDefinition = ghidraDefineLabelsSubroutineTemplate.Replace("%LABEL_DEFINITION_SUB_CONTENT%", labelSubImportBuilder.ToString());
                        subRoutineDefinition = subRoutineDefinition.Replace("%CREATE_LABELS_SUBROUTINE_NUMBER%", String.Format("{0}", labelSubRoutineNumber));
                        labelImportBuilder.Append(subRoutineDefinition);
                        labelImportCallBuilder.Append(String.Format("\t\tCreateLabels{0}(dataTypeMap);", labelSubRoutineNumber));
                        labelImportCallBuilder.Append(Environment.NewLine);
                        labelSubImportBuilder.Clear();
                        labelsInSubCount = 0;
                        labelSubRoutineNumber++;
                    }
                }
            }

            if (convertExternalDefinitionsToLabels)
            {
                int xNum = 0;
                foreach (CExternalDeclaration x in externs)
                {
                    // create the label for the extern/static
                    string externName = BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeCNameWithDefaultReplacement(x.GetName());
                    labelSubImportBuilder.Append(String.Format("\t\tSymbol sym{0} = symbolTable.createLabel(toAddr(0x{1:X8}), \"{2}\", SourceType.USER_DEFINED);", labelNum, x.offset_in_memory, externName));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append(String.Format("\t\tsym{0}.setPrimary();", labelNum));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    // TKTK: set the data type for the bytes of the thing
                    // note: array size 0 in a struct definition means basically "occupies zero space in the struct itself, 
                    //       but the struct may be followed by an arbitrary number of whatever this thing is"
                    //       https://gcc.gnu.org/onlinedocs/gcc/Zero-Length.html
                    // GetDataTypeFromMap(dataTypeMap, "struct AadSeqEvent *")
                    // createData(addr, wordDataType);
                    int arrayCount = x.GetCount();
                    string typeName = x.GetTypeName();
                    labelSubImportBuilder.Append(String.Format("\t\tlong xAddr{0} = {1}L;", labelNum, x.offset_in_memory));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append(String.Format("\t\tDataType xDataType{0} = GetDataTypeFromMap(dataTypeMap, \"{1}\");", labelNum, BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(typeName)));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append(String.Format("\t\tif (xDataType{0} != null);", labelNum));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append("\t\t{");
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append(String.Format("\t\t\tlong typeSize = xDataType{0}.getLength();", labelNum));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    //for (int i = 0; i < arrayCount; i++)
                    //{
                    labelSubImportBuilder.Append(String.Format("\t\t\tfor (int i = 0; i < {0}; i++)", arrayCount));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append("\t\t\t{");
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append(String.Format("\t\t\t\tClearAndCreateData(xAddr{0}, typeSize, xDataType{0});", labelNum));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append(String.Format("\t\t\t\txAddr{0} = xAddr{0} + typeSize;", labelNum));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append("\t\t\t}");
                    labelSubImportBuilder.Append(Environment.NewLine);
                    //}
                    labelSubImportBuilder.Append("\t\t}");
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelNum++;
                    labelsInSubCount++;
                    xNum++;
                    if ((labelsInSubCount > MaxLabelDefinitionsPerGhidraScriptSubroutine) || (xNum >= (externs.Count - 1)))
                    {
                        string subRoutineDefinition = ghidraDefineLabelsSubroutineTemplate.Replace("%LABEL_DEFINITION_SUB_CONTENT%", labelSubImportBuilder.ToString());
                        subRoutineDefinition = subRoutineDefinition.Replace("%CREATE_LABELS_SUBROUTINE_NUMBER%", String.Format("{0}", labelSubRoutineNumber));
                        labelImportBuilder.Append(subRoutineDefinition);
                        labelImportCallBuilder.Append(String.Format("\t\tCreateLabels{0}(dataTypeMap);", labelSubRoutineNumber));
                        labelImportCallBuilder.Append(Environment.NewLine);
                        labelSubImportBuilder.Clear();
                        labelsInSubCount = 0;
                        labelSubRoutineNumber++;
                    }
                }
            }

            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%LABEL_CREATION_SUBROUTINES%", labelImportBuilder.ToString());

            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%LABEL_CREATION_SUBROUTINE_CALLS%", labelImportCallBuilder.ToString());

            int dtImportSubNum = 0;

            AddDataTypeByNameResult enumAddResult = GhidraScriptGenerator.GetAddDataTypesScriptCode(ghidraImportDataTypeSubroutineTemplate,
                enums, projectName, "enum", MaxDataTypeImportsPerGhidraScriptSubroutine, dtImportSubNum);
            dtImportSubNum = enumAddResult.NextSubroutineNumber;
            List<string> addedEnums = enumAddResult.AddedNames;
            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%ENUM_IMPORT_SUBROUTINE_CALLS%", enumAddResult.GhidraCodeSubroutineCalls);
            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%ENUM_IMPORT_SUBROUTINES%", enumAddResult.GhidraCodeSubroutines);

            AddDataTypeByNameResult structAddResult = GhidraScriptGenerator.GetAddDataTypesScriptCode(ghidraImportDataTypeSubroutineTemplate,
                structs, projectName, "struct", MaxDataTypeImportsPerGhidraScriptSubroutine, dtImportSubNum);
            dtImportSubNum = structAddResult.NextSubroutineNumber;
            List<string> addedStructs = structAddResult.AddedNames;
            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%STRUCT_IMPORT_SUBROUTINE_CALLS%", structAddResult.GhidraCodeSubroutineCalls);
            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%STRUCT_IMPORT_SUBROUTINES%", structAddResult.GhidraCodeSubroutines);

            AddDataTypeByNameResult unionAddResult = GhidraScriptGenerator.GetAddDataTypesScriptCode(ghidraImportDataTypeSubroutineTemplate,
                unions, projectName, "union", MaxDataTypeImportsPerGhidraScriptSubroutine, dtImportSubNum);
            dtImportSubNum = unionAddResult.NextSubroutineNumber;
            List<string> addedUnions = unionAddResult.AddedNames;
            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%UNION_IMPORT_SUBROUTINE_CALLS%", unionAddResult.GhidraCodeSubroutineCalls);
            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%UNION_IMPORT_SUBROUTINES%", unionAddResult.GhidraCodeSubroutines);

            StringBuilder functionImportBuilder = new StringBuilder();
            StringBuilder functionSubImportBuilder = new StringBuilder();
            StringBuilder functionSubSubImportBuilder = new StringBuilder();
            int functionNum = 0;
            int subRoutineNumber = 0;
            int subroutineFunctionCount = 0;
            StringBuilder fpFunctionImportBuilder = new StringBuilder();
            StringBuilder fpFunctionSubImportBuilder = new StringBuilder();
            StringBuilder fpFunctionSubSubImportBuilder = new StringBuilder();
            int fpFunctionNum = 0;
            int fpSubRoutineNumber = 0;
            int fpSubroutineFunctionCount = 0;
            List<CFunction> functionsWithFunctionPointerParams = new List<CFunction>();
            List<string> functionsReferredToByPointers = new List<string>();
            //Dictionary<string, List<CFunction>> sourceFiles = new Dictionary<string, List<CFunction>>();
            //BenLincoln.DebugSymbols.FunctionSourceTracker fsTracker = new BenLincoln.DebugSymbols.FunctionSourceTracker();

            //quick hack to handle function pointer parameters, this should probably be done better
            List<string> functionPointerFunctionNames = new List<string>();
            List<string> functionPointerNames = new List<string>();

            result.FunctionPointerTypeDefs.Add(-1, new List<CTypeDef>());
            foreach (PsyQOverlayDescription ol in includedOverlays)
            {
                result.FunctionPointerTypeDefs.Add(ol.overlay_id, new List<CTypeDef>());
            }

            foreach (CFunction func in functions)
            {
                try
                {
                    if (func.HasFunctionPointerParameters())
                    {
                        manualChangesBuilder.Append(String.Format("Method {0} has a return type and/or parameters which are function pointers. You should manually verify in Ghidra that they were imported correctly, as they can be slightly problematic.", func.name));
                        manualChangesBuilder.Append(Environment.NewLine);
                        manualChangesBuilder.Append(String.Format("\tMethod signature: {0}", func.method_signature));
                        manualChangesBuilder.Append(Environment.NewLine);
                        manualChangesBuilder.Append(String.Format("\tMemory offset: 0x{0:X8}", func.offset_in_memory));
                        manualChangesBuilder.Append(Environment.NewLine);
                        manualChangesBuilder.Append(String.Format("\tFile: {0}", func.source_file));
                        manualChangesBuilder.Append(Environment.NewLine);
                        manualChangesBuilder.Append(String.Format("\tOriginal line number: {0}", func.source_file_line_number));
                        manualChangesBuilder.Append(Environment.NewLine);
                        manualChangesBuilder.Append(String.Format("\tParameters which are function pointers:"));
                        manualChangesBuilder.Append(Environment.NewLine);
                        if (func.return_value.is_function_pointer)
                        {
                            //string fptd = GetFunctionPointerTypeDefFromParameter(func.return_value);
                            string fptd = func.return_value.parameter_signature;
                            // this is a hack, but I can't think of a better approach ATM
                            string newTypeName = string.Format("{0}{1}", FunctionPointerTypeDefPrefix, func.name);
                            string oldTypeName = func.return_value.type_name;
                            func.return_value.type_name = newTypeName;
                            func.method_signature = func.method_signature.Replace(oldTypeName, newTypeName);
                            if (!functionPointerNames.Contains(newTypeName))
                            {
                                functionPointerNames.Add(newTypeName);
                            }
                            fptd = fptd.Replace("*)(", string.Format("*{0})(", newTypeName));
                            int lastSpacePos = fptd.LastIndexOf(" ");
                            if (lastSpacePos > 0)
                            {
                                //fptd = fptd.Substring(0, lastSpacePos).Trim() + ";";
                                fptd = fptd.Substring(0, lastSpacePos).Trim();
                            }
                            CTypeDef newTypeDef = new CTypeDef();
                            newTypeDef.typedef_signature = fptd;
                            newTypeDef.name = fptd;
                            newTypeDef.source_file = Path.ChangeExtension(func.source_file, "H");
                            foreach (string objectFile in func.object_files)
                            {
                                newTypeDef.object_files.Add(objectFile);
                            }
                            newTypeDef.c_type = newTypeName;

                            if (!result.FunctionPointerTypeDefs[func.overlay_id].Contains(newTypeDef))
                            {
                                result.FunctionPointerTypeDefs[func.overlay_id].Add(newTypeDef);
                            }
                            manualChangesBuilder.Append(String.Format("\t\t(The return type of the method)"));
                            manualChangesBuilder.Append(Environment.NewLine);
                        }
                        int paramNum = 0;
                        for (int i = 0; i < func.parameters.Count; i++)
                        {
                            CParameter p = func.parameters[i];
                            if (p.is_function_pointer)
                            {
                                //string fptd = GetFunctionPointerTypeDefFromParameter(p);
                                string fptd = p.parameter_signature;

                                string newTypeName = string.Format("{0}{1}{2}{3}", FunctionPointerTypeDefPrefix, func.name, paramNum, p.name);
                                string oldTypeName = p.type_name;
                                p.type_name = newTypeName;
                                string oldTypeNameWithVarName = p.parameter_signature;
                                int paramSigCommentPos = p.parameter_signature.LastIndexOf("/*");
                                string paramSigComment = "";
                                if (paramSigCommentPos > 0)
                                {
                                    paramSigComment = p.parameter_signature.Substring(paramSigCommentPos);
                                    oldTypeNameWithVarName = oldTypeNameWithVarName.Substring(0, paramSigCommentPos - 1).Trim();
                                }
                                string newTypeNameWithVarName = string.Format("{0} {1}", newTypeName, p.name);
                                p.parameter_signature = p.parameter_signature.Replace(oldTypeNameWithVarName, newTypeNameWithVarName);
                                func.method_signature = func.method_signature.Replace(oldTypeNameWithVarName, newTypeNameWithVarName);
                                if (!functionPointerNames.Contains(newTypeName))
                                {
                                    functionPointerNames.Add(newTypeName);
                                }
                                //fptd = string.Format("typedef {0};", oldTypeName);
                                fptd = oldTypeName;
                                fptd = fptd.Replace("*)(", string.Format("*{0})(", newTypeName));
                                

                                CTypeDef newTypeDef = new CTypeDef();
                                newTypeDef.typedef_signature = fptd;
                                newTypeDef.name = fptd;
                                newTypeDef.source_file = Path.ChangeExtension(func.source_file, "H");
                                foreach (string objectFile in func.object_files)
                                {
                                    newTypeDef.object_files.Add(objectFile);
                                }
                                newTypeDef.c_type = newTypeName;

                                if (!result.FunctionPointerTypeDefs[func.overlay_id].Contains(newTypeDef))
                                {
                                    result.FunctionPointerTypeDefs[func.overlay_id].Add(newTypeDef);
                                }
                                if (!functionPointerFunctionNames.Contains(p.name))
                                {
                                    functionPointerFunctionNames.Add(p.name);
                                }
                                if (!functionPointerNames.Contains(p.name))
                                {
                                    functionPointerNames.Add(p.name);
                                }

                                manualChangesBuilder.Append(String.Format("\t\t{0}", p.name));
                                manualChangesBuilder.Append(Environment.NewLine);
                            }
                            func.parameters[i] = p;
                            paramNum++;
                        }
                        manualChangesBuilder.Append(Environment.NewLine);
                        functionsWithFunctionPointerParams.Add(func);
                        GenerateFunctionSectionsResult functionPointerParamsFuncResult = GenerateFunctionSections(func, fpFunctionImportBuilder,
                            fpFunctionSubImportBuilder, fpFunctionSubSubImportBuilder, fpFunctionNum, fpSubRoutineNumber, fpSubroutineFunctionCount,
                            ghidraDefineFunctionsSubroutineTemplate, false, "FunctionPointerFunctions", assignReturnAddressStorage, assignParameterStorage, snGlobalPointerValue, assumeSNGPBase);
                        fpFunctionImportBuilder = functionPointerParamsFuncResult.FunctionImportBuilder;
                        fpFunctionSubImportBuilder = functionPointerParamsFuncResult.FunctionSubImportBuilder;
                        fpFunctionNum = functionPointerParamsFuncResult.FunctionNumber;
                        fpSubRoutineNumber = functionPointerParamsFuncResult.SubRoutineNumber;
                        fpSubroutineFunctionCount = functionPointerParamsFuncResult.SubroutineFunctionCount;
                        fpFunctionNum++;
                    }
                    else
                    {
                        GenerateFunctionSectionsResult noFunctionPointerParamsFuncResult = GenerateFunctionSections(func, functionImportBuilder,
                            functionSubImportBuilder, functionSubSubImportBuilder, functionNum, subRoutineNumber, subroutineFunctionCount,
                            ghidraDefineFunctionsSubroutineTemplate, false, "NormalFunctions", assignReturnAddressStorage, assignParameterStorage, snGlobalPointerValue, assumeSNGPBase);
                        functionImportBuilder = noFunctionPointerParamsFuncResult.FunctionImportBuilder;
                        functionSubImportBuilder = noFunctionPointerParamsFuncResult.FunctionSubImportBuilder;
                        functionNum = noFunctionPointerParamsFuncResult.FunctionNumber;
                        subRoutineNumber = noFunctionPointerParamsFuncResult.SubRoutineNumber;
                        subroutineFunctionCount = noFunctionPointerParamsFuncResult.SubroutineFunctionCount;
                        functionNum++;
                    }
                    //}
                }
                // disabled for debugging by using an exception that's never thrown
                catch (OutOfMemoryException ex)
                {
                    Console.WriteLine(String.Format("Error parsing method '{0}', signature '{1}': {2}", func.name, func.method_signature, ex.Message));
                }
            }

            // flush any remaining entries to the StringBuilders
            GenerateFunctionSectionsResult finalFunctionPointerParamsFuncResult = FlushLastEntriesForFunctionSections(fpFunctionImportBuilder,
                           fpFunctionSubImportBuilder, fpFunctionSubSubImportBuilder, fpFunctionNum, fpSubRoutineNumber, fpSubroutineFunctionCount,
                           ghidraDefineFunctionsSubroutineTemplate, "FunctionPointerFunctions");
            fpFunctionImportBuilder = finalFunctionPointerParamsFuncResult.FunctionImportBuilder;
            fpFunctionSubImportBuilder = finalFunctionPointerParamsFuncResult.FunctionSubImportBuilder;
            fpFunctionNum = finalFunctionPointerParamsFuncResult.FunctionNumber;
            fpSubRoutineNumber = finalFunctionPointerParamsFuncResult.SubRoutineNumber;
            fpSubroutineFunctionCount = finalFunctionPointerParamsFuncResult.SubroutineFunctionCount;
            fpFunctionNum++;

            GenerateFunctionSectionsResult finalNoFunctionPointerParamsFuncResult = FlushLastEntriesForFunctionSections(functionImportBuilder,
               functionSubImportBuilder, functionSubSubImportBuilder, functionNum, subRoutineNumber, subroutineFunctionCount,
               ghidraDefineFunctionsSubroutineTemplate, "NormalFunctions");
            functionImportBuilder = finalNoFunctionPointerParamsFuncResult.FunctionImportBuilder;
            functionSubImportBuilder = finalNoFunctionPointerParamsFuncResult.FunctionSubImportBuilder;
            functionNum = finalNoFunctionPointerParamsFuncResult.FunctionNumber;
            subRoutineNumber = finalNoFunctionPointerParamsFuncResult.SubRoutineNumber;
            subroutineFunctionCount = finalNoFunctionPointerParamsFuncResult.SubroutineFunctionCount;
            functionNum++;

            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%FUNCTION_IMPORT_SUBROUTINE_CALLS_1%", functionImportBuilder.ToString());

            functionImportBuilder.Clear();

            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%FUNCTIONS_WITH_FUNCTION_POINTER_PARAMETERS_CALLS%", fpFunctionImportBuilder.ToString());

            fpFunctionImportBuilder.Clear();

            StringBuilder functionPointerImportBuilder = new StringBuilder();
            List<string> addedFunctionPointers = new List<string>();
            foreach (string fp in functionPointerNames)
            {
                if (!addedFunctionPointers.Contains(fp))
                {
                    functionPointerImportBuilder.Append(String.Format("\t\tdataTypeMap = addDataTypeByName(dataTypeMap, \"/{0}.H/{1}\", \"{1}\", \"\");", projectName, BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(fp)));
                    functionPointerImportBuilder.Append(Environment.NewLine);
                    addedFunctionPointers.Add(fp);
                }
            }
            //
            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%FUNCTION_POINTER_IMPORTS%", functionPointerImportBuilder.ToString());

            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%FUNCTION_IMPORT_SUBROUTINES%", functionSubImportBuilder.ToString());
            templatesForStringReplacement = ReplacePlaceholderInTemplates(templatesForStringReplacement, "%FUNCTION_POINTER_FUNCTION_IMPORT_SUBROUTINES%", fpFunctionSubImportBuilder.ToString());

            BenLincoln.DebugSymbols.Utilities.WriteToOutput(ghidraMapMemoryAndLabelsScriptPath, templatesForStringReplacement["ghidraMemoryMapAndLabelScriptTemplate"], FileMode.Create);
            BenLincoln.DebugSymbols.Utilities.WriteToOutput(ghidraDefineFunctionsScriptPath, templatesForStringReplacement["ghidraFunctionDefinitionScriptTemplate"], FileMode.Create);
            BenLincoln.DebugSymbols.Utilities.WriteToOutput(ghidraArrayIdentificationScriptPath, templatesForStringReplacement["ghidraArrayIdentificationTemplate"], FileMode.Create);
            BenLincoln.DebugSymbols.Utilities.WriteToOutput(ghidraCodeExportScriptPath, templatesForStringReplacement["ghidraCustomDecompileExportTemplate"], FileMode.Create);
            BenLincoln.DebugSymbols.Utilities.WriteToOutput(ghidraDataExportScriptPath, templatesForStringReplacement["ghidraDataExportTemplate"], FileMode.Create);

            // create Ghidra debug symbol XML file even though this doesn't work right now
            BenLincoln.DebugSymbols.Utilities.WriteToOutput(ghidraXMLFilePath, ghidraXMLSymbols, FileMode.Create);

            foreach (string enumName in addedEnums)
            {
                foreach (string structName in addedStructs)
                {
                    if (enumName == structName)
                    {
                        manualChangesBuilder.Append(string.Format("The name '{0}' is used to identify both an enum and a struct. You should manually rename one or both in the JSON file to resolve the conflict, then re-run CreateSkeleton.", enumName));
                        manualChangesBuilder.Append(Environment.NewLine);
                        result.EnumStructDupeNames.Add(enumName);
                    }
                }
                foreach (string unionName in addedUnions)
                {
                    if (enumName == unionName)
                    {
                        manualChangesBuilder.Append(string.Format("The name '{0}' is used to identify both an enum and a union. You should manually rename one or both in the JSON file to resolve the conflict, then re-run CreateSkeleton.", enumName));
                        manualChangesBuilder.Append(Environment.NewLine);
                        result.EnumUnionDupeNames.Add(enumName);
                    }
                }
            }
            foreach (string structName in addedStructs)
            {
                foreach (string unionName in addedUnions)
                {
                    if (structName == unionName)
                    {
                        manualChangesBuilder.Append(string.Format("The name '{0}' is used to identify both an struct and a union. You should manually rename one or both in the JSON file to resolve the conflict, then re-run CreateSkeleton.", structName));
                        manualChangesBuilder.Append(Environment.NewLine);
                        result.StructUnionDupeNames.Add(structName);
                    }
                }
            }

            DupeCounter dc = new DupeCounter();
            foreach (CEnum e in enums)
            {
                dc.HandleName(e.name);
            }
            manualChangesBuilder.Append(dc.GetResultOutput("enum"));
            result.DuplicateEnumNames = dc.GetDuplicateNames();
            dc = new DupeCounter();

            foreach (CStruct s in structs)
            {
                dc.HandleName(s.name);
            }
            manualChangesBuilder.Append(dc.GetResultOutput("struct"));
            result.DuplicateStructNames = dc.GetDuplicateNames();
            dc = new DupeCounter();

            foreach (CUnion u in unions)
            {
                dc.HandleName(u.name);
            }

            manualChangesBuilder.Append(dc.GetResultOutput("union"));
            result.DuplicateUnionNames = dc.GetDuplicateNames();

            // create the master list of all typedefs
            foreach (long olID in result.FunctionPointerTypeDefs.Keys)
            {
                foreach (CTypeDef t in result.FunctionPointerTypeDefs[olID])
                {
                    if (!result.AllFunctionPointerTypeDefs.Contains(t))
                    {
                        result.AllFunctionPointerTypeDefs.Add(t);
                    }
                }
            }

            MonolithicHeaderGenerationResult mhgr = SourceCodeGenerator.GenerateMonolithicHeader(enums, structs, unions, typedefs, 
                result.AllFunctionPointerTypeDefs, externs, projectName,
                manualChangesBuilder);
            manualChangesBuilder = mhgr.ManualChangesBuilder;
            BenLincoln.DebugSymbols.Utilities.WriteToOutput(monolithicHeaderFilePath, mhgr.SourceCode, FileMode.Create);

            // create a list of known manual actions that need to be taken by the user
            manualChangesBuilder.Append(Environment.NewLine);
            result.ManualChangesBuilder = manualChangesBuilder;
            
            return result;
        }

        protected static string GetFunctionPointerTypeDefFromParameter(CParameter param)
        {
            //string pTypeName = param.type_name;
            //if (pTypeName.EndsWith(" *"))
            //{
            //    pTypeName = pTypeName.Substring(0, pTypeName.Length - 2);
            //}
            //string result = string.Format("typedef {0} (*{1})();", pTypeName, param.name);
            string result = string.Format("typedef {0};", param.parameter_signature);
            return result;
        }

        protected static string GetFunctionPointerRefInGhidraFormatFromParameter(CParameter param)
        {
            string result = string.Format("{0} *", param.name);
            return result;
        }

        protected struct GenerateFunctionSectionsResult
        {
            public StringBuilder FunctionImportBuilder;
            public StringBuilder FunctionSubImportBuilder;
            public int FunctionNumber;
            public int SubRoutineNumber;
            public int SubroutineFunctionCount;
        }

        protected static GenerateFunctionSectionsResult GenerateFunctionSections(CFunction func, StringBuilder functionImportBuilder,
            StringBuilder functionSubImportBuilder, StringBuilder functionSubSubImportBuilder, int functionNum, int subRoutineNumber, int subroutineFunctionCount,
            string ghidraDefineFunctionsSubroutineTemplate, bool replaceFunctionPointersWithVoidPointers, string prefix, bool assignReturnAddressStorage, bool assignParameterStorage, long snGPValue, bool assumeSNGPValue)
        {
            StringBuilder subBuilder = new StringBuilder();
            string functionSuffix = String.Format("{0}Function{1:X8}", prefix, functionNum);
            subBuilder.Append(String.Format("\t\t\t// 0x{0:X8}: {1}", func.offset_in_memory, func.method_signature));
            subBuilder.Append(Environment.NewLine);

            string returnTypeName = func.return_value.type_name;

            string returnValueVarName = string.Format("returnVal{0}", functionSuffix);
            string parameterArrayVarName = string.Format("functionParams{0}", functionSuffix);
            string returnValIsRegisterParam = "false";
            if ((func.return_value.register_name != null) && (func.return_value.register_name != ""))
            {
                returnValIsRegisterParam = "true";
            }

            string returnValueTypeNameVar = string.Format("returnType{0}", functionSuffix);
            subBuilder.Append(String.Format("\t\t\tDataType {0} = GetDataTypeFromMap(dataTypeMap, \"{1}\");", returnValueTypeNameVar, BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(returnTypeName)));
            subBuilder.Append(Environment.NewLine);
            subBuilder.Append(String.Format("\t\t\tFunctionParameter {0} = new FunctionParameter(\"{1}\", {2}, \"{3}\", {4}, {5});", returnValueVarName, 
                BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(func.return_value.name), 
                returnValueTypeNameVar, BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(CFunction.GetGhidrafiedRegisterName(func.return_value)), 
                func.return_value.stack_offset, returnValIsRegisterParam));
            subBuilder.Append(Environment.NewLine);
            subBuilder.Append(String.Format("\t\t\tFunctionParameter[] {0} = new FunctionParameter[{1}];", parameterArrayVarName, func.parameters.Count));
            subBuilder.Append(Environment.NewLine);
            for (int i = 0; i < func.parameters.Count; i++)
            {
                string isRegisterParam = "false";
                if ((func.parameters[i].register_name != null) && (func.parameters[i].register_name != ""))
                {
                    isRegisterParam = "true";
                }
                string paramTypeName = func.parameters[i].type_name;
                string parameterName = func.parameters[i].name;
                string paramTypeVar = string.Format("paramType{0}{1}", functionSuffix, i);
                List<int> arraySizes = new List<int>();
                if (paramTypeName.Contains("["))
                {
                    if (paramTypeName.Contains("])"))
                    {
                        Console.WriteLine(string.Format("Error: the parameter '{0}' in function '{1}' is too complex for this version of TDR to handle automatically. You will need to manually define this function parameter in Ghidra.", func.parameters[i].parameter_signature, func.method_signature));
                    }
                    else
                    {
                        Regex arrayRegex = new Regex("\\[(?<arraySize>[0-9]+)\\]");
                        MatchCollection arrayMatches = arrayRegex.Matches(paramTypeName);
                        foreach (Match m in arrayMatches)
                        {
                            if (m.Groups["arraySize"] != null)
                            {
                                try
                                {
                                    int sz = int.Parse(m.Groups["arraySize"].Value);
                                    arraySizes.Add(sz);
                                }
                                catch (Exception ex)
                                {
                                    Console.WriteLine(string.Format("Error: could not parse array size '{0}' from parameter type '{1}' in function with signature '{2}'", m.Groups["arraySize"].Value, paramTypeName, func.method_signature));
                                }
                            }
                        }
                        paramTypeName = paramTypeName.Substring(0, paramTypeName.IndexOf("["));
                        // otherwise multidimensional arrays will be backwards
                        arraySizes.Reverse();
                    }
                }
                string originalParamTypeName = paramTypeName;
                bool handledParamType = false;
                if (paramTypeName.EndsWith("(*)"))
                {
                    // pointer to something (probably an array)
                    paramTypeName = paramTypeName.Substring(0, paramTypeName.Length - 3).Trim();
                    subBuilder.Append(String.Format("\t\t\tDataType {0} = GetDataTypeFromMap(dataTypeMap, \"{1}\");", paramTypeVar, 
                        BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(paramTypeName)));
                    subBuilder.Append(Environment.NewLine);
                    GetArrayWrapperResult awr = GetArrayWrapper(subBuilder, arraySizes, paramTypeVar);
                    subBuilder = awr.SubBuilder;
                    paramTypeVar = awr.NewOuterVariableName;
                    string pointerTypeName = string.Format("{0}Pointer", paramTypeVar);
                    subBuilder.Append(String.Format("\t\t\tPointerDataType {0} = new PointerDataType({1});", pointerTypeName, paramTypeVar));
                    subBuilder.Append(Environment.NewLine);
                    paramTypeVar = pointerTypeName;
                    handledParamType = true;
                }
                if ((!handledParamType) && (paramTypeName.EndsWith("*")))
                {
                    // it's either a pointer to a single thing, or an array of pointers to something else
                    if (arraySizes.Count == 0)
                    {
                        // pointer to something
                        subBuilder.Append(String.Format("\t\t\tDataType {0} = GetDataTypeFromMap(dataTypeMap, \"{1}\");", paramTypeVar, 
                            BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(paramTypeName)));
                        subBuilder.Append(Environment.NewLine);
                    }
                    else
                    {
                        // array of pointers
                        paramTypeName = paramTypeName.Substring(0, paramTypeName.Length - 1).Trim();
                        subBuilder.Append(String.Format("\t\t\tDataType {0} = GetDataTypeFromMap(dataTypeMap, \"{1}\");", paramTypeVar, 
                            BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(paramTypeName)));
                        subBuilder.Append(Environment.NewLine);
                        string pointerTypeName = string.Format("{0}Pointer", paramTypeVar);
                        subBuilder.Append(String.Format("\t\t\tPointerDataType {0} = new PointerDataType({1});", pointerTypeName, paramTypeVar));
                        subBuilder.Append(Environment.NewLine);
                        GetArrayWrapperResult awr = GetArrayWrapper(subBuilder, arraySizes, paramTypeVar);
                        subBuilder = awr.SubBuilder;
                        paramTypeVar = awr.NewOuterVariableName;
                        paramTypeVar = pointerTypeName;
                    }
                    handledParamType = true;
                }
                if (!handledParamType)
                {
                    subBuilder.Append(String.Format("\t\t\tDataType {0} = GetDataTypeFromMap(dataTypeMap, \"{1}\");", paramTypeVar, 
                        BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(paramTypeName)));
                    subBuilder.Append(Environment.NewLine);
                }
                subBuilder.Append(String.Format("\t\t\t{0}[{1}] = new FunctionParameter(\"{2}\", {3}, \"{4}\", {5}, {6});", parameterArrayVarName, i, 
                    BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(parameterName), paramTypeVar,
                    BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(CFunction.GetGhidrafiedRegisterName(func.parameters[i])), 
                    func.parameters[i].stack_offset, isRegisterParam));
                subBuilder.Append(Environment.NewLine);
            }
            string assignRAStorage = "false";
            if (assignReturnAddressStorage)
            {
                assignRAStorage = "true";
            }
            string assignAnyStorage = "false";
            if (assignParameterStorage)
            {
                assignAnyStorage = "true";
            }
            subBuilder.Append(String.Format("\t\t\tDefineFunction2(dataTypeMap, 0x{0:X8}L, \"{1}\", {2}, {3}, {4}, {5});", func.offset_in_memory, 
                BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(func.name), returnValueVarName, parameterArrayVarName, assignRAStorage, assignAnyStorage));
            subBuilder.Append(Environment.NewLine);
            if (assumeSNGPValue)
            {
                subBuilder.Append(String.Format("\t\t\tSetRegisterAtAddress(ctx, \"gp\", 0x{0:X8}L, 0x{1:X8}L);", func.offset_in_memory, snGPValue));
                subBuilder.Append(Environment.NewLine);
            }
            subBuilder.Append(Environment.NewLine);

            // only add this block of lines if the entire set of operations for this function succeeded
            functionSubSubImportBuilder.Append(subBuilder.ToString());
            functionNum++;
            subroutineFunctionCount++;
            if (subroutineFunctionCount > MaxFunctionDefinitionsPerGhidraScriptSubroutine)
            {
                string subDefinition = ghidraDefineFunctionsSubroutineTemplate.Replace("%FUNCTION_IMPORT_SUB_CONTENT%", functionSubSubImportBuilder.ToString());
                functionSubSubImportBuilder.Clear();
                subDefinition = subDefinition.Replace("%PREFIX%", prefix);
                subDefinition = subDefinition.Replace("%DEFINE_FUNCTIONS_SUBROUTINE_NUMBER%", String.Format("{0}", subRoutineNumber));
                functionSubImportBuilder.Append(subDefinition);
                functionImportBuilder.Append(String.Format("\t\t{0}DefineFunctionsSub{1}(dataTypeMap, ctx);", prefix, subRoutineNumber));
                functionImportBuilder.Append(Environment.NewLine);
                subRoutineNumber++;
                subroutineFunctionCount = 0;
            }

            GenerateFunctionSectionsResult result = new GenerateFunctionSectionsResult();
            result.FunctionImportBuilder = functionImportBuilder;
            result.FunctionSubImportBuilder = functionSubImportBuilder;
            result.FunctionNumber = functionNum;
            result.SubRoutineNumber = subRoutineNumber;
            result.SubroutineFunctionCount = subroutineFunctionCount;

            return result;
        }

        protected static GenerateFunctionSectionsResult FlushLastEntriesForFunctionSections(StringBuilder functionImportBuilder,
            StringBuilder functionSubImportBuilder, StringBuilder functionSubSubImportBuilder, int functionNum, int subRoutineNumber, int subroutineFunctionCount,
            string ghidraDefineFunctionsSubroutineTemplate, string prefix)
        {
            StringBuilder subBuilder = new StringBuilder();
            string functionSuffix = String.Format("{0}Function{1:X8}", prefix, functionNum);

            string subDefinition = ghidraDefineFunctionsSubroutineTemplate.Replace("%FUNCTION_IMPORT_SUB_CONTENT%", functionSubSubImportBuilder.ToString());
            functionSubSubImportBuilder.Clear();
            subDefinition = subDefinition.Replace("%PREFIX%", prefix);
            subDefinition = subDefinition.Replace("%DEFINE_FUNCTIONS_SUBROUTINE_NUMBER%", String.Format("{0}", subRoutineNumber));
            functionSubImportBuilder.Append(subDefinition);
            functionImportBuilder.Append(String.Format("\t\t{0}DefineFunctionsSub{1}(dataTypeMap, ctx);", prefix, subRoutineNumber));
            functionImportBuilder.Append(Environment.NewLine);
            subRoutineNumber++;
            subroutineFunctionCount = 0;

            GenerateFunctionSectionsResult result = new GenerateFunctionSectionsResult();
            result.FunctionImportBuilder = functionImportBuilder;
            result.FunctionSubImportBuilder = functionSubImportBuilder;
            result.FunctionNumber = functionNum;
            result.SubRoutineNumber = subRoutineNumber;
            result.SubroutineFunctionCount = subroutineFunctionCount;

            return result;
        }

        public struct GetArrayWrapperResult
        {
            public StringBuilder SubBuilder;
            public string NewOuterVariableName;
        }

        protected static GetArrayWrapperResult GetArrayWrapper(StringBuilder subBuilder, List<int> arraySizes, string paramTypeVar)
        {
            if (arraySizes.Count > 0)
            {
                string baseTypeName = paramTypeVar;
                for (int arrayNum = 0; arrayNum < arraySizes.Count; arrayNum++)
                {
                    string currentArrayTypeName = paramTypeVar;
                    if (arrayNum > 0)
                    {
                        currentArrayTypeName = string.Format("{0}Array{1}", baseTypeName, arrayNum - 1);
                    }
                    string newArrayTypeName = string.Format("{0}Array{1}", baseTypeName, arrayNum);
                    subBuilder.Append(String.Format("\t\t\tArrayDataType {0} = new ArrayDataType({1}, {2}, {3}.getLength());", newArrayTypeName, currentArrayTypeName, arraySizes[arrayNum], baseTypeName));
                    subBuilder.Append(Environment.NewLine);
                    paramTypeVar = newArrayTypeName;
                }
            }

            GetArrayWrapperResult result = new GetArrayWrapperResult();
            result.SubBuilder = subBuilder;
            result.NewOuterVariableName = paramTypeVar;
            return result;
        }
    }
}
