﻿// PopulateSkeleton
// Populates a skeleton PlayStation dev project based on the output of SymDumpTE and CreateSkeleton
// Copyright 2019 Ben Lincoln
// https://www.beneaththewaves.net/

// This file is part of PopulateSkeleton.

// PopulateSkeleton is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// PopulateSkeleton is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with PopulateSkeleton (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using Newtonsoft.Json;
using BenLincoln.DebugSymbols;
using BenLincoln.DebugSymbols.SymDump;

namespace PopulateSkeleton
{
    public struct DecompiledSourceCodeMatch
    {
        public CFunction DecompiledFunction;
        public float MatchPercent;
    }

    class Program
    {
        protected const string ProgramName = "PopulateSkeleton";
        protected static string Copyright = "Copyright 2019 Ben Lincoln";
        protected const string InfoURL = "https://www.beneaththewaves.net/Software/This_Dust_Remembers_What_It_Once_Was.html";
        protected static string[] ThanksList = new string[] {
            "The United States National Security Agency for Ghidra (https://www.ghidra-sre.org/)",
            "Gh0stBlade (https://github.com/Gh0stBlade)",
            "Andrew Fradley (https://github.com/afradley)",
            "Vladimir Kononovich (https://github.com/lab313ru)",
            };
        protected const string HelpMessageFile = "PopulateSkeleton.HelpMessage.txt";
        protected const string ProjectNameValidationRegexString = @"^[A-Za-z0-9_]{1,16}$";
        protected const string ProjectNameExplanation = "alphanumeric characters and underscore (_)";

        static void Main(string[] args)
        {
            Assembly asm = Assembly.GetExecutingAssembly();

            if ((args.Length < 5))
            {
                BenLincoln.DebugSymbols.Utilities.DisplayHelpMessage(asm, ProgramName, Copyright, InfoURL, ThanksList, HelpMessageFile);
                Environment.Exit(1);
            }

            string projectName = "";
            string baseOutputDirectoryPath = "";
            string inputJSONFilePath = "";
            string inputSourceCodeFilePath = "";
            string inputExportedDataCodeFilePath = "";
            List<long> overlayIDs = new List<long>();
            bool forceOverwrite = false;

            ArrayList arguments = new ArrayList();
            for (int i = 0; i < args.Length; i++)
            {
                // if the argument is one with a second component, catch that now
                string nextArg = "";
                switch (args[i])
                {
                    case "--name":
                        nextArg = args[i + 1];
                        projectName = nextArg;
                        Regex pnRex = new Regex(ProjectNameValidationRegexString);
                        if (!pnRex.IsMatch(projectName))
                        {
                            BenLincoln.DebugSymbols.Utilities.ExitWithError(String.Format("The specified project name is invalid. The name must consist only of {0}.", 
                                ProjectNameExplanation), 1, asm, ProgramName, Copyright, InfoURL, ThanksList);
                        }
                        i++;
                        break;
                    case "--input-json":
                        nextArg = args[i + 1];
                        inputJSONFilePath = Path.GetFullPath(nextArg);
                        i++;
                        break;
                    case "--input-source":
                        nextArg = args[i + 1];
                        inputSourceCodeFilePath = Path.GetFullPath(nextArg);
                        i++;
                        break;
                    case "--input-data":
                        nextArg = args[i + 1];
                        inputExportedDataCodeFilePath = Path.GetFullPath(nextArg);
                        i++;
                        break;
                    case "--output":
                        nextArg = args[i + 1];
                        baseOutputDirectoryPath = Path.GetFullPath(nextArg);
                        i++;
                        break;
                    case "--overlay":
                        nextArg = args[i + 1];
                        try
                        {
                            long newOverlayID = BenLincoln.DebugSymbols.Utilities.GetHexOrDecimalLong(nextArg);
                            if (!overlayIDs.Contains(newOverlayID))
                            {
                                overlayIDs.Add(newOverlayID);
                            }
                        }
                        catch (Exception ex)
                        {
                            BenLincoln.DebugSymbols.Utilities.ExitWithError(ex.Message, 1, asm, ProgramName, Copyright, InfoURL, ThanksList);
                        }
                        i++;
                        break;
                    default:
                        arguments.Add(args[i]);
                        break;
                }
            }

            ArrayList positionalArgs = new ArrayList();

            foreach (string arg in arguments)
            {
                if ((arg == "-?") || (arg == "-h") || (arg == "--help"))
                {
                    BenLincoln.DebugSymbols.Utilities.DisplayHelpMessage(asm, ProgramName, Copyright, InfoURL, ThanksList, HelpMessageFile);
                    Environment.Exit(0);
                }
                bool handled = false;
                if (arg == "--overwrite")
                {
                    forceOverwrite = true;
                    handled = true;
                }
                // right now, all positional arguments are file paths
                if (!handled)
                {
                    positionalArgs.Add(arg);
                }
            }

            if (inputJSONFilePath == "")
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("No input SymDump file path was specified.", 
                    1, asm, ProgramName, Copyright, InfoURL, ThanksList);
            }

            if (!File.Exists(inputJSONFilePath))
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("The specified input SymDump file does not exist.", 
                    1, asm, ProgramName, Copyright, InfoURL, ThanksList);
            }

            if (inputSourceCodeFilePath == "")
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("No input decompiled source code file path was specified.", 
                    1, asm, ProgramName, Copyright, InfoURL, ThanksList);
            }

            if (!File.Exists(inputSourceCodeFilePath))
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("The specified input decompiled source code file does not exist.", 
                    1, asm, ProgramName, Copyright, InfoURL, ThanksList);
            }

            if (inputExportedDataCodeFilePath != "")
            {
                if (!File.Exists(inputExportedDataCodeFilePath))
                {
                    BenLincoln.DebugSymbols.Utilities.ExitWithError("The specified input global-variables-as-data source code file does not exist.",
                        1, asm, ProgramName, Copyright, InfoURL, ThanksList);
                }
            }

            if (baseOutputDirectoryPath == "")
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError("No output directory was specified.", 
                    1, asm, ProgramName, Copyright, InfoURL, ThanksList);
            }
            if (forceOverwrite)
            {
                if (Directory.Exists(baseOutputDirectoryPath))
                {
                    try
                    {
                        Directory.Delete(baseOutputDirectoryPath, true);
                    }
                    catch (Exception ex)
                    {
                        BenLincoln.DebugSymbols.Utilities.ExitWithError(String.Format("The output directory '{0}' already exists, and could not be deleted: {1}", 
                            baseOutputDirectoryPath, ex.Message), 2, asm, ProgramName, Copyright, InfoURL, ThanksList);
                    }
                }
            }
            /*
            else
            {
                if (Directory.Exists(baseOutputDirectoryPath))
                {
                    ExitWithError("The specified output directory already exists. Use the --overwrite option to replace it.", 1);
                }
            }
            */

            StringBuilder manualChangesBuilder = new StringBuilder();

            // create base output directory structure
            //string[] outputDirs = new string[] { "source-decompiled" };
            //BenLincoln.DebugSymbols.Utilities.CreateOutputDirectories(baseOutputDirectoryPath, outputDirs, Assembly.GetExecutingAssembly(), ProgramName, Copyright, InfoURL, ThanksList);

            string manualChangesPath = Path.Combine(baseOutputDirectoryPath, String.Format("README-{0}-PopulateSkeleton-Manual_Changes_Required.txt", projectName));

            // get SymDumpTE output
            string symDumpRaw = "";
            try
            {
                symDumpRaw = BenLincoln.DebugSymbols.Utilities.GetFileAsString(inputJSONFilePath);
            }
            catch (Exception ex)
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError(String.Format("Could not read the input file '{0}': {1}", inputJSONFilePath, ex.Message), 
                    2, asm, ProgramName, Copyright, InfoURL, ThanksList);
            }

            // get decompiled source
            string decompiledSource = "";
            try
            {
                decompiledSource = BenLincoln.DebugSymbols.Utilities.GetFileAsString(inputSourceCodeFilePath);
            }
            catch (Exception ex)
            {
                BenLincoln.DebugSymbols.Utilities.ExitWithError(String.Format("Could not read the input source code file '{0}': {1}", inputSourceCodeFilePath, ex.Message), 
                    2, asm, ProgramName, Copyright, InfoURL, ThanksList);
            }

            // get exported data
            string exportedDataRaw = "";
            if (inputExportedDataCodeFilePath != "")
            {
                try
                {
                    exportedDataRaw = BenLincoln.DebugSymbols.Utilities.GetFileAsString(inputExportedDataCodeFilePath);
                }
                catch (Exception ex)
                {
                    BenLincoln.DebugSymbols.Utilities.ExitWithError(String.Format("Could not read the input exported data file '{0}': {1}", inputSourceCodeFilePath, ex.Message), 
                        2, asm, ProgramName, Copyright, InfoURL, ThanksList);
                }
            }

            RenameJSONEntriesForGhidraResult rnjr = BenLincoln.DebugSymbols.GhidraScriptGenerator.RenameJSONEntriesForGhidra(symDumpRaw, manualChangesBuilder);
            manualChangesBuilder = rnjr.ManualChangesBuilder;
            symDumpRaw = rnjr.JSONText;

            SymDumpSymbolCollection symbolCollection = JsonConvert.DeserializeObject<SymDumpSymbolCollection>(symDumpRaw);

            List<PsyQOverlayDescription> includedOverlays = BenLincoln.DebugSymbols.SourceCodeGenerator.GetOverlayList(symbolCollection, overlayIDs,
                asm, ProgramName, Copyright, InfoURL, ThanksList);

            BenLincoln.DebugSymbols.SourceCodeSourceFileTracker codeSourceTracker = new BenLincoln.DebugSymbols.SourceCodeSourceFileTracker();

            //foreach (CFunction func in symbolCollection.functions)
            //{
            //    try
            //    {
            //        codeSourceTracker.AddOrUpdateEntry(func);
            //    }
            //    // disabled for debugging by using an exception that's never thrown
            //    catch (OutOfMemoryException ex)
            //    {
            //        Console.WriteLine(String.Format("Error parsing method '{0}', signature '{1}': {2}", func.name, func.method_signature, ex.Message));
            //    }
            //}

            List<CFunction> decompiledFunctions = BenLincoln.DebugSymbols.SourceCodeParser.GetAllFunctionsFromSource(decompiledSource);
            List<string> matchedFunctionGUIDs = new List<string>();
            List<CExternalDeclaration> exportedData = BenLincoln.DebugSymbols.SourceCodeParser.GetExportedData(exportedDataRaw);
            List<string> matchedExternGUIDs = new List<string>();

            for (int funcNum = 0; funcNum < decompiledFunctions.Count; funcNum++)
            {
                bool foundFunction = false;
                for (int existingFuncNum = 0; existingFuncNum < symbolCollection.functions.Count; existingFuncNum++)
                {
                    if (decompiledFunctions[funcNum].offset_in_memory == symbolCollection.functions[existingFuncNum].offset_in_memory)
                    {
                        Console.WriteLine(string.Format("Debug: found function '{0}' @0x{1:X8} in the original debug symbol data, and matched it to the decompiled code exported from Ghidra.",
                            symbolCollection.functions[existingFuncNum].name, symbolCollection.functions[existingFuncNum].offset_in_memory));
                        symbolCollection.functions[existingFuncNum].source_code = decompiledFunctions[funcNum].source_code;
                        matchedFunctionGUIDs.Add(symbolCollection.functions[existingFuncNum].guid.ToString());
                        matchedFunctionGUIDs.Add(decompiledFunctions[funcNum].guid.ToString());
                        foundFunction = true;                        
                        //break;
                    }
                }
                if (!foundFunction)
                {
                    for (int overlayNum = 0; overlayNum < symbolCollection.overlays.Count; overlayNum++)
                    {
                        if ((!foundFunction) && (overlayIDs.Contains(symbolCollection.overlays[overlayNum].overlay_id)))
                        {
                            for (int existingFuncNum = 0; existingFuncNum < symbolCollection.overlays[overlayNum].functions.Count; existingFuncNum++)
                            {
                                if (decompiledFunctions[funcNum].offset_in_memory == symbolCollection.overlays[overlayNum].functions[existingFuncNum].offset_in_memory)
                                {
                                    Console.WriteLine(string.Format("Debug: found function '{0}' @0x{1:X8} in overlay 0x{2:X2} of the original debug symbol data, and matched it to the decompiled code exported from Ghidra.",
                                        symbolCollection.overlays[overlayNum].functions[existingFuncNum].name,
                                        symbolCollection.overlays[overlayNum].functions[existingFuncNum].offset_in_memory,
                                        symbolCollection.overlays[overlayNum].overlay_id));
                                    symbolCollection.overlays[overlayNum].functions[existingFuncNum].source_code = decompiledFunctions[funcNum].source_code;
                                    matchedFunctionGUIDs.Add(symbolCollection.overlays[overlayNum].functions[existingFuncNum].guid.ToString());
                                    matchedFunctionGUIDs.Add(decompiledFunctions[funcNum].guid.ToString());
                                    foundFunction = true;
                                    //break;
                                }
                            }
                        }
                    }
                }
                if (!foundFunction)
                {
                    decompiledFunctions[funcNum].source_code = string.Format("// function '{0}' @0x{1:X8} was found in the decompiled code exported from Ghidra, but not the original debug symbol data.{2}{3}",
                        decompiledFunctions[funcNum].name, decompiledFunctions[funcNum].offset_in_memory, Environment.NewLine, decompiledFunctions[funcNum].source_code);
                    if (decompiledFunctions[funcNum].source_file == "")
                    {
                        decompiledFunctions[funcNum].source_file = BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedSourceCodeFileName(BenLincoln.DebugSymbols.SourceCodeGenerator.DefaultOrUnmappedFilename);
                    }
                    symbolCollection.functions.Add(decompiledFunctions[funcNum]);
                    matchedFunctionGUIDs.Add(decompiledFunctions[funcNum].guid.ToString());
                }
            }

            for (int existingFuncNum = 0; existingFuncNum < symbolCollection.functions.Count; existingFuncNum++)
            {
                if (!matchedFunctionGUIDs.Contains(symbolCollection.functions[existingFuncNum].guid.ToString()))
                {
                    Console.WriteLine(string.Format("Warning: function '{0}' @0x{1:X8} was part of the original debug symbol data, but was not found in the decompiled code exported from Ghidra.",
                        symbolCollection.functions[existingFuncNum].name, symbolCollection.functions[existingFuncNum].offset_in_memory));
                }
            }
            for (int overlayNum = 0; overlayNum < symbolCollection.overlays.Count; overlayNum++)
            {
                if (overlayIDs.Contains(symbolCollection.overlays[overlayNum].overlay_id))
                {
                    for (int existingFuncNum = 0; existingFuncNum < symbolCollection.overlays[overlayNum].functions.Count; existingFuncNum++)
                    {
                        if (!matchedFunctionGUIDs.Contains(symbolCollection.overlays[overlayNum].functions[existingFuncNum].guid.ToString()))
                        {
                            Console.WriteLine(string.Format("Warning: function '{0}' @0x{1:X8} was part of the original debug symbol data in overlay 0x{2:X2}, but was not found in the decompiled code exported from Ghidra.",
                                symbolCollection.overlays[overlayNum].functions[existingFuncNum].name, 
                                symbolCollection.overlays[overlayNum].functions[existingFuncNum].offset_in_memory,
                                symbolCollection.overlays[overlayNum].overlay_id));
                        }
                    }
                }
            }

            for (int externNum = 0; externNum < exportedData.Count; externNum++)
            {
                bool foundExtern = false;
                for (int existingExternNum = 0; existingExternNum < symbolCollection.external_declarations.Count; existingExternNum++)
                {
                    if (exportedData[externNum].offset_in_memory == symbolCollection.external_declarations[existingExternNum].offset_in_memory)
                    {
                        Console.WriteLine(string.Format("Debug: found external declaration '{0}' @0x{1:X8} in the original debug symbol data, and matched it to the decompiled code exported from Ghidra.",
                            symbolCollection.external_declarations[existingExternNum].name, symbolCollection.external_declarations[existingExternNum].offset_in_memory));
                        symbolCollection.external_declarations[existingExternNum].source_code = exportedData[externNum].source_code;
                        matchedExternGUIDs.Add(symbolCollection.external_declarations[existingExternNum].guid.ToString());
                        matchedExternGUIDs.Add(exportedData[externNum].guid.ToString());
                        foundExtern = true;
                        //break;
                    }
                }
                if (!foundExtern)
                {
                    for (int overlayNum = 0; overlayNum < symbolCollection.overlays.Count; overlayNum++)
                    {
                        if ((!foundExtern) && (overlayIDs.Contains(symbolCollection.overlays[overlayNum].overlay_id)))
                        {
                            for (int existingExternNum = 0; existingExternNum < symbolCollection.overlays[overlayNum].external_declarations.Count; existingExternNum++)
                            {
                                if (exportedData[externNum].offset_in_memory == symbolCollection.overlays[overlayNum].external_declarations[existingExternNum].offset_in_memory)
                                {
                                    Console.WriteLine(string.Format("Debug: found external declaration '{0}' @0x{1:X8} in overlay 0x{2:X2} of the original debug symbol data, and matched it to the decompiled code exported from Ghidra.",
                                        symbolCollection.overlays[overlayNum].external_declarations[existingExternNum].name,
                                        symbolCollection.overlays[overlayNum].external_declarations[existingExternNum].offset_in_memory,
                                        symbolCollection.overlays[overlayNum].overlay_id));
                                    symbolCollection.overlays[overlayNum].external_declarations[existingExternNum].source_code = exportedData[externNum].source_code;
                                    matchedExternGUIDs.Add(symbolCollection.overlays[overlayNum].external_declarations[existingExternNum].guid.ToString());
                                    matchedExternGUIDs.Add(exportedData[externNum].guid.ToString());
                                    foundExtern = true;
                                    //break;
                                }
                            }
                        }
                    }
                }
                if (!foundExtern)
                {
                    string warningPrefix = string.Format("// external declaration '{0}' @0x{1:X8} was found in the data exported from Ghidra, but not the original debug symbol data.{2}",
                        exportedData[externNum].name, exportedData[externNum].offset_in_memory, Environment.NewLine);
                    exportedData[externNum].declaration = string.Format("{0}{1}", warningPrefix, exportedData[externNum].declaration);
                    exportedData[externNum].source_code = string.Format("{0}{1}", warningPrefix, exportedData[externNum].source_code);
                    if (exportedData[externNum].source_file == "")
                    {
                        exportedData[externNum].source_file = BenLincoln.DebugSymbols.SourceCodeGenerator.GetUnmappedSourceCodeFileName(BenLincoln.DebugSymbols.SourceCodeGenerator.DefaultOrUnmappedFilename);
                    }
                    symbolCollection.external_declarations.Add(exportedData[externNum]);
                    matchedExternGUIDs.Add(exportedData[externNum].guid.ToString());
                }
            }

            for (int existingExternNum = 0; existingExternNum < symbolCollection.external_declarations.Count; existingExternNum++)
            {
                if (!matchedExternGUIDs.Contains(symbolCollection.external_declarations[existingExternNum].guid.ToString()))
                {
                    Console.WriteLine(string.Format("Warning: external declaration '{0}' @0x{1:X8} was part of the original debug symbol data, but was not found in the decompiled code exported from Ghidra.",
                        symbolCollection.external_declarations[existingExternNum].name, symbolCollection.external_declarations[existingExternNum].offset_in_memory));
                }
            }

            for (int overlayNum = 0; overlayNum < symbolCollection.overlays.Count; overlayNum++)
            {
                if (overlayIDs.Contains(symbolCollection.overlays[overlayNum].overlay_id))
                {
                    for (int existingExternNum = 0; existingExternNum < symbolCollection.overlays[overlayNum].external_declarations.Count; existingExternNum++)
                    {
                        if (!matchedExternGUIDs.Contains(symbolCollection.overlays[overlayNum].external_declarations[existingExternNum].guid.ToString()))
                        {
                            Console.WriteLine(string.Format("Warning: external declaration '{0}' @0x{1:X8} was part of the original debug symbol data in overlay 0x{2:X2}, but was not found in the decompiled code exported from Ghidra.",
                                symbolCollection.overlays[overlayNum].external_declarations[existingExternNum].name, 
                                symbolCollection.overlays[overlayNum].external_declarations[existingExternNum].offset_in_memory,
                                symbolCollection.overlays[overlayNum].overlay_id));
                        }
                    }
                }
            }

            // create and populate individual source files
            string baseProgramDirectoryName = BenLincoln.DebugSymbols.MetaGenerator.SubsetNameMainBinary;
            //if (includedOverlays.Count == 0)
            //{
            //    baseProgramDirectoryName = "";
            //}
            GenerateSkeletonFilesForSubsetResult gsffs = BenLincoln.DebugSymbols.MetaGenerator.GenerateSkeletonFilesForSubset(projectName, -1, baseProgramDirectoryName, baseOutputDirectoryPath,
                null, false, symbolCollection.functions, symbolCollection.sld_data, symbolCollection.labels, symbolCollection.enums,
                symbolCollection.structs, symbolCollection.unions, symbolCollection.typedefs, symbolCollection.generated_typedefs,
                symbolCollection.external_declarations, "PopulateSkeleton", BenLincoln.DebugSymbols.MetaGenerator.DirectoryNameSourceDecompiled,
                asm, ProgramName, Copyright, InfoURL, ThanksList);
            //symbolCollection.AddFunctions(gsffs.AddedFunctions);
            // overlays
            foreach (PsyQOverlayDescription ol in includedOverlays)
            {
                GenerateSkeletonFilesForSubsetResult gsffso = BenLincoln.DebugSymbols.MetaGenerator.GenerateSkeletonFilesForSubset(projectName, ol.overlay_id, ol.GetDirectoryNameForDecompilation(),
                    baseOutputDirectoryPath, null, false, ol.functions, ol.sld_data, ol.labels, ol.enums,
                    ol.structs, ol.unions, ol.typedefs, ol.generated_typedefs, ol.external_declarations, "PopulateSkeleton", 
                    BenLincoln.DebugSymbols.MetaGenerator.DirectoryNameSourceDecompiled,
                    asm, ProgramName, Copyright, InfoURL, ThanksList);
                ol.AddFunctions(gsffso.AddedFunctions);
            }

            if (includedOverlays.Count == 0)
            {
                BenLincoln.DebugSymbols.Utilities.HandleWarningForOverlays(symbolCollection);
            }

            BenLincoln.DebugSymbols.Utilities.WriteToOutput(manualChangesPath, manualChangesBuilder.ToString(), FileMode.Create);

            //// create and populate individual source files
            //foreach (string sourceFileName in codeSourceTracker.SourceFiles.Keys)
            //{
            //    string stubSourceFilePath = sourceFileName;
            //    // remove any leading drive letter if the code was compiled on Windows
            //    Regex driveLetterRex = new Regex(@"^[A-Za-z]:");
            //    if (driveLetterRex.IsMatch(stubSourceFilePath))
            //    {
            //        stubSourceFilePath = stubSourceFilePath.Substring(2);
            //    }
            //    // replace forward and backslashes with the path separator for the current platform
            //    stubSourceFilePath = stubSourceFilePath.Replace("/", Path.DirectorySeparatorChar.ToString());
            //    stubSourceFilePath = stubSourceFilePath.Replace(@"\", Path.DirectorySeparatorChar.ToString());
            //    // make path relative to output directory
            //    while (stubSourceFilePath.StartsWith(Path.DirectorySeparatorChar.ToString()))
            //    {
            //        stubSourceFilePath = stubSourceFilePath.Substring(1);
            //    }
            //    string stubTemp = stubSourceFilePath;
            //    stubSourceFilePath = Path.Combine(baseOutputDirectoryPath, "source-decompiled", stubTemp);

            //    // get a list of unique line numbers for methods from the file, and sort by that
            //    List<long> lineNumbers = new List<long>();
            //    foreach (CFunction f in codeSourceTracker.SourceFiles[sourceFileName])
            //    {
            //        if (!lineNumbers.Contains(f.source_file_line_number))
            //        {
            //            lineNumbers.Add(f.source_file_line_number);
            //        }
            //    }
            //    lineNumbers.Sort();

            //    StringBuilder headerFileBuilder = new StringBuilder();
            //    StringBuilder sourceFileStubBuilder = new StringBuilder();

            //    string headerFilePath = Path.ChangeExtension(stubSourceFilePath, "H");
            //    string headerFileName = Path.GetFileName(headerFilePath);
            //    string headerFileDefString = headerFileName.Replace(".", "_");

            //    bool headerAndSourceAreSameFile = false;
            //    if (Path.GetFileName(stubSourceFilePath) == headerFileName)
            //    {
            //        headerAndSourceAreSameFile = true;
            //    }

            //    headerFileBuilder.Append(String.Format("#ifndef {0}", headerFileDefString));
            //    headerFileBuilder.Append(Environment.NewLine);
            //    headerFileBuilder.Append(String.Format("#define {0}", headerFileDefString));
            //    headerFileBuilder.Append(Environment.NewLine);
            //    headerFileBuilder.Append(Environment.NewLine);

            //    // handle edge case: source file IS header file
            //    // i.e. if functions are defined in a .H file instead of just declared in it.
            //    if (!headerAndSourceAreSameFile)
            //    {
            //        sourceFileStubBuilder.Append(String.Format("#include \"{0}.H\"", projectName));
            //        sourceFileStubBuilder.Append(Environment.NewLine);
            //        sourceFileStubBuilder.Append(String.Format("#include \"{0}\"", headerFileName));
            //        sourceFileStubBuilder.Append(Environment.NewLine);
            //    }
            //    sourceFileStubBuilder.Append(Environment.NewLine);

            //    foreach (int lineNumber in lineNumbers)
            //    {
            //        foreach (CFunction f in codeSourceTracker.FunctionSourceFiles[sourceFileName])
            //        {
            //            if (f.source_file_line_number == lineNumber)
            //            {
            //                List<DecompiledSourceCodeMatch> matches = new List<DecompiledSourceCodeMatch>();
            //                float highestMatchScore = 0.0f;
            //                foreach (CFunction otherFunc in decompiledFunctions)
            //                {
            //                    if (f.offset_in_memory == otherFunc.offset_in_memory)
            //                    {
            //                        DecompiledSourceCodeMatch matchingFunc = new DecompiledSourceCodeMatch();
            //                        matchingFunc.DecompiledFunction = otherFunc;
            //                        matchingFunc.MatchPercent = 100.0f;
            //                        highestMatchScore = 100.0f;
            //                        matches.Add(matchingFunc);
            //                        break;
            //                    }
            //                }
            //                string functionCode = "";
            //                if (matches.Count > 0)
            //                {
            //                    DecompiledSourceCodeMatch selectedFunc = matches[0];
            //                    if (matches.Count > 1)
            //                    {
            //                        bool foundBestMatch = false;
            //                        if (matches[0].MatchPercent == highestMatchScore)
            //                        {
            //                            foundBestMatch = true;
            //                        }
            //                        bool multipleBestMatches = false;
            //                        for (int i = 1; i < matches.Count; i++)
            //                        {
            //                            if (matches[i].MatchPercent == highestMatchScore)
            //                            {
            //                                if (!foundBestMatch)
            //                                {
            //                                    selectedFunc = matches[i];
            //                                    foundBestMatch = true;
            //                                }
            //                                else
            //                                {
            //                                    multipleBestMatches = true;
            //                                }
            //                                break;
            //                            }
            //                        }
            //                        if (multipleBestMatches)
            //                        {
            //                            Console.WriteLine(string.Format("Warning: {0} matches when searching for decompiled function '{1}'. The best match was {2:F2} percent, and there were multiple matches with this score. You should manually verify the result.", matches.Count, f.name, highestMatchScore));
            //                        }
            //                        else
            //                        {
            //                            Console.WriteLine(string.Format("Warning: {0} matches when searching for decompiled function '{1}'. The best match was {2:F2} percent. You should manually verify the result.", matches.Count, f.name, highestMatchScore));
            //                        }
            //                    }
            //                    else
            //                    {
            //                        Console.WriteLine(string.Format("Found 1 match for function '{0}', with a score of {1:F2} percent.", f.name, highestMatchScore));
            //                    }
            //                    matchedFunctionGUIDs.Add(selectedFunc.DecompiledFunction.guid);
            //                    StringBuilder functionCodeBuilder = new StringBuilder();
            //                    functionCodeBuilder.Append(matches[0].DecompiledFunction.GetFullCode());
            //                    functionCode = functionCodeBuilder.ToString();
            //                    f.source_code = functionCode;
            //                }
            //                else
            //                {
            //                    Console.WriteLine(string.Format("Did not find any matches for function '{0}' in the decompiled code. A stub will be created instead.", f.name));
            //                }

            //                headerFileBuilder.Append(f.GetHeaderCode());
            //                headerFileBuilder.Append(Environment.NewLine);
            //                headerFileBuilder.Append(Environment.NewLine);

            //                sourceFileStubBuilder.Append(f.GetSourceCode());
            //                sourceFileStubBuilder.Append(Environment.NewLine);
            //                sourceFileStubBuilder.Append(Environment.NewLine);
            //            }
            //        }
            //    }

            //    if (headerAndSourceAreSameFile)
            //    {
            //        headerFileBuilder.Append("%STUBCONTENT%");
            //        headerFileBuilder.Append(Environment.NewLine);
            //    }

            //    headerFileBuilder.Append(String.Format("#endif", headerFileDefString));

            //    string parentPath = Path.GetDirectoryName(stubSourceFilePath);
            //    if (!Directory.Exists(parentPath))
            //    {
            //        try
            //        {
            //            Directory.CreateDirectory(parentPath);
            //        }
            //        catch (Exception ex)
            //        {
            //            Console.WriteLine(String.Format("Could not create a required output path ('{0}'): {1}", parentPath, ex.Message));
            //            Environment.Exit(5);
            //        }
            //    }
            //    string headerFileContent = headerFileBuilder.ToString();
            //    string sourceFileStubContent = sourceFileStubBuilder.ToString();
            //    if (headerAndSourceAreSameFile)
            //    {
            //        headerFileContent = headerFileContent.Replace("%STUBCONTENT%", sourceFileStubContent);
            //        BenLincoln.DebugSymbols.Utilities.WriteToOutput(headerFilePath, headerFileContent, FileMode.Create);
            //    }
            //    else
            //    {
            //        BenLincoln.DebugSymbols.Utilities.WriteToOutput(headerFilePath, headerFileContent, FileMode.Create);
            //        BenLincoln.DebugSymbols.Utilities.WriteToOutput(stubSourceFilePath, sourceFileStubContent, FileMode.Create);
            //    }
            //}

            //// create a file with all of the unmatched functions from the decompiled source
            //string unmatchedFunctionSourceFilePath = Path.Combine(baseOutputDirectoryPath, "source-decompiled", "Unmatched_Decompiled_Functions.C");
            //StringBuilder unmatchedBuilder = new StringBuilder();
            //for (int i = 0; i < decompiledFunctions.Count; i++)
            //{
            //    if (!matchedFunctionGUIDs.Contains(decompiledFunctions[i].guid))
            //    {
            //        manualChangesBuilder.Append(string.Format("The decompiled function '{0}' was not matched to a source file, and has been placed in '{1}'", decompiledFunctions[i].name, unmatchedFunctionSourceFilePath));
            //        manualChangesBuilder.Append(Environment.NewLine);
            //        unmatchedBuilder.Append(decompiledFunctions[i].GetFunctionStub(decompiledFunctions[i].GetFullCode()));
            //        unmatchedBuilder.Append(Environment.NewLine);
            //        unmatchedBuilder.Append(Environment.NewLine);
            //        unmatchedBuilder.Append(Environment.NewLine);
            //    }
            //}

            //BenLincoln.DebugSymbols.Utilities.WriteToOutput(unmatchedFunctionSourceFilePath, unmatchedBuilder.ToString(), FileMode.Create);

            // create a list of known manual actions that need to be taken by the user
            manualChangesBuilder.Append(Environment.NewLine);
            BenLincoln.DebugSymbols.Utilities.WriteToOutput(manualChangesPath, manualChangesBuilder.ToString(), FileMode.Create);

            Console.WriteLine("");
        }
    }
}
