﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using symdump.symfile;

namespace symdump
{
    public class TypeDef : IEquatable<TypeDef>
    {
        public readonly BaseType BaseType;
        private readonly DerivedType[] _derivedTypes = new DerivedType[6];
        public readonly int SYMFileSequenceNumber;

        public TypeDef(BinaryReader fs, BenLincoln.DebugSymbols.SequenceIssuer issuer)
        {
            SYMFileSequenceNumber = issuer.GetNextNumber();
            var val = fs.ReadUInt16();
            BaseType = (BaseType)(val & 0x0f);
            for (var i = 0; i < 6; ++i)
            {
                var x = (val >> (i * 2 + 4)) & 3;
                _derivedTypes[i] = (DerivedType)x;
            }
        }

        public bool IsFunctionReturnType => _derivedTypes.Contains(DerivedType.FunctionReturnType);

        public bool Equals(TypeDef other)
        {
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;
            return BaseType == other.BaseType && _derivedTypes.SequenceEqual(other._derivedTypes);
        }

        public override string ToString()
        {
            var attributes = string.Join(",", _derivedTypes.Where(e => e != DerivedType.None));
            return attributes.Length == 0 ? BaseType.ToString() : $"{BaseType}({attributes})";
        }

        public string GetCType(TypeInfo typeInfo, string referencingItemName)
        {
            string cType;
            switch (BaseType)
            {
                case BaseType.StructDef:
                    Debug.Assert(!string.IsNullOrEmpty(typeInfo.Tag));
                    cType = $"struct {typeInfo.Tag}";
                    break;
                case BaseType.UnionDef:
                    Debug.Assert(!string.IsNullOrEmpty(typeInfo.Tag));
                    cType = $"union {typeInfo.Tag}";
                    break;
                case BaseType.EnumDef:
                    Debug.Assert(!string.IsNullOrEmpty(typeInfo.Tag));
                    cType = $"enum {typeInfo.Tag}";
                    break;
                case BaseType.Char:
                    cType = "char";
                    break;
                case BaseType.Short:
                    cType = "short";
                    break;
                case BaseType.Int:
                    cType = "int";
                    break;
                case BaseType.Long:
                    cType = "long";
                    break;
                case BaseType.Float:
                    cType = "float";
                    break;
                case BaseType.Double:
                    cType = "double";
                    break;
                case BaseType.UChar:
                    cType = "unsigned char";
                    break;
                case BaseType.UShort:
                    cType = "unsigned short";
                    break;
                case BaseType.UInt:
                    cType = "unsigned int";
                    break;
                case BaseType.ULong:
                    cType = "unsigned long";
                    break;
                case BaseType.Void:
                    cType = "void";
                    break;
                case BaseType.Null:
                    // I don't know if any of these are correct
                    // but telling Ghidra a parameter is "void" doesn't work
                    //cType = "NULL";
                    //cType = "void";
                    cType = "void *";
                    Console.WriteLine(string.Format("Warning: a type with a basetype of NULL was encountered when processing '{0}'. It has been approximated as '{1}' to prevent issues in Ghidra. You should manually correct this definition in the decompiled code if necessary.", referencingItemName, cType));
                    break;
                default:
                    throw new Exception($"Unexpected base type {BaseType}");
            }
            return cType;
        }

        public string GetName(string name, TypeInfo typeInfo)
        {
            var dimIdx = 0;
            var needsParens = false;
            int arrayDTCount = 0;
            int pointerDTCount = 0;
            int functionPointerDTCount = 0;
            foreach (DerivedType dt in _derivedTypes)
            {
                if (dt == DerivedType.Array)
                {
                    arrayDTCount++;
                }
                if (dt == DerivedType.Pointer)
                {
                    pointerDTCount++;
                }
                if (dt == DerivedType.FunctionReturnType)
                {
                    functionPointerDTCount++;
                }
            }
            bool innerPointerIsWrapped = false;
            foreach (var dt in _derivedTypes)
                switch (dt)
                {
                    case DerivedType.None:
                        continue;
                    case DerivedType.Array:
                        Debug.Assert(name != null);
                        //if (name != "")
                        //{
                        //    if (needsParens)
                        //    {
                        //        name = $"({name})";
                        //    }
                        //}
                        name += $"[{typeInfo.Dims[dimIdx]}]";
                        ++dimIdx;
                        needsParens = true;
                        break;
                    case DerivedType.FunctionReturnType:
                        if (name != "")
                        {
                            //name = needsParens ? $"({name})()" : $"{name}()";
                            if (needsParens)
                            {
                                name = $"({name})()";
                            }
                            else
                            {
                                name = $"{name}()";
                            }
                            needsParens = true;
                        }

                        break;
                    case DerivedType.Pointer:
                        //name = $"*{name}";
                        if (!innerPointerIsWrapped)
                        {
                            if (arrayDTCount > 0)
                            {
                                name = $"(%POINTER%{name})";
                            }
                            else
                            {
                                name = $"%POINTER%{name}";
                            }
                            innerPointerIsWrapped = true;
                        }
                        needsParens = true;
                        break;
                }
            string pointerStars = "";
            for (int i = 0; i < pointerDTCount; i++)
            {
                pointerStars += "*";
            }
            name = name.Replace("%POINTER%", pointerStars);
            return name;
        }

        public List<string> GetDerivedTypeList(string name, TypeInfo typeInfo)
        {
            List<string> result = new List<string>();
            foreach (var dt in _derivedTypes)
                switch (dt)
                {
                    case DerivedType.None:
                        result.Add("None");
                        break;
                    case DerivedType.Array:
                        result.Add("Array");
                        break;
                    case DerivedType.FunctionReturnType:
                        result.Add("FunctionReturnType");
                        break;
                    case DerivedType.Pointer:
                        result.Add("Pointer");
                        break;
                }
            return result;
        }

        public string AsCode(string name, TypeInfo typeInfo)
        {
            string cType = GetCType(typeInfo, name);

            name = GetName(name, typeInfo);

            return $"{cType} {name}";
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != GetType()) return false;
            return Equals((TypeDef) obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                return ((int) BaseType * 397) ^ (_derivedTypes != null ? _derivedTypes.GetHashCode() : 0);
            }
        }
    }
}
