using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using symdump.symfile.util;

namespace symdump.symfile
{
    public class TypeInfo : IEquatable<TypeInfo>
    {
        public readonly ClassType ClassType;
        public readonly uint[] Dims;
        public readonly uint Size;
        public string Tag;
        public readonly TypeDef TypeDef;
        public readonly int SYMFileSequenceNumber;
        public string SLDSourceFileName;
        public List<string> ObjectFileNames;
        public bool OriginalWasFake;
        protected SymDumpProcessingConfiguration ProcessingConfig;

        public TypeInfo(BinaryReader reader, bool extended, SymDumpProcessingConfiguration config)
        {
            // comments in this section refer to the decompiled version of DUMPSYM.EXE
            // https://github.com/lab313ru/dumpsym_src/blob/master/main.c
            ClassType = reader.ReadClassType();             // sub_39B(read_word(input));
            TypeDef = reader.ReadTypeDef(config.Issuer);           // sub_504(read_word(input));
            Console.WriteLine(string.Format("Debug: file offset = {0:X8}.", reader.BaseStream.Position));
            Console.WriteLine(string.Format("Debug: size = {0}.", Size));
            Size = reader.ReadUInt32();                     // printf("size %ld ", read_dword(input));
            SYMFileSequenceNumber = config.Issuer.GetNextNumber();
            ProcessingConfig = config;
            ObjectFileNames = new List<string>();
            ObjectFileNames.Add(config.Tracker.CurrentObjectFileData.FileName);

            if (extended)
            {
                Console.WriteLine(string.Format("Debug: file offset = {0:X8}.", reader.BaseStream.Position));
                var n = reader.ReadUInt16();
                Console.WriteLine(string.Format("Debug: n = {0}.", n));
                Dims = new uint[n];
                for (var i = 0; i < n; ++i)
                {
                    Console.WriteLine(string.Format("Debug: file offset = {0:X8}.", reader.BaseStream.Position));
                    Dims[i] = reader.ReadUInt32();
                    Console.WriteLine(string.Format("Debug: Dims[{0}] = {1:X8}.", i, Dims[i]));
                }

                Console.WriteLine(string.Format("Debug: file offset = {0:X8}.", reader.BaseStream.Position));
                Tag = reader.ReadPascalString();
                Console.WriteLine(string.Format("Debug: got tag '{0}'.", Tag));
                //if (Tag == ".0fake")
                //{
                //    Console.WriteLine("Debug");
                //}
                OriginalWasFake = IsFake;
                if (config.RenameForCompatibility)
                {
                    if (OriginalWasFake)
                    {
                        Tag = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(Tag, true);
                    }
                }
                if (config.AutoRenameFakes)
                {
                    if (OriginalWasFake)
                    {
                        string lastObjectFileName = "";
                        if (config.Tracker.ObjectFileDataEntries.Count > 0)
                        {
                            lastObjectFileName = config.Tracker.ObjectFileDataEntries[config.Tracker.ObjectFileDataEntries.Count - 1].FileName;
                        }
                        string lastSLDSourceFileName = "";
                        if (config.Tracker.SLDEntries.Count > 0)
                        {
                            lastSLDSourceFileName = config.Tracker.SLDEntries[config.Tracker.SLDEntries.Count - 1].SLDFile;
                        }
                        Tag = BenLincoln.DebugSymbols.Utilities.GetNewFakeName("TypeInfo", Tag, config.Tracker.CurrentObjectFileData.FileName, config.Tracker.CurrentSLDFile, 
                            lastObjectFileName, lastSLDSourceFileName);
                    }
                }
                //bool doneGettingTags = false;
                //if (Tag == "")
                //{
                //    doneGettingTags = true;
                //}
                //while (!doneGettingTags)
                //{
                //    string nextTag = reader.ReadPascalString();
                //    if (nextTag == "")
                //    {
                //        doneGettingTags = true;
                //    }
                //    else
                //    {
                //        Console.WriteLine(string.Format("Debug: got additional tag '{0}'.", nextTag));
                //    }
                //}
            }
            else
            {
                Dims = new uint[0];
                Tag = null;
            }
        }

        public bool IsFake => Tag != null && new Regex(@"^(\.|_)\d+fake$").IsMatch(Tag);

        public bool Equals(TypeInfo other)
        {
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;

            bool isMatch = true;

            if (ClassType != other.ClassType)
            {
                SymFile.OutputDebugMessage(string.Format("TypeInfo objects don't have matching ClassTypes ('{0}' vs '{1}')", ClassType.ToString(), other.ClassType.ToString()), ProcessingConfig);
                isMatch = false;
            }

            if (!TypeDef.Equals(other.TypeDef))
            {
                SymFile.OutputDebugMessage(string.Format("TypeInfo objects don't have matching TypeDefs ('{0}' vs '{1}')", TypeDef.ToString(), other.TypeDef.ToString()), ProcessingConfig);
                isMatch = false;
            }

            if (!Dims.SequenceEqual(other.Dims))
            {
                SymFile.OutputDebugMessage("TypeInfo objects don't have matching Dims", ProcessingConfig);
                isMatch = false;
            }

            // sometimes PsyQ sets the size in one version of a struct to 0 for some reason that I haven't figured out yet
            if ((Size != 0) && (other.Size != 0)) 
            {
                if (Size != other.Size)
                {
                    SymFile.OutputDebugMessage(string.Format("TypeInfo objects don't have matching sizes ({0} vs {1})", Size, other.Size), ProcessingConfig);
                    isMatch = false;
                }
            }
            if ((!OriginalWasFake) || (!other.OriginalWasFake))
            {
                if (!string.Equals(Tag, other.Tag))
                {
                    SymFile.OutputDebugMessage(string.Format("TypeInfo objects don't have matching tags ('{0}' vs '{1}')", Tag, other.Tag), ProcessingConfig);
                    isMatch = false;
                }
            }
            return isMatch;
        }

        public override string ToString()
        {
            return $"classType={ClassType} typeDef={TypeDef} size={Size}, dims=[{string.Join(",", Dims)}] // {BenLincoln.DebugSymbols.Utilities.GetListAsString(ObjectFileNames)}";
        }

        public string AsCode(string name)
        {
            return TypeDef.AsCode(name, this).Trim();
        }

        public string GetCType()
        {
            return TypeDef.GetCType(this, ToString());
        }

        public string GetName(string name)
        {
            return TypeDef.GetName(name, this);
        }

        public List<string> GetDerivedTypes(string name)
        {
            return TypeDef.GetDerivedTypeList(name, this);
        }

        public BenLincoln.DebugSymbols.SymDump.CTypeDef ToBLDSCTypeDef(string name, string defaultSourceFile)
        {
            BenLincoln.DebugSymbols.SymDump.CTypeDef result = new BenLincoln.DebugSymbols.SymDump.CTypeDef();

            result.name = GetName(name);
            result.c_type = GetCType();
            result.sym_file_sequence_number = SYMFileSequenceNumber;
            result.source_file = defaultSourceFile;
            result.fake = OriginalWasFake;
            if ((SLDSourceFileName != null) && (SLDSourceFileName != ""))
            {
                Console.WriteLine(string.Format("Info: using SLD source file '{0}' for typedef '{2}'", SLDSourceFileName, defaultSourceFile, result.name));
                result.source_file = SLDSourceFileName;
            }
            result.object_files = ObjectFileNames;
            //result.hashcode = GetHashCode();
            result.typedef_signature = AsCode(name);

            return result;
        }

        public string DumpJSON(string name, string sourceFile, string lineIndent)
        {
            BenLincoln.DebugSymbols.SymDump.CTypeDef resultDef = ToBLDSCTypeDef(name, sourceFile);

            return BenLincoln.DebugSymbols.Utilities.IndentString(resultDef.ToJSON(), lineIndent);
            //StringBuilder builder = new StringBuilder();

            //builder.Append(lineIndent);
            //builder.Append("{");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"name\": \"{0}\",", name));
            //builder.Append(Environment.NewLine);
            //if (sourceFile != "")
            //{
            //    builder.Append(lineIndent);
            //    builder.Append(String.Format("\t\"source_file\": \"{0}\",", sourceFile));
            //    builder.Append(Environment.NewLine);
            //}
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"definition\": \"{0}\",", AsCode(name)));
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("}");

            //return builder.ToString();
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != GetType()) return false;
            return Equals((TypeInfo) obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                var hashCode = (int) ClassType;
                hashCode = (hashCode * 397) ^ TypeDef.GetHashCode();
                hashCode = (hashCode * 397) ^ (int) Size;
                hashCode = (hashCode * 397) ^ Dims.GetHashCode();
                if (Tag != null)
                {
                    hashCode = (hashCode * 397) ^ (!IsFake ? Tag.GetHashCode() : 0);
                }
                return hashCode;
            }
        }
    }
}
